// *************************************************************************************************
//
// Horde3D
//   Next-Generation Graphics Engine
// ---------------------------------
// Copyright (C) 2006 Nicolas Schulz
//
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//
// *************************************************************************************************

#pragma once

#ifndef DLL
#	if defined( WIN32 ) || defined( _WINDOWS )
#		define DLL extern "C" __declspec( dllimport )
#	else
#		define DLL
#	endif
#endif

#ifndef DLL_CLASS
#	if defined( WIN32 ) || defined( _WINDOWS )
#		define DLL_CLASS __declspec( dllimport )
#	else
#		define DLL_CLASS
#	endif
#endif


// *************************************************************************************************
// Core library
// *************************************************************************************************

// Types
typedef unsigned int ResHandle;
typedef unsigned int NodeHandle;

// Predefined constants
const NodeHandle RootNode = 1;
const NodeHandle PrimeTimeCam = 2;

// Flags
namespace ResourceFlags
{
	enum List
	{
		NoQuery = 1,
		NoCompression = 2
	};
}


namespace EngineOptions
{
	enum List
	{
		TrilinearFiltering,
		AnisotropyFactor,
		TexCompression,
		LoadTextures,
		FastAnimation,
		OcclusionCulling,
		ShadowMapSize,
		DebugViewMode
	};
}


namespace ResourceTypes
{
	enum List
	{
		SceneGraph,
		Geometry,
		Animation,
		Material,
		Code,
		Shader,
		Texture2D,
		TextureCube,
		Effect,
		Undefined = 9999
	};
}


namespace ResourceData
{
	enum List
	{
		AnimFrameCount,
		Tex2DPixelData
	};
}


namespace SceneNodeTypes
{
	enum List
	{
		Group,
		Model,
		Mesh,
		Joint,
		Light,
		Camera,
		Emitter,
		Undefined = 9999
	};
}


namespace GroupNodeParams
{
	enum List
	{
		MinDist,
		MaxDist
	};
}


namespace MeshNodeParams
{
	enum List
	{
		MaterialRes
	};
}


namespace LightNodeParams
{
	enum List
	{
		MaterialRes,
		Radius,
		FOV,
		Col_R,
		Col_G,
		Col_B,
		ShadowMapCount,
		ShadowSplitLambda,
		ShadowMapBias
	};
}


namespace CameraNodeParams
{
	enum List
	{
		LeftPlane,
		RightPlane,
		BottomPlane,
		TopPlane,
		NearPlane,
		FarPlane
	};
}


namespace EmitterNodeParams
{
	enum List
	{
		Delay,
		EmissionRate,
		SpreadAngle,
		ForceX,
		ForceY,
		ForceZ
	};
}


namespace Horde3D
{
	// --- Basic funtions ---
	DLL const char *getVersionString();
	DLL	bool init();
	DLL void release();
	DLL void resize( int x, int y, int width, int height );
	DLL void render();
	DLL void clear();
	DLL bool loadPipelineConfig( const char *configFilename );


	// --- General functions ---
	DLL const char *getMessage( unsigned int *level, float *time );
	DLL NodeHandle getActiveCamera();
	DLL bool setActiveCamera( NodeHandle camNode );
	DLL float getOption( EngineOptions::List param );
	DLL bool setOption( EngineOptions::List param, float value );
	DLL void showOverlay( float x_ll, float y_ll, float u_ll, float v_ll,
						  float x_lr, float y_lr, float u_lr, float v_lr,
						  float x_ur, float y_ur, float u_ur, float v_ur,
						  float x_ul, float y_ul, float u_ul, float v_ul,
						  unsigned int layer, ResHandle material );


	// --- Resource functions ---
	DLL ResourceTypes::List getResourceType( ResHandle res );
	DLL ResHandle findResource( ResourceTypes::List type, const char *name );
	DLL ResHandle addResource( ResourceTypes::List type, const char *name, unsigned int flags );
	DLL bool removeResource( ResHandle res );
	DLL bool loadResource( const char *name, const char *data, unsigned int size );
	DLL bool unloadResource( ResHandle res );
	DLL const void *getResourceData( ResHandle res, ResourceData::List param );
	DLL bool updateResourceData( ResHandle res, ResourceData::List param, const void *data, unsigned int size );
	DLL const char *queryUnloadedResource();
	DLL const char *getResourcePath( ResourceTypes::List type );
	DLL void setResourcePath( ResourceTypes::List type, const char *path );
	DLL void releaseUnusedResources();

	// Material specific
	DLL bool setMaterialUniform( ResHandle matRes, const char *name, float a, float b, float c, float d );


	// --- Scene graph functions ---
	DLL SceneNodeTypes::List getNodeType( NodeHandle node );
	DLL const char *getNodeName( NodeHandle node );
	DLL bool setNodeName( NodeHandle node, const char *name );
	DLL NodeHandle getNodeParent( NodeHandle node );
	DLL NodeHandle getNodeChild( NodeHandle parent, const char *name, unsigned int index, bool recursive );
	DLL NodeHandle addNodes( NodeHandle parent, ResHandle sgRes );
	DLL bool removeNode( NodeHandle node );
	DLL bool setNodeActivation( NodeHandle node, bool active );
	DLL bool getNodeTransform( NodeHandle node, float *px, float *py, float *pz,
							   float *rx, float *ry, float *rz, float *sx, float *sy, float *sz );
	DLL bool setNodeTransform( NodeHandle node, float px, float py, float pz,
							   float rx, float ry, float rz, float sx, float sy, float sz );
	DLL bool getNodeTransformMatrices( NodeHandle node, const float **relMat, const float **absMat );
	DLL bool setNodeTransformMatrix( NodeHandle node, const float *mat4x4 );
	DLL bool getNodeAABB( NodeHandle node, float *minX, float *minY, float *minZ,
						  float *maxX, float *maxY, float *maxZ );

	// Group specific
	DLL NodeHandle addGroupNode( NodeHandle parent, const char *name );
	DLL float getGroupParam( NodeHandle node, GroupNodeParams::List param );
	DLL bool setGroupParam( NodeHandle node, GroupNodeParams::List param, float value );
	
	// Model specific
	DLL NodeHandle addModelNode( NodeHandle parent, const char *name, ResHandle geoRes );
	DLL bool setupModelAnimStage( NodeHandle node, unsigned int stage, ResHandle res,
								  const char *animMask, bool additive );
	DLL bool setModelAnimParams( NodeHandle node, unsigned int stage, float time, float weight );
	DLL bool setModelMorpher( NodeHandle node, const char *target, float weight );
	
	// Mesh specific
	DLL NodeHandle addMeshNode( NodeHandle parent, const char *name, ResHandle matRes, 
								unsigned int batchStart, unsigned int batchCount,
								unsigned int vertRStart, unsigned int vertREnd );
	DLL float getMeshParam( NodeHandle node, MeshNodeParams::List param );
	DLL bool setMeshParam( NodeHandle node, MeshNodeParams::List param, float value );
	DLL bool getMeshData( NodeHandle node, unsigned int *vertPosArrayCount, const float **vertPosArrayData,
						  unsigned int *indexArrayCount, const unsigned int **indexArayData, unsigned int *indexOffset );
	DLL bool setMeshUniform( NodeHandle node, const char *name, float a, float b, float c, float d, bool recursive );
	
	// Joint specific
	DLL NodeHandle addJointNode( NodeHandle parent, const char *name, unsigned int jointIndex );
	
	// Light specific
	DLL NodeHandle addLightNode( NodeHandle parent, const char *name, ResHandle materialRes,
								 const char *lightingContext, const char *shadowContext );
	DLL float getLightParam( NodeHandle node, LightNodeParams::List param );
	DLL bool setLightParam( NodeHandle node, LightNodeParams::List param, float value );

	// Camera specific
	DLL NodeHandle addCameraNode( NodeHandle parent, const char *name );
	DLL float getCameraParam( NodeHandle node, CameraNodeParams::List param );
	DLL bool setCameraParam( NodeHandle node, CameraNodeParams::List param, float value );
	DLL bool setupCameraView( NodeHandle node, float fov, float aspect, float nearDist, float farDist );

	// Emitter specific
	DLL NodeHandle addEmitterNode( NodeHandle parent, const char *name,
								   ResHandle matRes, ResHandle effectRes,
								   unsigned int maxParticleCount, int respawnCount );
	DLL float getEmitterParam( NodeHandle node, EmitterNodeParams::List param );
	DLL bool setEmitterParam( NodeHandle node, EmitterNodeParams::List param, float value );
	DLL bool advanceEmitterTime( NodeHandle node, float timeDelta );
}


// *************************************************************************************************
// Integration library
// *************************************************************************************************

typedef void (*NodeAttFactoryFunc)( unsigned int node, const char *attachmentString );
typedef void (*NodeAttCollectorFunc)( unsigned int node );


class DLL_CLASS NodeAttachment
{
protected:
	
	unsigned int	_node;

public:

	static NodeAttFactoryFunc getFactory();
	static void setFactory( NodeAttFactoryFunc func );
	static NodeAttCollectorFunc getCollector();
	static void setCollector( NodeAttCollectorFunc func );
	static NodeAttachment *getFromNode( NodeHandle node );
	
	NodeAttachment( NodeHandle node );
	virtual ~NodeAttachment();
	virtual void onUpdate();

	void updateSceneGraph();
};
