// *************************************************************************************************
//
// Horde3D
//   Next-Generation Graphics Engine
//
// Sample Application
// ---------------------------------
// Copyright (C) 2006 Nicolas Schulz
//
//
// This sample source file is not covered by the LGPL as the rest of the SDK
// and may be used without any restrictions
//
// *************************************************************************************************

#include "app.h"
#include "Horde3D.h"
#include "Horde3DUtils.h"
#include <math.h>
#include <iomanip>


// Convert from degrees to radians
inline float degToRad( float f ) 
{
	return f * (3.1415926f / 180.0f);
}


Application::Application()
{
	for( unsigned int i = 0; i < 256; ++i ) _keys[i] = false;

	_x = 0; _y = 2; _z = 0; _rx = 0; _ry = 0; _velocity = 0.3f;
	_curFPS = 30; _timer = 0;

	_freeze = false; _showFPS = false; _debugViewMode = false;
}


bool Application::init()
{	
	// Initialize engine
	if( !Horde3D::init() )
	{	
		Horde3DUtils::dumpMessages();
		return false;
	}
	
	// Set paths for resources
	Horde3D::setResourcePath( ResourceTypes::SceneGraph, "models" );
	Horde3D::setResourcePath( ResourceTypes::Geometry, "models" );
	Horde3D::setResourcePath( ResourceTypes::Animation, "models" );
	Horde3D::setResourcePath( ResourceTypes::Material, "materials" );
	Horde3D::setResourcePath( ResourceTypes::Code, "shaders" );
	Horde3D::setResourcePath( ResourceTypes::Shader, "shaders" );
	Horde3D::setResourcePath( ResourceTypes::Texture2D, "textures" );
	Horde3D::setResourcePath( ResourceTypes::TextureCube, "textures" );
	Horde3D::setResourcePath( ResourceTypes::Effect, "effects" );

	// Load pipeline configuration
	if( !Horde3D::loadPipelineConfig( "../pipeline_Chicago.xml" ) )
	{	
		Horde3DUtils::dumpMessages();
		return false;
	}

	// Set options
	Horde3D::setOption( EngineOptions::LoadTextures, 1 );
	Horde3D::setOption( EngineOptions::TexCompression, 0 );
	Horde3D::setOption( EngineOptions::AnisotropyFactor, 8 );
	Horde3D::setOption( EngineOptions::ShadowMapSize, 2048 );

	// Add resources
	// Font
	_fontMatRes = Horde3D::addResource( ResourceTypes::Material, "font.material.xml", 0 );
	// Logo
	_logoMatRes = Horde3D::addResource( ResourceTypes::Material, "logo.material.xml", 0 );
	// Shader for deferred shading
	ResHandle lightMatRes = Horde3D::addResource( ResourceTypes::Material, "light.material.xml", 0 );
	// Environment
	ResHandle envRes = Horde3D::addResource( ResourceTypes::SceneGraph, "scene.scene.xml", 0 );
	// Skybox
	ResHandle skyBoxRes = Horde3D::addResource( ResourceTypes::SceneGraph, "skybox.scene.xml", 0 );
	
	// Load resources
	Horde3DUtils::loadResourcesFromDisk( "../Content" );

	// Add scene nodes
	// Add environment
	Horde3D::addNodes( RootNode, envRes );
	// Add skybox
	NodeHandle sky = Horde3D::addNodes( RootNode, skyBoxRes );
	Horde3D::setNodeTransform( sky, 0, 0, 0, 0, 0, 0, 210, 50, 210 );
	// Add light source
	NodeHandle light = Horde3D::addLightNode( RootNode, "Light1", lightMatRes, "LIGHTING", "SHADOWMAP" );
	Horde3D::setNodeTransform( light, 0, 25, -25, -120, 0, 0, 1, 1, 1 );
	Horde3D::setLightParam( light, LightNodeParams::Radius, 100 );
	Horde3D::setLightParam( light, LightNodeParams::FOV, 90 );
	Horde3D::setLightParam( light, LightNodeParams::ShadowMapCount, 3 );
	Horde3D::setLightParam( light, LightNodeParams::ShadowSplitLambda, 0.95f );
	Horde3D::setLightParam( light, LightNodeParams::Col_R, 0.98f );
	Horde3D::setLightParam( light, LightNodeParams::Col_G, 0.6f );
	Horde3D::setLightParam( light, LightNodeParams::Col_B, 0.5f );

	// Add light source
	light = Horde3D::addLightNode( RootNode, "Light2", lightMatRes, "LIGHTING", "SHADOWMAP" );
	Horde3D::setNodeTransform( light, 0, 25, 30, -60, 0, 0, 1, 1, 1 );
	Horde3D::setLightParam( light, LightNodeParams::Radius, 50 );
	Horde3D::setLightParam( light, LightNodeParams::FOV, 90 );
	Horde3D::setLightParam( light, LightNodeParams::ShadowMapCount, 3 );
	Horde3D::setLightParam( light, LightNodeParams::ShadowSplitLambda, 0.95f );
	Horde3D::setLightParam( light, LightNodeParams::Col_R, 0.5f );
	Horde3D::setLightParam( light, LightNodeParams::Col_G, 0.6f );
	Horde3D::setLightParam( light, LightNodeParams::Col_B, 0.98f );

	_crowdSim = new CrowdSim();
	_crowdSim->init();

	return true;
}


void Application::mainLoop( float fps )
{
	_curFPS = fps;
	_timer += 1 / fps;

	keyHandler();
	
	Horde3D::setOption( EngineOptions::DebugViewMode, _debugViewMode ? 1.0f : 0.0f );
	
	if( !_freeze )
	{
		_crowdSim->update( _curFPS );
	}
	
	// Set camera parameters
	Horde3D::setNodeTransform( PrimeTimeCam, _x, _y, _z, _rx ,_ry, 0, 1, 1, 1 );
	
	if( _showFPS )
	{
		// Avoid updating FPS text every frame to make it readable
		if( _timer > 0.3f )
		{
			_fpsText.str( "" );
			_fpsText << "FPS: " << fixed << setprecision( 2 ) << _curFPS;
			_timer = 0;
		}
		
		// Show text
		Horde3DUtils::showText( _fpsText.str().c_str(), 0, 0.95f, 0.03f, 0, _fontMatRes );
	}

	// Show logo
	Horde3D::showOverlay( 0.75f, 0, 0, 0, 1, 0, 1, 0,
						  1, 0.2f, 1, 1, 0.75f, 0.2f, 0, 1,
						  7, _logoMatRes );
	
	// Render scene
	Horde3D::render();

	// Write all mesages to log file
	Horde3DUtils::dumpMessages();
}


void Application::release()
{
	delete _crowdSim; _crowdSim = 0x0;
	
	// Release engine
	Horde3D::release();
}


void Application::resize( int width, int height )
{
	// Resize viewport
	Horde3D::resize( 0, 0, width, height );
	
	// Set virtual camera parameters
	Horde3D::setupCameraView( PrimeTimeCam, 45.0f, (float)width / height, 0.1f, 1000.0f );
}


void Application::keyPressEvent( int keyCode )
{
	if( keyCode == 32 )		// Space
		_freeze = !_freeze;
	
	if( keyCode == 288 )	// F7
		_debugViewMode = !_debugViewMode;
	
	if( keyCode == 290 )	// F9
		_showFPS = !_showFPS;
}


void Application::keyHandler()
{
	float curVel = _velocity * (30 / _curFPS);
	
	if( _keys[119] )	// w
	{
		// Move forward
		_x -= sinf( degToRad( _ry ) ) * cosf( -degToRad( _rx ) ) * curVel;
		_y -= sinf( -degToRad( _rx ) ) * curVel;
		_z -= cosf( degToRad( _ry ) ) * cosf( -degToRad( _rx ) ) * curVel;
	}

	if( _keys[115] )	// s
	{
		// Move backward
		_x += sinf( degToRad( _ry ) ) * cosf( -degToRad( _rx ) ) * curVel;
		_y += sinf( -degToRad( _rx ) ) * curVel;
		_z += cosf( degToRad( _ry ) ) * cosf( -degToRad( _rx ) ) * curVel;
	}

	if( _keys[97] )		// s
	{
		// Strafe left
		_x += sinf( degToRad( _ry - 90) ) * curVel;
		_z += cosf( degToRad( _ry - 90 ) ) * curVel;
	}

	if( _keys[100] )	// d
	{
		// Strafe right
		_x += sinf( degToRad( _ry + 90 ) ) * curVel;
		_z += cosf( degToRad( _ry + 90 ) ) * curVel;
	}
}


void Application::mouseMoveEvent( float dX, float dY )
{
	// Look left/right
	_ry -= dX / 100 * 30;
	
	// Loop up/down but only in a limited range
	_rx += dY / 100 * 30;
	if( _rx > 90 ) _rx = 90; 
	if( _rx < -90 ) _rx = -90;
}
