// *************************************************************************************************
//
// Horde3D
//   Next-Generation Graphics Engine
//
// Sample Application
// ---------------------------------
// Copyright (C) 2006 Nicolas Schulz
//
//
// This sample source file is not covered by the LGPL as the rest of the SDK
// and may be used without any restrictions
//
// *************************************************************************************************

#include "app.h"
#include "Horde3D.h"
#include "Horde3DUtils.h"
#include <math.h>
#include <iomanip>


// Convert from degrees to radians
inline float degToRad( float f ) 
{
	return f * (3.1415926f / 180.0f);
}


Application::Application()
{
	for( unsigned int i = 0; i < 256; ++i ) _keys[i] = false;

	_x = -5; _y = 3; _z = -19; _rx = 7; _ry = 195; _velocity = 0.3f;
	_curFPS = 30; _timer = 0;

	_freeze = false; _showFPS = false; _debugViewMode = false;
	_animTime = 0; _weight = 1.0f;
}


bool Application::init()
{	
	// Initialize engine
	if( !Horde3D::init() )
	{	
		Horde3DUtils::dumpMessages();
		return false;
	}
	
	// Set paths for resources
	Horde3D::setResourcePath( ResourceTypes::SceneGraph, "models" );
	Horde3D::setResourcePath( ResourceTypes::Geometry, "models" );
	Horde3D::setResourcePath( ResourceTypes::Animation, "models" );
	Horde3D::setResourcePath( ResourceTypes::Material, "materials" );
	Horde3D::setResourcePath( ResourceTypes::Code, "shaders" );
	Horde3D::setResourcePath( ResourceTypes::Shader, "shaders" );
	Horde3D::setResourcePath( ResourceTypes::Texture2D, "textures" );
	Horde3D::setResourcePath( ResourceTypes::TextureCube, "textures" );
	Horde3D::setResourcePath( ResourceTypes::Effect, "effects" );

	// Load pipeline configuration
	if( !Horde3D::loadPipelineConfig( "../pipeline_Knight.xml" ) )
	{	
		Horde3DUtils::dumpMessages();
		return false;
	}

	// Set options
	Horde3D::setOption( EngineOptions::LoadTextures, 1 );
	Horde3D::setOption( EngineOptions::TexCompression, 0 );
	Horde3D::setOption( EngineOptions::FastAnimation, 0 );
	Horde3D::setOption( EngineOptions::AnisotropyFactor, 8 );
	Horde3D::setOption( EngineOptions::ShadowMapSize, 2048 );

	// Add resources
	// Font
	_fontMatRes = Horde3D::addResource( ResourceTypes::Material, "font.material.xml", 0 );
	// Logo
	_logoMatRes = Horde3D::addResource( ResourceTypes::Material, "logo.material.xml", 0 );
	// Environment
	ResHandle envRes = Horde3D::addResource( ResourceTypes::SceneGraph, "scene.scene.xml", 0 );
	// Knight
	ResHandle knightRes = Horde3D::addResource( ResourceTypes::SceneGraph, "knight.scene.xml", 0 );
	ResHandle knightAnim1Res = Horde3D::addResource( ResourceTypes::Animation, "knight_order.anim", 0 );
	ResHandle knightAnim2Res = Horde3D::addResource( ResourceTypes::Animation, "knight_attack.anim", 0 );
	// Particle system
	ResHandle particleSysRes = Horde3D::addResource( ResourceTypes::SceneGraph, "particleSys1.scene.xml", 0 );
	
	// Load resources
	Horde3DUtils::loadResourcesFromDisk( "../Content" );

	// Add scene nodes
	// Add environment
	Horde3D::addNodes( RootNode, envRes );
	// Add knight
	_knight = Horde3D::addNodes( RootNode, knightRes );
	Horde3D::setNodeTransform( _knight, 0, 0, 0, 0, 0, 0, 0.1f, 0.1f, 0.1f );
	Horde3D::setupModelAnimStage( _knight, 0, knightAnim1Res, "", false );
	Horde3D::setupModelAnimStage( _knight, 1, knightAnim2Res, "", false );
	// Attach particle system to hand joint
	NodeHandle hand = Horde3D::getNodeChild( _knight, "Bip01_R_Hand", 0, true );
	_particleSys = Horde3D::addNodes( hand, particleSysRes );
	Horde3D::setNodeTransform( _particleSys, 0, 40, 0, 90, 0, 0, 1, 1, 1 );

	// Add light source
	NodeHandle light = Horde3D::addLightNode( RootNode, "Light1", 0, "LIGHTING", "SHADOWMAP" );
	Horde3D::setNodeTransform( light, 0, 1, -15, 30, 180, 0, 1, 1, 1 );
	Horde3D::setLightParam( light, LightNodeParams::Radius, 30 );
	Horde3D::setLightParam( light, LightNodeParams::FOV, 90 );
	Horde3D::setLightParam( light, LightNodeParams::ShadowMapCount, 3 );
	Horde3D::setLightParam( light, LightNodeParams::ShadowMapBias, 0.01f );
	Horde3D::setLightParam( light, LightNodeParams::Col_R, 1.0f );
	Horde3D::setLightParam( light, LightNodeParams::Col_G, 0.7f );
	Horde3D::setLightParam( light, LightNodeParams::Col_B, 0.7f );

	// Customize post processing effects
	NodeHandle matRes = Horde3D::findResource( ResourceTypes::Material, "postprocessing.material.xml" );
	Horde3D::setMaterialUniform( matRes, "radblurParams", 0.007f, 0, 0, 0 );
	
	return true;
}


void Application::mainLoop( float fps )
{
	_curFPS = fps;
	_timer += 1 / fps;

	keyHandler();
	
	Horde3D::setOption( EngineOptions::DebugViewMode, _debugViewMode ? 1.0f : 0.0f );
	
	if( !_freeze )
	{
		_animTime += 1.0f / _curFPS;

		// Do animation blending
		Horde3D::setModelAnimParams( _knight, 0, _animTime * 24.0f, _weight );
		Horde3D::setModelAnimParams( _knight, 1, _animTime * 24.0f, 1.0f - _weight );

		// Animate particle system
		Horde3D::advanceEmitterTime( _particleSys, 1.0f / _curFPS );
	}
	
	// Set camera parameters
	Horde3D::setNodeTransform( PrimeTimeCam, _x, _y, _z, _rx ,_ry, 0, 1, 1, 1 );
	
	if( _showFPS )
	{
		// Avoid updating FPS text every frame to make it readable
		if( _timer > 0.3f )
		{
			_fpsText.str( "" );
			_fpsText << "FPS: " << fixed << setprecision( 2 ) << _curFPS;
			_timer = 0;
		}
		
		// Show text
		Horde3DUtils::showText( _fpsText.str().c_str(), 0, 0.95f, 0.03f, 0, _fontMatRes );

		stringstream text;
		text << "Weight: " << _weight;
		Horde3DUtils::showText( text.str().c_str(), 0, 0.91f, 0.03f, 0, _fontMatRes );
	}

	// Show logo
	Horde3D::showOverlay( 0.75f, 0, 0, 0, 1, 0, 1, 0,
						  1, 0.2f, 1, 1, 0.75f, 0.2f, 0, 1,
						  7, _logoMatRes );
	
	// Render scene
	Horde3D::render();

	// Write all mesages to log file
	Horde3DUtils::dumpMessages();
}


void Application::release()
{
	// Release engine
	Horde3D::release();
}


void Application::resize( int width, int height )
{
	// Resize viewport
	Horde3D::resize( 0, 0, width, height );
	
	// Set virtual camera parameters
	Horde3D::setupCameraView( PrimeTimeCam, 45.0f, (float)width / height, 0.1f, 1000.0f );
}


void Application::keyPressEvent( int keyCode )
{
	if( keyCode == 32 )		// Space
		_freeze = !_freeze;
	
	if( keyCode == 288 )	// F7
		_debugViewMode = !_debugViewMode;
	
	if( keyCode == 290 )	// F9
		_showFPS = !_showFPS;
}


void Application::keyHandler()
{
	float curVel = _velocity * (30 / _curFPS);
	
	if( _keys[119] )	// w
	{
		// Move forward
		_x -= sinf( degToRad( _ry ) ) * cosf( -degToRad( _rx ) ) * curVel;
		_y -= sinf( -degToRad( _rx ) ) * curVel;
		_z -= cosf( degToRad( _ry ) ) * cosf( -degToRad( _rx ) ) * curVel;
	}

	if( _keys[115] )	// s
	{
		// Move backward
		_x += sinf( degToRad( _ry ) ) * cosf( -degToRad( _rx ) ) * curVel;
		_y += sinf( -degToRad( _rx ) ) * curVel;
		_z += cosf( degToRad( _ry ) ) * cosf( -degToRad( _rx ) ) * curVel;
	}

	if( _keys[97] )		// a
	{
		// Strafe left
		_x += sinf( degToRad( _ry - 90) ) * curVel;
		_z += cosf( degToRad( _ry - 90 ) ) * curVel;
	}

	if( _keys[100] )	// d
	{
		// Strafe right
		_x += sinf( degToRad( _ry + 90 ) ) * curVel;
		_z += cosf( degToRad( _ry + 90 ) ) * curVel;
	}

	if( _keys[49] )	// 1
	{
		_weight += 2 / _curFPS;
		if( _weight > 1 ) _weight = 1;
	}
	if( _keys[50] )	// 2
	{
		_weight -= 2 / _curFPS;
		if( _weight < 0 ) _weight = 0;
	}
}


void Application::mouseMoveEvent( float dX, float dY )
{
	// Look left/right
	_ry -= dX / 100 * 30;
	
	// Loop up/down but only in a limited range
	_rx += dY / 100 * 30;
	if( _rx > 90 ) _rx = 90; 
	if( _rx < -90 ) _rx = -90;
}
