// *************************************************************************************************
//
// Horde3D
//   Next-Generation Graphics Engine
// --------------------------------------
// Copyright (C) 2006-2007 Nicolas Schulz
//
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//
// *************************************************************************************************

#if defined( _MSC_VER )
#	if _MSC_VER >= 1400
#		define _CRT_SECURE_NO_DEPRECATE
#	endif
#endif


#include "converter.h"
#include "NVTriStrip.h"
#include "utPlatform.h"
#include <fstream>
#include <sstream>
using namespace std;

#ifdef PLATFORM_WIN
#	include <direct.h>
#else
#	include <sys/stat.h>
#endif


Converter::Converter()
{
	_numAnimFrames = 0;
}


Converter::~Converter()
{
	for( unsigned int i = 0; i < _joints.size(); ++i ) delete _joints[i];
	for( unsigned int i = 0; i < _meshes.size(); ++i ) delete _meshes[i];
}


void Converter::calcTangentSpaceBasis( vector<Vertex> &verts )
{
	for( unsigned int i = 0; i < verts.size(); ++i )
	{
		verts[i].normal = Vec3f( 0, 0, 0 );
		verts[i].tangent = Vec3f( 0, 0, 0 );
		verts[i].bitangent = Vec3f( 0, 0, 0 );
	}
	
	// Algorithm: Eric Lengyel, Mathematics for 3D Game Programming & Computer Graphics
	// Note: Smoothing groups are currently not respected at texture seams
	for( unsigned int i = 0; i < _meshes.size(); ++i )
	{
		for( unsigned int j = 0; j < _meshes[i]->triGroups.size(); ++j )
		{
			TriGroup &triGroup = _meshes[i]->triGroups[j];
			
			for( unsigned int k = triGroup.first; k < triGroup.first + triGroup.count; k += 3 )
			{
				Vec3f &v0 = verts[_indices[k + 0]].pos;
				Vec3f &v1 = verts[_indices[k + 1]].pos;
				Vec3f &v2 = verts[_indices[k + 2]].pos;

				Vec3f &v0_str = verts[_indices[k + 0]].texCoords[0];
				Vec3f &v1_str = verts[_indices[k + 1]].texCoords[0];
				Vec3f &v2_str = verts[_indices[k + 2]].texCoords[0];

				Plane plane( v0, v1, v2 );
				Vec3f xyz0 = v1 - v0;
				Vec3f xyz1 = v2 - v0;
				Vec3f st0 = v1_str - v0_str;
				Vec3f st1 = v2_str - v0_str;

				float r = 0.0f;
				if( st0.x * st1.y - st1.x * st0.y != 0 )
					r = 1.0f / (st0.x * st1.y - st1.x * st0.y);
				Vec3f sdir( (st1.y * xyz0.x - st0.y * xyz1.x) * r,
							(st1.y * xyz0.y - st0.y * xyz1.y) * r, (st1.y * xyz0.z - st0.y * xyz1.z) * r );
				Vec3f tdir( (st0.x * xyz1.x - st1.x * xyz0.x) * r,
							(st0.x * xyz1.y - st1.x * xyz0.y) * r, (st0.x * xyz1.z - st1.x * xyz0.z) * r );

				verts[_indices[k + 0]].normal += plane.normal;
				verts[_indices[k + 1]].normal += plane.normal;
				verts[_indices[k + 2]].normal += plane.normal;
				verts[_indices[k + 0]].tangent += sdir;
				verts[_indices[k + 1]].tangent += sdir;
				verts[_indices[k + 2]].tangent += sdir;
				verts[_indices[k + 0]].bitangent += tdir;
				verts[_indices[k + 1]].bitangent += tdir;
				verts[_indices[k + 2]].bitangent += tdir;
			}
		}
	}
	// Normalize tangent space basis
	for( unsigned int i = 0; i < verts.size(); ++i )
	{
		const Vec3f &n = verts[i].normal.normalized();
		const Vec3f &t = verts[i].tangent;
		verts[i].tangent = (t - n * (n * t)).normalized();
		verts[i].normal = n;
				
		if( n.crossProduct( t ) * verts[i].bitangent < 0.0f )
			verts[i].bitangent = (n * -1).crossProduct( t );
		else
			verts[i].bitangent = n.crossProduct( t );
	}
}


bool Converter::convertModel( SimpleColladaFile &collada, bool optimize )
{
	if( collada.scene == 0x0 ) return true;		// Nothing to convert
	
	DaeVisualScene &visScene = *collada.scene;

	// *********************************************************************************************
	// Add scene nodes (meshes and joints)
	// *********************************************************************************************

	_numAnimFrames = 0;
	
	// Check that all nodes have same number of animation frames (animation sampling)
	for( unsigned int i = 0; i < visScene.nodes.size(); ++i )
	{
		DaeAnimation *anim = visScene.nodes[i]->animation;

		if( anim != 0x0 && anim->matrices->floatArray.size() > 0 )
		{
			unsigned int numFrames = anim->matrices->floatArray.size() / anim->matrices->elemsPerEntry;
			if( _numAnimFrames == 0 ) _numAnimFrames = numFrames;

			if( numFrames != _numAnimFrames )
			{
				log( "Error: Not all animated nodes have same number of animation frames; use sampling for export!" );
				return false;
			}
		}
	}

	// Output default pose if no animation is available
	if( _numAnimFrames == 0 ) _numAnimFrames = 1;
	
	// Loop through all nodes in Collada document
	for( unsigned int i = 0; i < visScene.nodes.size(); ++i )
	{
		DaeNode &iNode = *visScene.nodes[i];
		SceneNode *oNode = 0x0;

		if( iNode.joint )
		{
			oNode = new Joint();
			_joints.push_back( (Joint *)oNode );
		}
		else
		{
			// Check if node is a mesh
			if( iNode.instanceId != "" )
			{
				oNode = new Mesh();
				_meshes.push_back( (Mesh *)oNode );
			}
		}

		// Ignore pure transformation nodes
		if( oNode == 0x0 ) continue;

		// *** Read common attributes ***
		// Name
		if( iNode.name.length() > 255 )
		{
			log( "Warning: node name is too long" );
			iNode.name.erase( 255, iNode.name.length() - 255 );
		}
		strcpy( oNode->name, iNode.name.c_str() );

		// Collada node
		oNode->daeNode = &iNode;

		// Relative transformation
		oNode->matRel = makeMatrix4f( iNode.transMat, collada.y_up );
		oNode->matAbs = oNode->matRel;

		// Apply transformation from pure transformation nodes
		DaeNode *parent = iNode.parent;
		while( parent != 0x0 )
		{
			// Check if hierarchie is valid
			if( !iNode.joint && parent->joint )
			{
				log( "Warning: Mesh " + iNode.name + " is directly attached to a joint; use skin instead!" );
				break;
			}

			if( iNode.joint && parent->joint ) break;
			else if( !iNode.joint && parent->instanceId != "" ) break;
			
			Matrix4f matParent = makeMatrix4f( parent->transMat, collada.y_up );
			oNode->matRel = matParent * oNode->matRel;
			parent = parent->parent;
		}

		// Add parent/child relations
		if( parent != 0x0 )
		{
			// Parent joints and meshes are definately already created since the Collada node
			// list is built by performing a DFS

			if( iNode.joint )
			{
				for( unsigned int j = 0; j < _joints.size(); ++j )
				{
					if( _joints[j]->daeNode == parent )
					{
						_joints[j]->children.push_back( oNode );
						oNode->parent = _joints[j];
						break;
					}
				}
			}
			else
			{
				for( unsigned int j = 0; j < _meshes.size(); ++j )
				{
					if( _meshes[j]->daeNode == parent )
					{
						_meshes[j]->children.push_back( oNode );
						oNode->parent = _meshes[j];
						break;
					}
				}
			}
		}
		
		// Absolute transformation
		parent = iNode.parent;
		while( parent != 0x0 )
		{
			Matrix4f matParent = makeMatrix4f( parent->transMat, collada.y_up );
			oNode->matAbs = matParent * oNode->matAbs;
			parent = parent->parent;
		}
		
		// Animation (assume sampling)
		for( unsigned int j = 0; j < _numAnimFrames; ++j )
		{
			Matrix4f mat;
			if( iNode.animation != 0x0 && iNode.animation->matrices->floatArray.size() > 0 &&
				iNode.animation->matrices->elemsPerEntry == 16 ) 
			{
				mat = makeMatrix4f( &iNode.animation->matrices->floatArray[j * 16], collada.y_up );
			}
			else
			{
				mat = makeMatrix4f( iNode.transMat, collada.y_up );
			}
			
			// Apply transformation from pure transformation nodes
			parent = iNode.parent;
			while( parent != 0x0 )
			{
				if( iNode.joint && parent->joint ) break;
				else if( !iNode.joint && parent->instanceId != "" ) break;
				
				Matrix4f matParent;

				if( parent->animation != 0x0 && parent->animation->matrices->floatArray.size() > 0 )
					matParent = makeMatrix4f( &parent->animation->matrices->floatArray[j * 16], collada.y_up );
				else
					matParent = makeMatrix4f( parent->transMat, collada.y_up );

				mat = matParent * mat;
				parent = parent->parent;
			}
			
			oNode->frames.push_back( mat );
		}
	}

	
	// *********************************************************************************************
	// Joints
	// *********************************************************************************************

	// Check that all joints have same number of animation frames (animation sampling)
	for( unsigned int i = 0; i < _joints.size(); ++i )
	{
		_joints[i]->index = i;
		_joints[i]->invBindMat = _joints[i]->matAbs.inverted();
	}

	if( _joints.size() > 75 ) log( "Warning: Model has more than 75 joints" );


	// *********************************************************************************************
	// Meshes
	// *********************************************************************************************
	
	// Note: At the moment the geometry for all nodes is copied and not referenced
	for( unsigned int i = 0; i < _meshes.size(); ++i )
	{
		// Find geometry/controller for node
		string id = _meshes[i]->daeNode->instanceId;
		
		DaeSkin *skin = collada.libControllers.findSkin( id );
		DaeMorph *morpher = collada.libControllers.findMorph( id );
		
		// Resolve controller stack
		if( skin != 0x0 )
		{	
			id = skin->ownerId;
			morpher = collada.libControllers.findMorph( id );
			if( morpher != 0x0 ) id = morpher->ownerId;
		}
		else if( morpher != 0x0 )
		{
			id = morpher->ownerId;
			skin = collada.libControllers.findSkin( id );
			if( skin != 0x0 ) id = skin->ownerId;
		}

		// Check that skin has all required arrays
		if( skin != 0x0 )
		{
			if( skin->jointArray == 0x0 || skin->bindMatArray == 0x0 || skin->weightArray == 0x0 )
			{
				log( "Skin controller '" + skin->id + "' is missing information and is ignored" );
				skin = 0x0;
			}
		}
		
		DaeGeometry *geo = collada.libGeometries.findGeometry( id );
		if( geo == 0x0 )
		{
			log( "Warning: Geometry or controller " + id + " referenced by scene node not found" );
			continue;
		}
		
		vector< unsigned int > jointLookup;
		if( skin != 0x0 )
		{
			// Build lookup table
			for( unsigned int j = 0; j < skin->jointArray->stringArray.size(); ++j )
			{
				string id = skin->jointArray->stringArray[j];

				jointLookup.push_back( 0 );
				
				bool found = false;
				for( unsigned int k = 0; k < _joints.size(); ++k )
				{
					if( _joints[k]->daeNode != 0x0 && _joints[k]->daeNode->id == id )
					{
						jointLookup[j] = k;
						found = true;
						break;
					}
				}

				if( !found )
				{
					log( "Warning: joint '" + id + "' used in skin controller not found" );
				}
			}
			
			// Find bind matrices
			Matrix4f bindShapeMat = makeMatrix4f( &skin->bindShapeMat[0], collada.y_up );
			for( unsigned int j = 0; j < skin->jointArray->stringArray.size(); ++j )
			{
				if( jointLookup[j] >= 0 )
				{
					_joints[jointLookup[j]]->used = true;
					
					_joints[jointLookup[j]]->daeInvBindMat =
						makeMatrix4f( &skin->bindMatArray->floatArray[j * 16], collada.y_up );

					// Multiply bind matrices with bind shape matrix
					_joints[jointLookup[j]]->daeInvBindMat =
						_joints[jointLookup[j]]->daeInvBindMat * bindShapeMat;
				}
			}
		}
		
		unsigned int firstGeoVert = _vertices.size();
		
		for( unsigned int j = 0; j < geo->triGroups.size(); ++j )
		{
			DaeTriGroup &iTriGroup = geo->triGroups[j];
			TriGroup oTriGroup;

			oTriGroup.matName = _meshes[i]->daeNode->materialBindings[iTriGroup.matId];
			DaeMaterial *mat = collada.libMaterials.findMaterial( oTriGroup.matName );
			if( mat == 0x0 )
				log( "Warning: Material '" + oTriGroup.matName + "' not found" );
			else
				mat->used = true;

			oTriGroup.first = (unsigned int)_indices.size();
			oTriGroup.count = (unsigned int)iTriGroup.indices.size();
			oTriGroup.vertRStart = (unsigned int)_vertices.size();

			// Add indices and vertices
			unsigned int numPosIndices = iTriGroup.vSource->posSource->floatArray.size() /
										 iTriGroup.vSource->posSource->elemsPerEntry;
			vector< unsigned int > *posIndexToVertices = new vector< unsigned int >[numPosIndices];
			
			for( unsigned int k = 0; k < iTriGroup.indices.size(); ++k )
			{
				// Try to find vertex
				vector< unsigned int > &vertList = posIndexToVertices[iTriGroup.indices[k].posIndex];
				bool found = false;
				unsigned int index = (unsigned int)_vertices.size();
				
				// Only check vertices that have the same Collada position index
				for( unsigned int l = 0; l < vertList.size(); ++l )
				{
					Vertex &v = _vertices[vertList[l]];
					
					if( v.storedPos == iTriGroup.getPos( iTriGroup.indices[k].posIndex ) &&
						v.storedNormal == iTriGroup.getNormal( iTriGroup.indices[k].normIndex ) &&
						v.texCoords[0] == iTriGroup.getTexCoords( iTriGroup.indices[k].texIndex[0], 0 ) &&
						v.texCoords[1] == iTriGroup.getTexCoords( iTriGroup.indices[k].texIndex[1], 1 ) &&
						v.texCoords[2] == iTriGroup.getTexCoords( iTriGroup.indices[k].texIndex[2], 2 ) &&
						v.texCoords[3] == iTriGroup.getTexCoords( iTriGroup.indices[k].texIndex[3], 3 ) )
	
					{
						found = true;
						index = vertList[l];
						break;
					}
				}

				if( found )
				{
					_indices.push_back( index );	
				}
				else
				{
					Vertex v;

					v.posIndex = iTriGroup.indices[k].posIndex;
					int normIndex = iTriGroup.indices[k].normIndex;

					// Position
					v.storedPos = iTriGroup.getPos( v.posIndex );
					v.pos = v.storedPos;
					if( !collada.y_up )
					{	
						swap( v.pos.y, v.pos.z );
						v.pos.z *= -1;
					}

					// Texture coordinates
					v.texCoords[0] = iTriGroup.getTexCoords( iTriGroup.indices[k].texIndex[0], 0 );
					v.texCoords[1] = iTriGroup.getTexCoords( iTriGroup.indices[k].texIndex[1], 1 );
					v.texCoords[2] = iTriGroup.getTexCoords( iTriGroup.indices[k].texIndex[2], 2 );
					v.texCoords[3] = iTriGroup.getTexCoords( iTriGroup.indices[k].texIndex[3], 3 );

					// Normal
					v.storedNormal =  iTriGroup.getNormal( normIndex );

					// Skinning
					if( skin != 0x0 )
					{
						DaeVertWeight vertWeights = skin->vertWeights[v.posIndex];
						
						// Sort weights if necessary
						if( vertWeights.size() > 4 )
						{
							for( unsigned int xx = 0; xx < vertWeights.size(); ++xx )
							{
								for( unsigned int yy = 0; yy < vertWeights.size(); ++yy )
								{
									if( skin->weightArray->floatArray[(int)vertWeights[xx].weight] >
										skin->weightArray->floatArray[(int)vertWeights[yy].weight] )
									{
										swap( vertWeights[xx], vertWeights[yy] );
									}
								}
							}
						}
						
						for( unsigned int l = 0; l < vertWeights.size(); ++l )
						{
							if( l == 4 ) break;
							v.weights[l] = skin->weightArray->floatArray[(int)vertWeights[l].weight];
							v.joints[l] = jointLookup[vertWeights[l].joint];
						}

						if( vertWeights.size() > 4 )
						{
							v.weights[0] = 1.0f - (v.weights[1] + v.weights[2] + v.weights[3]);
						}

						// Apply skinning to vertex
						if( !_joints.empty() )
						{
							v.pos =
								_joints[v.joints[0]]->matAbs * _joints[v.joints[0]]->daeInvBindMat * v.pos * v.weights[0] +
								_joints[v.joints[1]]->matAbs * _joints[v.joints[1]]->daeInvBindMat * v.pos * v.weights[1] +
								_joints[v.joints[2]]->matAbs * _joints[v.joints[2]]->daeInvBindMat * v.pos * v.weights[2] +
								_joints[v.joints[3]]->matAbs * _joints[v.joints[3]]->daeInvBindMat * v.pos * v.weights[3];
						}
					}
					
					_vertices.push_back( v );
					_indices.push_back( index );

					vertList.push_back( _vertices.size() - 1 );
				}
			}
			delete[] posIndexToVertices; posIndexToVertices = 0x0;

			oTriGroup.vertREnd = (unsigned int)_vertices.size() - 1;
			
			// Remove degenerated triangles
			unsigned int numDegTris = 0;
			for( unsigned int k = oTriGroup.first; k < oTriGroup.first + oTriGroup.count; k += 3 )
			{
				Vec3f &v0 = _vertices[_indices[k + 0]].pos;
				Vec3f &v1 = _vertices[_indices[k + 1]].pos;
				Vec3f &v2 = _vertices[_indices[k + 2]].pos;

				if( (v2 - v0).crossProduct( v1 - v0 ) == Vec3f( 0, 0, 0 ) )
				{
					++numDegTris;
					// Remove triangle indices
					_indices.erase( _indices.begin() + k + 2 );
					_indices.erase( _indices.begin() + k + 1 );
					_indices.erase( _indices.begin() + k + 0 );
					k -= 3;
					oTriGroup.count -= 3;
				}
			}
			if( numDegTris > 0 )
			{
				stringstream ss;
				ss << numDegTris;
				log( "Removed " + ss.str() + " degenerated triangles from mesh " + visScene.nodes[i]->name );
			}
			
			// Optimize order of indices for vertex cache using NVTriStrip
			if( optimize && oTriGroup.count <= 65535 )
			{
				PrimitiveGroup *primGroups = 0x0, *primGroups2 = 0x0;
				unsigned short numGroups;
				
				SetListsOnly( true );
				SetCacheSize( 16 );
				SetStitchStrips( false );
				
				// Create index list
				unsigned short *indexList = new unsigned short[oTriGroup.count];
				for( unsigned int k = 0; k < oTriGroup.count; ++k )
				{
					indexList[k] = _indices[oTriGroup.first + k] - oTriGroup.vertRStart;
				}
				
				// Optimize indices
				if( GenerateStrips( indexList, oTriGroup.count, &primGroups, &numGroups ) )
				{ 
					unsigned int numVerts = oTriGroup.vertREnd - oTriGroup.vertRStart + 1;
				
					// Remap indices to be as linear-access as possible
					RemapIndices( primGroups, numGroups, numVerts, &primGroups2 );

					// Copy vertices
					vector< Vertex > oldVerts( numVerts );
					for( unsigned int k = 0; k < numVerts; ++k )
					{
						oldVerts[k] = _vertices[oTriGroup.vertRStart + k];
					}
					
					// Reorder vertices
					for( unsigned int k = 0; k < oTriGroup.count; ++k )
					{
						_vertices[primGroups2[0].indices[k] + oTriGroup.vertRStart] =
							oldVerts[primGroups[0].indices[k]];
					}

					// Reorder indices
					for( unsigned int k = 0; k < oTriGroup.count; ++k )
					{
						_indices[oTriGroup.first + k] =
							(unsigned int)primGroups2[0].indices[k] + oTriGroup.vertRStart;
					}
				}

				delete[] indexList;
				delete[] primGroups;
				delete[] primGroups2;
			}

			_meshes[i]->triGroups.push_back( oTriGroup );
		}

		unsigned int numGeoVerts = _vertices.size() - firstGeoVert;

		// Morph targets
		if( morpher != 0x0 && morpher->targetArray != 0x0  )
		{
			for( unsigned int j = 0; j < morpher->targetArray->stringArray.size(); ++j )
			{
				string targetId = morpher->targetArray->stringArray[j];
				DaeGeometry *targetGeo = collada.libGeometries.findGeometry( targetId );
				if( targetGeo == 0x0 ) continue;
				
				// Try to find morph target
				int index = -1;
				for( unsigned int k = 0; k < _morphTargets.size(); ++k )
				{
					if( targetId == _morphTargets[k].name )
					{
						index = k;
						break;
					}
				}
				
				// Add new morph target if not found
				if( index < 0 )
				{
					MorphTarget mt;
					
					if( targetGeo->name != "" )
						strcpy( mt.name, targetGeo->name.c_str() );
					else
						strcpy( mt.name, targetId.c_str() );
					
					_morphTargets.push_back( mt );
					index = _morphTargets.size() - 1;
				}

				// Loop over all vertices (Collada allows only exactly one vertex source per mesh)
				if( geo->vsources[0].posSource->floatArray.size() !=
					targetGeo->vsources[0].posSource->floatArray.size() )
				{
					log( "Warning: morph target geometry differs from base geometry" );
					continue;
				}
				
				for( unsigned int k = 0; k < numGeoVerts; ++k )
				{
					Vertex &v = _vertices[firstGeoVert + k];
					
					Vec3f basePos = geo->getPos( v.posIndex );
					Vec3f targetPos = targetGeo->getPos( v.posIndex );
					
					if( targetPos != basePos )
					{
						if( !collada.y_up )
						{	
							swap( targetPos.y, targetPos.z );
							targetPos.z *= -1;
						}
						
						// Apply skinning to morph target vertex
						if( !_joints.empty() )
						{
							targetPos =
								_joints[v.joints[0]]->matAbs * _joints[v.joints[0]]->daeInvBindMat * targetPos * v.weights[0] +
								_joints[v.joints[1]]->matAbs * _joints[v.joints[1]]->daeInvBindMat * targetPos * v.weights[1] +
								_joints[v.joints[2]]->matAbs * _joints[v.joints[2]]->daeInvBindMat * targetPos * v.weights[2] +
								_joints[v.joints[3]]->matAbs * _joints[v.joints[3]]->daeInvBindMat * targetPos * v.weights[3];
						}
						
						Vec3f diff = targetPos - v.pos;
						
						MorphDiff md;
						md.vertIndex = firstGeoVert + k;
						md.posDiff = diff;
						_morphTargets[index].diffs.push_back( md );
					}
				}
			}	
		}
	}

	// Calculate tangent space basis for base mesh
	calcTangentSpaceBasis( _vertices );

	// Calculate tangent space basis for morph targets
	vector< Vertex > verts( _vertices );
	for( unsigned int i = 0; i < _morphTargets.size(); ++i )
	{
		// Morph
		for( unsigned int j = 0; j < _morphTargets[i].diffs.size(); ++j )
		{	
			verts[_morphTargets[i].diffs[j].vertIndex].pos += _morphTargets[i].diffs[j].posDiff;
		}

		calcTangentSpaceBasis( verts );

		// Find basis differences and undo morphing
		for( unsigned int j = 0; j < _morphTargets[i].diffs.size(); ++j )
		{	
			MorphDiff &md = _morphTargets[i].diffs[j];
			
			verts[md.vertIndex].pos -= md.posDiff;
			md.normDiff = verts[md.vertIndex].normal - _vertices[md.vertIndex].normal;
			md.tanDiff = verts[md.vertIndex].tangent - _vertices[md.vertIndex].tangent;
			md.bitanDiff = verts[md.vertIndex].bitangent - _vertices[md.vertIndex].bitangent;
		}
	}
	
	return true;
}


bool Converter::writeGeometry( string name )
{
	FILE *f = fopen( (name + ".geo").c_str(), "wb" );

	// Write header
	unsigned int version = 4;
	fwrite( "H3DG", 4, 1, f );
	fwrite( &version, sizeof( int ), 1, f ); 
	
	// Write joints
	unsigned int count = (unsigned int)_joints.size();
	fwrite( &count, sizeof( int ), 1, f );

	for( unsigned int i = 0; i < _joints.size(); ++i )
	{
		// Inverse bind matrix
		for( unsigned int j = 0; j < 4; ++j )
		{
			for( unsigned int k = 0; k < 4; ++k )
			{
				fwrite( &_joints[i]->invBindMat.c[k][j], sizeof( float ), 1, f );
			}
		}
	}
	
	// Write vertices
	count = (unsigned int)_vertices.size();
	fwrite( &count, sizeof( int ), 1, f );

	for( unsigned int i = 0; i < _vertices.size(); ++i )
	{
		char c;
		
		fwrite( &_vertices[i].pos.x, sizeof( float ), 1, f );
		fwrite( &_vertices[i].pos.y, sizeof( float ), 1, f );
		fwrite( &_vertices[i].pos.z, sizeof( float ), 1, f );
		c = (char)(_vertices[i].normal.x * 127); fwrite( &c, sizeof( char ), 1, f );
		c = (char)(_vertices[i].normal.y * 127); fwrite( &c, sizeof( char ), 1, f );
		c = (char)(_vertices[i].normal.z * 127); fwrite( &c, sizeof( char ), 1, f );
		c = (char)(_vertices[i].tangent.x * 127); fwrite( &c, sizeof( char ), 1, f );
		c = (char)(_vertices[i].tangent.y * 127); fwrite( &c, sizeof( char ), 1, f );
		c = (char)(_vertices[i].tangent.z * 127); fwrite( &c, sizeof( char ), 1, f );
		c = (char)(_vertices[i].bitangent.x * 127); fwrite( &c, sizeof( char ), 1, f );
		c = (char)(_vertices[i].bitangent.y * 127); fwrite( &c, sizeof( char ), 1, f );
		c = (char)(_vertices[i].bitangent.z * 127); fwrite( &c, sizeof( char ), 1, f );

		fwrite( &_vertices[i].texCoords[0].x, sizeof( float ), 1, f );
		fwrite( &_vertices[i].texCoords[0].y, sizeof( float ), 1, f );
		fwrite( &_vertices[i].texCoords[1].x, sizeof( float ), 1, f );
		fwrite( &_vertices[i].texCoords[1].y, sizeof( float ), 1, f );
		fwrite( &_vertices[i].texCoords[2].x, sizeof( float ), 1, f );
		fwrite( &_vertices[i].texCoords[2].y, sizeof( float ), 1, f );
		fwrite( &_vertices[i].texCoords[3].x, sizeof( float ), 1, f );
		fwrite( &_vertices[i].texCoords[3].y, sizeof( float ), 1, f );

		if( !_joints.empty() )
		{
			for( unsigned int j = 0; j < 4; ++j )
			{
				unsigned char uc = (unsigned char)_vertices[i].joints[j];
				fwrite( &uc, sizeof( char ), 1, f );
			}
			for( unsigned int j = 0; j < 4; ++j )
			{
				unsigned char uc = (unsigned char)(_vertices[i].weights[j] * 255);
				fwrite( &uc, sizeof( char ), 1, f );
			}
		}
	}

	// Write indices
	count = (unsigned int)_indices.size();
	fwrite( &count, sizeof( int ), 1, f );

	for( unsigned int i = 0; i < _indices.size(); ++i )
	{
		fwrite( &_indices[i], sizeof( int ), 1, f );
	}

	// Write morph targets
	count = (unsigned int)_morphTargets.size();
	fwrite( &count, sizeof( int ), 1, f );

	for( unsigned int i = 0; i < _morphTargets.size(); ++i )
	{
		fwrite( &_morphTargets[i].name, 256, 1, f );
		count = (unsigned int)_morphTargets[i].diffs.size();
		fwrite( &count, sizeof( int ), 1, f );
		
		for( unsigned int j = 0; j < _morphTargets[i].diffs.size(); ++j )
		{
			fwrite( &_morphTargets[i].diffs[j].vertIndex, sizeof( int ), 1, f );
			fwrite( &_morphTargets[i].diffs[j].posDiff.x, sizeof( float ), 1, f );
			fwrite( &_morphTargets[i].diffs[j].posDiff.y, sizeof( float ), 1, f );
			fwrite( &_morphTargets[i].diffs[j].posDiff.z, sizeof( float ), 1, f );
			fwrite( &_morphTargets[i].diffs[j].normDiff.x, sizeof( float ), 1, f );
			fwrite( &_morphTargets[i].diffs[j].normDiff.y, sizeof( float ), 1, f );
			fwrite( &_morphTargets[i].diffs[j].normDiff.z, sizeof( float ), 1, f );
			fwrite( &_morphTargets[i].diffs[j].tanDiff.x, sizeof( float ), 1, f );
			fwrite( &_morphTargets[i].diffs[j].tanDiff.y, sizeof( float ), 1, f );
			fwrite( &_morphTargets[i].diffs[j].tanDiff.z, sizeof( float ), 1, f );
			fwrite( &_morphTargets[i].diffs[j].bitanDiff.x, sizeof( float ), 1, f );
			fwrite( &_morphTargets[i].diffs[j].bitanDiff.y, sizeof( float ), 1, f );
			fwrite( &_morphTargets[i].diffs[j].bitanDiff.z, sizeof( float ), 1, f );
		}
	}
	
	fclose( f );

	return true;
}


void Converter::writeSGJoint( Joint *joint, unsigned int depth, ofstream &outf )
{
	Vec3f trans, rot, scale;
	joint->matRel.decompose( trans, rot, scale );
	rot.x = radToDeg( rot.x );
	rot.y = radToDeg( rot.y );
	rot.z = radToDeg( rot.z );
	
	for( unsigned int i = 0; i < depth + 1; ++i ) outf << "\t";
	outf << "<Joint ";
	outf << "name=\"" << joint->name << "\" ";
	if( trans != Vec3f( 0, 0, 0 ) )
		outf << "px=\"" << trans.x << "\" py=\"" << trans.y << "\" pz=\"" << trans.z << "\" ";
	if( rot != Vec3f( 0, 0, 0 ) )
		outf << "rx=\"" << rot.x << "\" ry=\"" << rot.y << "\" rz=\"" << rot.z << "\" ";
	if( scale != Vec3f( 1, 1, 1 ) )
		outf << "sx=\"" << scale.x << "\" sy=\"" << scale.y << "\" sz=\"" << scale.z << "\" ";
	outf << "jointIndex=\"" << joint->index << "\"";
	
	if( joint->children.size() == 0 )
	{
		outf << " />\n";
	}
	else
	{
		outf << ">\n";
		for( unsigned int i = 0; i < joint->children.size(); ++i )
			writeSGJoint( (Joint *)joint->children[i], depth + 1, outf );

		for( unsigned int i = 0; i < depth + 1; ++i ) outf << "\t";
		outf << "</Joint>\n";	
	}
}


void Converter::writeSGMesh( string modelName, Mesh *mesh, unsigned int depth, ofstream &outf )
{
	Vec3f trans, rot, scale;
	mesh->matRel.decompose( trans, rot, scale );
	rot.x = radToDeg( rot.x );
	rot.y = radToDeg( rot.y );
	rot.z = radToDeg( rot.z );

	for( unsigned int i = 0; i < mesh->triGroups.size(); ++i )
	{
		for( unsigned int j = 0; j < depth + 1; ++j ) outf << "\t";
		outf << "<Mesh ";
		outf << "name=\"" << mesh->name << "\" ";
		outf << "material=\"";
		outf << modelName + "/" + mesh->triGroups[i].matName + ".material.xml\" ";
		if( trans != Vec3f( 0, 0, 0 ) )
			outf << "px=\"" << trans.x << "\" py=\"" << trans.y << "\" pz=\"" << trans.z << "\" ";
		if( rot != Vec3f( 0, 0, 0 ) )
			outf << "rx=\"" << rot.x << "\" ry=\"" << rot.y << "\" rz=\"" << rot.z << "\" ";
		if( scale != Vec3f( 1, 1, 1 ) )
			outf << "sx=\"" << scale.x << "\" sy=\"" << scale.y << "\" sz=\"" << scale.z << "\" ";
		outf << "batchStart=\"";
		outf << mesh->triGroups[i].first;
		outf << "\" batchCount=\"";
		outf << mesh->triGroups[i].count;
		outf << "\" vertRStart=\"";
		outf << mesh->triGroups[i].vertRStart;
		outf << "\" vertREnd=\"";
		outf << mesh->triGroups[i].vertREnd;
		outf << "\"";

		if( mesh->children.size() == 0 )
		{
			outf << " />\n";
		}
		else
		{
			outf << ">\n";
			for( unsigned int j = 0; j < mesh->children.size(); ++j )
				writeSGMesh( modelName, (Mesh *)mesh->children[j], depth + 1, outf );

			for( unsigned int j = 0; j < depth + 1; ++j ) outf << "\t";
			outf << "</Mesh>\n";	
		}
	}
}


bool Converter::writeSceneGraph( string name )
{
	ofstream outf;
	outf.open( (name + ".scene.xml").c_str(), ios::out );
	
	outf << "<Model name=\"" << name << "\" geometry=\"" << name << ".geo\">\n";
	
	// Joints
	for( unsigned int i = 0; i < _joints.size(); ++i )
	{
		if( _joints[i]->parent == 0x0 ) writeSGJoint( _joints[i], 0, outf );
	}

	outf << "\n";
	
	// Meshes
	for( unsigned int i = 0; i < _meshes.size(); ++i )
	{
		if( _meshes[i]->parent == 0x0 ) writeSGMesh( name, _meshes[i], 0, outf );
	}

	outf << "</Model>\n";

	outf.close();
	
	return true;
}


bool Converter::saveModel( string name )
{
	writeGeometry( name );
	writeSceneGraph( name );

	return true;
}


bool Converter::writeMaterials( SimpleColladaFile &collada, string name, string defShader )
{
	_mkdir( "materials" );
	_mkdir( ("materials/" + name).c_str() );
	
	for( unsigned int i = 0; i < collada.libMaterials.materials.size(); ++i )
	{
		DaeMaterial &material = *collada.libMaterials.materials[i];
		
		if( !material.used ) continue;
		
		ofstream outf;
		string fileName = "materials/" + name + "/" + material.id + ".material.xml";
		outf.open( fileName.c_str(), ios::out );

		outf << "<Material>\n";
		
		if( material.effect != 0x0 )
		{
			outf << "\t<Shader name=\"" + defShader + "\" />\n\n";
			
			if( material.effect->diffuseMap != 0x0 )
			{
				outf << "\t<TexUnit unit=\"0\" map=\"";
				outf << material.effect->diffuseMap->fileName;
				outf << "\" />\n";
			}
		}
		
		outf << "</Material>\n";

		outf.close();
	}
	
	return true;
}


bool Converter::hasAnimation()
{
	if( _numAnimFrames == 0 )
		return false;
	else
		return true;
}


void Converter::writeAnimFrames( SceneNode &node, FILE *f )
{
	fwrite( &node.name, 256, 1, f );
		
	for( unsigned int i = 0; i < node.frames.size(); ++i )
	{
		Vec3f transVec, rotVec, scaleVec;
		node.frames[i].decompose( transVec, rotVec, scaleVec );
		Quaternion rotQuat( rotVec.x, rotVec.y, rotVec.z );
		
		fwrite( &rotQuat.v.x, sizeof( float ), 1, f );
		fwrite( &rotQuat.v.y, sizeof( float ), 1, f );
		fwrite( &rotQuat.v.z, sizeof( float ), 1, f );
		fwrite( &rotQuat.w, sizeof( float ), 1, f );
		fwrite( &transVec.x, sizeof( float ), 1, f );
		fwrite( &transVec.y, sizeof( float ), 1, f );
		fwrite( &transVec.z, sizeof( float ), 1, f );
		fwrite( &scaleVec.x, sizeof( float ), 1, f );
		fwrite( &scaleVec.y, sizeof( float ), 1, f );
		fwrite( &scaleVec.z, sizeof( float ), 1, f );
	}
}


bool Converter::writeAnimation( string name )
{
	FILE *f = fopen( (name + ".anim").c_str(), "wb" );

	// Write header
	unsigned int version = 2;
	fwrite( "H3DA", 4, 1, f );
	fwrite( &version, sizeof( int ), 1, f ); 
	
	// Write number of nodes
	unsigned int count = 0;
	for( unsigned int i = 0; i < _joints.size(); ++i )
		if( _joints[i]->frames.size() > 0 ) ++count;
	for( unsigned int i = 0; i < _meshes.size(); ++i )
		if( _meshes[i]->frames.size() > 0 ) ++count;
	fwrite( &count, sizeof( int ), 1, f );
	fwrite( &_numAnimFrames, sizeof( int ), 1, f );

	for( unsigned int i = 0; i < _joints.size(); ++i )
	{
		if( _joints[i]->frames.size() == 0 ) continue;
		
		writeAnimFrames( *_joints[i], f );
	}

	for( unsigned int i = 0; i < _meshes.size(); ++i )
	{
		if( _meshes[i]->frames.size() == 0 ) continue;
		
		writeAnimFrames( *_meshes[i], f );
	}
	
	fclose( f );

	return true;
}
