// *************************************************************************************************
//
// Horde3D
//   Next-Generation Graphics Engine
// --------------------------------------
// Copyright (C) 2006-2007 Nicolas Schulz
//
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//
// *************************************************************************************************

#pragma once

#include "tinyxml.h"
#include "daeCommon.h"
#include "utils.h"
#include <string>
#include <vector>
using namespace std;


struct DaeAnimation
{
	string						id;
	string						targetId;
	vector< DaeSource >			sources;
	DaeSource					*times;
	DaeSource					*matrices;
	vector< DaeAnimation * >	children;


	DaeAnimation()
	{
		times = 0x0;
		matrices = 0x0;
	}
	

	~DaeAnimation()
	{
		for( unsigned int i = 0; i < children.size(); ++i ) delete children[i];
	}


	DaeSource *findSource( string id )
	{
		if( id == "" ) return 0x0;

		for( unsigned int i = 0; i < sources.size(); ++i )
		{
			if( sources[i].id == id ) return &sources[i];
		}

		return 0x0;
	}

	
	DaeAnimation *findAnimationForTarget( string targetId )
	{
		if( this->targetId == targetId ) return this;
		
		// Parse children
		for( unsigned int i = 0; i < children.size(); ++i )
		{
			DaeAnimation *anim = children[i]->findAnimationForTarget( targetId );

			if( anim != 0x0 ) return anim;
		}

		return 0x0;
	}


	bool parse( TiXmlElement *animNode )
	{
		id = animNode->Attribute( "id" );
		
		// Sources
		TiXmlElement *node1 = animNode->FirstChildElement( "source" );
		while( node1 != 0x0 )
		{
			DaeSource source;
			if( source.parse( node1 ) ) sources.push_back( source );

			node1 = node1->NextSiblingElement( "source" );
		}
		
		// Sampler
		node1 = animNode->FirstChildElement( "sampler" );
		if( node1 != 0x0 )
		{
			TiXmlElement *node2 = node1->FirstChildElement( "input" );

			while( node2 != 0x0 )
			{
				if( strcmp( node2->Attribute( "semantic" ), "INPUT" ) == 0x0 )
				{
					string id = node2->Attribute( "source" );
					removeGate( id );
					times = findSource( id );
				}
				else if( strcmp( node2->Attribute( "semantic" ), "OUTPUT" ) == 0x0 )
				{
					string id = node2->Attribute( "source" );
					removeGate( id );
					matrices = findSource( id );
				}
				
				node2 = node2->NextSiblingElement( "input" );
			}
		}

		// Channel target
		node1 = animNode->FirstChildElement( "channel" );
		if( node1 != 0x0 )
		{
			// TODO: Check the channel source attribute
			string s = node1->Attribute( "target" );
			if( s.find( "/transform" ) != string::npos )
				targetId = s.substr( 0, s.find( "/transform" ) );
			else
				log( "Warning: Animation channel not supported" );
		}
		
		// Parse children
		node1 = animNode->FirstChildElement( "animation" );
		while( node1 != 0x0 )
		{
			DaeAnimation *anim = new DaeAnimation();
			if( anim->parse( node1 ) ) children.push_back( anim );
			else delete anim;

			node1 = node1->NextSiblingElement( "animation" );
		}
		
		return true;
	}
};


struct DaeLibAnimations
{
	vector< DaeAnimation * >	animations;
	

	~DaeLibAnimations()
	{
		for( unsigned int i = 0; i < animations.size(); ++i ) delete animations[i];
	}
	

	DaeAnimation *findAnimationForTarget( string targetId )
	{
		for( unsigned int i = 0; i < animations.size(); ++i )
		{
			DaeAnimation *anim = animations[i]->findAnimationForTarget( targetId );
			if( anim != 0x0 ) return anim;
		}

		return 0x0;
	}

	
	bool parse( TiXmlElement *rootNode )
	{
		TiXmlElement *node1 = rootNode->FirstChildElement( "library_animations" );
		if( node1 == 0x0 ) return true;

		TiXmlElement *node2 = node1->FirstChildElement( "animation" );
		while( node2 != 0x0 )
		{
			DaeAnimation *anim = new DaeAnimation();

			if( anim->parse( node2 ) ) animations.push_back( anim );
			else delete anim;

			node2 = node2->NextSiblingElement( "animation" );
		}
		
		return true;
	}
};
