// *************************************************************************************************
//
// Horde3D
//   Next-Generation Graphics Engine
// --------------------------------------
// Copyright (C) 2006-2007 Nicolas Schulz
//
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//
// *************************************************************************************************

#pragma once

#include "tinyxml.h"
#include <string>
#include <vector>
using namespace std;


struct DaeWeight
{
	unsigned int			joint;
	float					weight;
};


typedef vector< DaeWeight > DaeVertWeight;


struct DaeSkin
{
	string					id;
	string					ownerId;
	float					bindShapeMat[16];
	vector< DaeSource >		sources;
	DaeSource				*jointArray;
	DaeSource				*weightArray;
	DaeSource				*bindMatArray;
	vector< DaeVertWeight >	vertWeights;

	
	DaeSkin()
	{
		jointArray = 0x0;
		weightArray = 0x0;
		bindMatArray = 0x0;
	}
	

	DaeSource *findSource( string id )
	{
		if( id == "" ) return 0x0;
		
		for( unsigned int i = 0; i < sources.size(); ++i )
		{
			if( sources[i].id == id ) return &sources[i];
		}

		return 0x0;
	}

	
	bool parse( TiXmlElement *ctrlNode )
	{
		TiXmlElement *node1 = ctrlNode->FirstChildElement( "skin" );
		if( node1 == 0x0 ) return false;
		
		id = ctrlNode->Attribute( "id" );
		ownerId = node1->Attribute( "source" );
		removeGate( ownerId );

		// Bind shape matrix
		TiXmlElement *node2 = node1->FirstChildElement( "bind_shape_matrix" );
		if( node2 == 0x0 ) return false;

		unsigned int pos = 0;
		char *s = (char *)node2->GetText();	
		for( int i = 0; i < 16; ++i )
		{
			float f;
			parseFloat( s, pos, f );
			bindShapeMat[i] = f;
		}
		
		// Sources
		node2 = node1->FirstChildElement( "source" );
		while( node2 != 0x0 )
		{
			DaeSource source;
			if( source.parse( node2 ) ) sources.push_back( source );
			
			node2 = node2->NextSiblingElement( "source" );
		}

		// Joints
		node2 = node1->FirstChildElement( "joints" );
		if( node2 != 0x0 )
		{
			TiXmlElement *node3 = node2->FirstChildElement( "input" );
			while( node3 != 0x0 )
			{
				if( strcmp( node3->Attribute( "semantic" ), "JOINT" ) == 0 )
				{
					// TODO: Does the joint have to be saved here??
				}
				else if( strcmp( node3->Attribute( "semantic" ), "INV_BIND_MATRIX" ) == 0 )
				{
					string sourceId = node3->Attribute( "source" );
					removeGate( sourceId );
					bindMatArray = findSource( sourceId );
				}

				node3 = node3->NextSiblingElement( "input" );
			}
		}
		
		// Vertex weights
		int jointOffset = 0, weightOffset = 0;
		unsigned int numInputs = 0;
		
		node2 = node1->FirstChildElement( "vertex_weights" );
		int count = atoi( node2->Attribute( "count" ) );
		TiXmlElement *node3 = node2->FirstChildElement( "input" );
		while( node3 != 0x0 )
		{
			++numInputs;

			if( strcmp( node3->Attribute( "semantic" ), "JOINT" ) == 0 )
			{
				if( node3->Attribute( "offset" ) != 0x0 )
					jointOffset = atoi( node3->Attribute( "offset" ) );
				string id = node3->Attribute( "source" );
				removeGate( id );
				jointArray = findSource( id );
			}
			else if( strcmp( node3->Attribute( "semantic" ), "WEIGHT" ) == 0 )
			{
				if( node3->Attribute( "offset" ) != 0x0 )
					weightOffset = atoi( node3->Attribute( "offset" ) );
				string id = node3->Attribute( "source" );
				removeGate( id );
				weightArray = findSource( id );
			}
			
			node3 = node3->NextSiblingElement( "input" );
		}

		node3 = node2->FirstChildElement( "vcount" );
		pos = 0;
		s = (char *)node3->GetText();
		for( int i = 0; i < count; ++i )
		{
			unsigned int ui;
			parseUInt( s, pos, ui );

			DaeVertWeight vertWeight;
			for( unsigned int j = 0; j < ui; ++j )
			{
				DaeWeight weight;
				vertWeight.push_back( weight );
			}

			vertWeights.push_back( vertWeight );
		}

		node3 = node2->FirstChildElement( "v" );
		pos = 0;
		s = (char *)node3->GetText();
		for( int i = 0; i < count; ++i )
		{
			for( unsigned int j = 0; j < vertWeights[i].size(); ++j )
			{
				for( unsigned int k = 0; k < numInputs; ++k )
				{
					float f;
					parseFloat( s, pos, f );

					if( k == jointOffset ) vertWeights[i][j].joint = (unsigned int)f;
					if( k == weightOffset ) vertWeights[i][j].weight = f;
				}
			}
		}
		
		return true;
	}
};


struct DaeMorph
{
	string					id;
	string					ownerId;
	vector< DaeSource >		sources;
	DaeSource				*targetArray;
	DaeSource				*weightArray;


	DaeMorph()
	{
		targetArray = 0x0;
		weightArray = 0x0;
	}
		

	DaeSource *findSource( string id )
	{
		if( id == "" ) return 0x0;
		
		for( unsigned int i = 0; i < sources.size(); ++i )
		{
			if( sources[i].id == id ) return &sources[i];
		}

		return 0x0;
	}

	
	bool parse( TiXmlElement *ctrlNode )
	{
		TiXmlElement *node1 = ctrlNode->FirstChildElement( "morph" );
		if( node1 == 0x0 ) return false;

		id = ctrlNode->Attribute( "id" );
		ownerId = node1->Attribute( "source" );
		removeGate( ownerId );
		
		// Sources
		TiXmlElement *node2 = node1->FirstChildElement( "source" );
		while( node2 != 0x0 )
		{
			DaeSource source;
			if( source.parse( node2 ) ) sources.push_back( source );
			
			node2 = node2->NextSiblingElement( "source" );
		}

		// Targets
		node2 = node1->FirstChildElement( "targets" );
		if( node2 != 0x0 )
		{
			TiXmlElement *node3 = node2->FirstChildElement( "input" );
			while( node3 != 0x0 )
			{
				if( strcmp( node3->Attribute( "semantic" ), "MORPH_TARGET" ) == 0 )
				{
					string sourceId = node3->Attribute( "source" );
					removeGate( sourceId );
					targetArray = findSource( sourceId );
				}
				else if( strcmp( node3->Attribute( "semantic" ), "MORPH_WEIGHT" ) == 0 )
				{
					string sourceId = node3->Attribute( "source" );
					removeGate( sourceId );
					weightArray = findSource( sourceId );
				}

				node3 = node3->NextSiblingElement( "input" );
			}
		}

		return true;
	}
};


struct DaeLibControllers
{
	vector< DaeSkin * >		skinControllers;
	vector< DaeMorph * >	morphControllers;


	~DaeLibControllers()
	{
		for( unsigned int i = 0; i < skinControllers.size(); ++i ) delete skinControllers[i];
		for( unsigned int i = 0; i < morphControllers.size(); ++i ) delete morphControllers[i];
	}

	
	DaeSkin *findSkin( string id )
	{
		if( id == "" ) return 0x0;
		
		for( unsigned int i = 0; i < skinControllers.size(); ++i )
		{
			if( skinControllers[i]->id == id ) return skinControllers[i];
		}
		
		return 0x0;
	}


	DaeMorph *findMorph( string id )
	{
		if( id == "" ) return 0x0;
		
		for( unsigned int i = 0; i < morphControllers.size(); ++i )
		{
			if( morphControllers[i]->id == id ) return morphControllers[i];
		}
		
		return 0x0;
	}
	
	
	bool parse( TiXmlElement *rootNode )
	{
		TiXmlElement *node1 = rootNode->FirstChildElement( "library_controllers" );
		if( node1 == 0x0 ) return true;

		TiXmlElement *node2 = node1->FirstChildElement( "controller" );
		while( node2 != 0x0 )
		{
			// Skin
			TiXmlElement *node3 = node2->FirstChildElement( "skin" );
			if( node3 != 0x0 )
			{
				DaeSkin *skin = new DaeSkin();
				if( skin->parse( node2 ) ) skinControllers.push_back( skin );
				else delete skin;
			}

			// Morph
			node3 = node2->FirstChildElement( "morph" );
			if( node3 != 0x0 )
			{
				DaeMorph *morph = new DaeMorph();
				if( morph->parse( node2 ) ) morphControllers.push_back( morph );
				else delete morph;
			}

			node2 = node2->NextSiblingElement( "controller" );
		}
		
		return true;
	}
};
