// *************************************************************************************************
//
// Horde3D
//   Next-Generation Graphics Engine
// --------------------------------------
// Copyright (C) 2006-2007 Nicolas Schulz
//
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//
// *************************************************************************************************

#pragma once

#include "tinyxml.h"
#include "daeCommon.h"
#include "utils.h"
#include "utMath.h"
#include <string>
#include <vector>
using namespace std;


struct DaeVSource
{
	string		id;
	string		posSourceId;
	DaeSource	*posSource;


	bool parse( TiXmlElement *verticesNode )
	{
		id = verticesNode->Attribute( "id" );
		
		TiXmlElement *node1 = verticesNode->FirstChildElement( "input" );
		while( node1 != 0x0 )
		{
			if( strcmp( node1->Attribute( "semantic" ), "POSITION" ) == 0 )
			{
				posSourceId = node1->Attribute( "source" );
				removeGate( posSourceId );
				return true;
			}
			
			node1 = node1->NextSiblingElement( "input" );
		}
		
		return false;
	}
};


struct IndexEntry
{
	unsigned int				posIndex;		// Required
	int							normIndex;		// Optional, can be -1
	int							texIndex[4];	// Optional, can be -1


	IndexEntry()
	{
		normIndex = -1;
		texIndex[0] = -1; texIndex[1] = -1; texIndex[2] = -1; texIndex[3] = -1;
	}
};


struct DaeTriGroup
{
	string						vSourceId;
	DaeVSource					*vSource;
	string						normSourceId;
	DaeSource					*normSource;
	string						texSourceId[4];
	DaeSource					*texSource[4];
	string						matId;
	vector< IndexEntry >		indices;


	DaeTriGroup()
	{
		vSource = 0x0;
		normSource = 0x0;
		texSource[0] = 0x0; texSource[1] = 0x0; texSource[2] = 0x0; texSource[3] = 0x0;
	}


	bool parse( TiXmlElement *trianglesNode )
	{
		int vertexOffset = 0, normOffset = -1, texCoordOffset[4] = { -1, -1, -1, -1 };
		unsigned int numInputs = 0;
		
		// Find the base mapping channel with the lowest set-id
		int baseChannel = 999999;
		TiXmlElement *node1 = trianglesNode->FirstChildElement( "input" );
		while( node1 != 0x0 )
		{
			if( strcmp( node1->Attribute( "semantic" ), "TEXCOORD" ) == 0 )
			{
				if( node1->Attribute( "set" ) != 0x0 )
				{
					if( atoi( node1->Attribute( "set" ) ) < baseChannel )
						baseChannel = atoi( node1->Attribute( "set" ) );
				}
				else
				{
					baseChannel = 0;
					break;
				}
			}
			node1 = node1->NextSiblingElement( "input" );
		}
		
		// Parse data
		node1 = trianglesNode->FirstChildElement( "input" );
		while( node1 != 0x0 )
		{
			++numInputs;
			
			if( strcmp( node1->Attribute( "semantic" ), "VERTEX" ) == 0 )
			{
				vertexOffset = atoi( node1->Attribute( "offset" ) );
				vSourceId = node1->Attribute( "source" );
				removeGate( vSourceId );
			}
			if( strcmp( node1->Attribute( "semantic" ), "TEXCOORD" ) == 0 )
			{
				int set = -1;
				if( node1->Attribute( "set" ) == 0x0 )
					set = 0;
				else
					set = atoi( node1->Attribute( "set" ) );
				set -= baseChannel;
				
				if( set >= 0 && set < 4 )
				{
					texCoordOffset[set] = atoi( node1->Attribute( "offset" ) );
					texSourceId[set] = node1->Attribute( "source" );
					removeGate( texSourceId[set] );
				}
			}
			if( strcmp( node1->Attribute( "semantic" ), "NORMAL" ) == 0 )
			{
				normOffset = atoi( node1->Attribute( "offset" ) );
				normSourceId = node1->Attribute( "source" );
				removeGate( normSourceId );
			}
			
			node1 = node1->NextSiblingElement( "input" );
		}
		
		if( vSourceId == "" ) return false;

		if( trianglesNode->Attribute( "material" ) != 0x0 )
			matId = trianglesNode->Attribute( "material" );
		
		int count = atoi( trianglesNode->Attribute( "count" ) ) * 3;
		node1 = trianglesNode->FirstChildElement( "p" );
		
		unsigned int pos = 0;
		char *s = (char *)node1->GetText();	
		for( int i = 0; i < count; ++i )
		{
			IndexEntry indexEntry;
			
			for( int j = 0; j < (int)numInputs; ++j )
			{
				unsigned int ui;
				parseUInt( s, pos, ui );
				
				// No else-if since offset sharing is possible
				if( j == vertexOffset )
					indexEntry.posIndex = ui;
				if( j == normOffset )
					indexEntry.normIndex = (int)ui;
				if( j == texCoordOffset[0] )	
					indexEntry.texIndex[0] = (int)ui;
				if( j == texCoordOffset[1] )	
					indexEntry.texIndex[1] = (int)ui;
				if( j == texCoordOffset[2] )	
					indexEntry.texIndex[2] = (int)ui;
				if( j == texCoordOffset[3] )	
					indexEntry.texIndex[3] = (int)ui;
			}

			indices.push_back( indexEntry );
		}

		return true;
	}


	Vec3f getPos( int posIndex )
	{
		Vec3f v;
		
		// Assume the float buffer has at least 3 values per element
		DaeSource *source = vSource->posSource;
		v.x = source->floatArray[posIndex * source->elemsPerEntry + 0];
		v.y = source->floatArray[posIndex * source->elemsPerEntry + 1];
		v.z = source->floatArray[posIndex * source->elemsPerEntry + 2];
		
		return v;
	}

	Vec3f getNormal( int normIndex )
	{
		Vec3f v;
		
		DaeSource *source = normSource;
		if( source != 0x0 && normIndex >= 0 )
		{
			// Assume the float buffer has at least 3 values per element
			v.x = source->floatArray[normIndex * source->elemsPerEntry + 0];
			v.y = source->floatArray[normIndex * source->elemsPerEntry + 1];
			v.z = source->floatArray[normIndex * source->elemsPerEntry + 2];
		}
		
		return v;
	}

	Vec3f getTexCoords( int texIndex, unsigned int set )
	{
		Vec3f v;
		
		if( set < 4 )
		{
			DaeSource *source = texSource[set];
			if( source != 0x0 && texIndex >= 0 )
			{
				// Assume the float buffer has at least 2 values per element
				v.x = source->floatArray[texIndex * source->elemsPerEntry + 0];
				v.y = source->floatArray[texIndex * source->elemsPerEntry + 1];
				
				if( source->elemsPerEntry >= 3 )
					v.z = source->floatArray[texIndex * source->elemsPerEntry + 2];
			}
		}
		
		return v;
	}
};


struct DaeGeometry
{
	string					id;
	string					name;
	vector< DaeSource >		sources;
	vector< DaeVSource >	vsources;
	vector< DaeTriGroup >	triGroups;


	DaeSource *findSource( string id )
	{
		if( id == "" ) return 0x0;
		
		for( unsigned int i = 0; i < sources.size(); ++i )
		{
			if( sources[i].id == id ) return &sources[i];
		}

		return 0x0;
	}


	DaeVSource *findVSource( string id )
	{
		if( id == "" ) return 0x0;
		
		for( unsigned int i = 0; i < vsources.size(); ++i )
		{
			if( vsources[i].id == id ) return &vsources[i];
		}

		return 0x0;
	}


	bool parse( TiXmlElement *geometryNode )
	{
		TiXmlElement *node1 = geometryNode->FirstChildElement( "mesh" );
		if( node1 == 0x0 ) return false;

		id = geometryNode->Attribute( "id" );
		
		if( geometryNode->Attribute( "name" ) != 0x0 )
			name = geometryNode->Attribute( "name" );

		TiXmlElement *node2 = node1->FirstChildElement( "source" );
		while( node2 != 0x0 )
		{
			DaeSource source;
			if( source.parse( node2 ) != 0x0 ) sources.push_back( source );
			
			node2 = node2->NextSiblingElement( "source" );
		}

		node2 = node1->FirstChildElement( "vertices" );
		while( node2 != 0x0 )
		{
			DaeVSource vsource;
			if( vsource.parse( node2 ) != 0x0 )
			{	
				vsource.posSource = findSource( vsource.posSourceId );
				if( vsource.posSource != 0x0 ) vsources.push_back( vsource );
			}
			
			node2 = node2->NextSiblingElement( "vertices" );
		}

		node2 = node1->FirstChildElement( "triangles" );
		while( node2 != 0x0 )
		{
			DaeTriGroup triGroup;
			if( triGroup.parse( node2 ) != 0x0 )
			{	
				triGroup.vSource = findVSource( triGroup.vSourceId );
				triGroup.normSource = findSource( triGroup.normSourceId );
				triGroup.texSource[0] = findSource( triGroup.texSourceId[0] );
				triGroup.texSource[1] = findSource( triGroup.texSourceId[1] );
				triGroup.texSource[2] = findSource( triGroup.texSourceId[2] );
				triGroup.texSource[3] = findSource( triGroup.texSourceId[3] );
				
				if( triGroup.vSource != 0x0 )
					triGroups.push_back( triGroup );

				if( triGroup.vSource == 0x0 )
					log( "Warning: Mesh '" + id + "' has no vertex coordinates and is ignored" );
			}
			
			node2 = node2->NextSiblingElement( "triangles" );
		}

		if( node1->FirstChildElement( "polygons" ) != 0x0 )
			log( "Warning: Ingnoring non-triangle data in mesh '" + id + "'" );
		if( node1->FirstChildElement( "polylist" ) != 0x0 )
			log( "Warning: Ingnoring non-triangle data in mesh '" + id + "'" );
		if( node1->FirstChildElement( "trifans" ) != 0x0 )
			log( "Warning: Ingnoring non-triangle data in mesh '" + id + "'" );
		if( node1->FirstChildElement( "tristrips" ) != 0x0 )
			log( "Warning: Ingnoring non-triangle data in mesh '" + id + "'" );

		return true;
	}


	Vec3f getPos( int posIndex )
	{
		Vec3f v;
		
		// Assume the float buffer has at least 3 values per element
		DaeSource *source = vsources[0].posSource;
		v.x = source->floatArray[posIndex * source->elemsPerEntry + 0];
		v.y = source->floatArray[posIndex * source->elemsPerEntry + 1];
		v.z = source->floatArray[posIndex * source->elemsPerEntry + 2];
		
		return v;
	}
};


struct DaeLibGeometries
{
public:

	vector< DaeGeometry* >	geometries;

	
	~DaeLibGeometries()
	{
		for( unsigned int i = 0; i < geometries.size(); ++i ) delete geometries[i];
	}
	

	DaeGeometry *findGeometry( string id )
	{
		if( id == "" ) return 0x0;
		
		for( unsigned int i = 0; i < geometries.size(); ++i )
		{
			if( geometries[i]->id == id ) return geometries[i];
		}

		return 0x0;
	}
	
	
	bool parse( TiXmlElement *rootNode )
	{
		TiXmlElement *node1 = rootNode->FirstChildElement( "library_geometries" );
		if( node1 == 0x0 ) return true;

		TiXmlElement *node2 = node1->FirstChildElement( "geometry" );
		while( node2 != 0x0 )
		{
			DaeGeometry *geometry = new DaeGeometry();
			if( geometry->parse( node2 ) ) geometries.push_back( geometry );
			else delete geometry;

			node2 = node2->NextSiblingElement( "geometry" );
		}
		
		return true;
	}
};
