// *************************************************************************************************
//
// Horde3D
//   Next-Generation Graphics Engine
// --------------------------------------
// Copyright (C) 2006-2007 Nicolas Schulz
//
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//
// *************************************************************************************************

#pragma once

#include "tinyxml.h"
#include "daeLibAnimations.h"
#include <string>
#include <vector>
#include <map>
using namespace std;


struct DaeNode
{
	string					id;
	string					name;
	string					instanceId;
	map< string, string >	materialBindings;
	bool					joint;
	float					transMat[16];		// Relative to parent
	DaeNode					*parent;
	DaeAnimation			*animation;

	
	DaeNode()
	{
		parent = 0x0;
		animation = 0x0;
		// Identity matrix
		for( unsigned int i = 0; i < 16; ++i ) transMat[i] = 0;
		for( unsigned int i = 0; i < 4; ++i ) transMat[i * 4 + i] = 1;
	}
	
	bool parse( TiXmlElement *nodeNode )
	{
		parent = 0x0;
		id = nodeNode->Attribute( "id" );
		if( nodeNode->Attribute( "name" ) != 0x0 )
			name = nodeNode->Attribute( "name" );
		
		if( nodeNode->Attribute( "type" ) != 0x0 &&
			strcmp( nodeNode->Attribute( "type" ), "JOINT" ) == 0 ) joint  = true;
		else joint = false;

		// Instance
		TiXmlElement *node1 = nodeNode->FirstChildElement( "instance_geometry" );
		if( node1 == 0x0 ) node1 = nodeNode->FirstChildElement( "instance_controller" );
		if( node1 != 0x0 )
		{
			instanceId = node1->Attribute( "url" );
			removeGate( instanceId );

			// Find material bindings
			TiXmlElement *node2 = node1->FirstChildElement( "bind_material" );
			if( node2 != 0x0 )
			{
				node2 = node2->FirstChildElement( "technique_common" );
				if( node2 != 0x0 )
				{
					node2 = node2->FirstChildElement( "instance_material" );
					while( node2 != 0x0 )
					{
						string s = node2->Attribute( "target" );
						removeGate( s );
						materialBindings[node2->Attribute( "symbol" )] = s;

						node2 = node2->NextSiblingElement( "instance_material" );
					}
				}
			}
		}
		
		// Matrix
		node1 = nodeNode->FirstChildElement( "matrix" );
		if( node1 != 0x0 )
		{
			unsigned int pos = 0;
			char *s = (char *)node1->GetText();	
			for( int i = 0; i < 16; ++i )
			{
				float f;
				parseFloat( s, pos, f );
				transMat[i] = f;
			}
		}
		
		return true;
	}
};


struct DaeVisualScene
{
	string					id;
	vector< DaeNode * >		nodes;

	
	~DaeVisualScene()
	{
		for( unsigned int i = 0; i < nodes.size(); ++i ) delete nodes[i];
	}
	
	DaeNode *parseNodes( TiXmlElement *nodeNode, DaeNode *parent )
	{
		DaeNode *node = new DaeNode();
		if( node->parse( nodeNode ) )
		{	
			node->parent = parent;
			nodes.push_back( node );

			// Parse children
			TiXmlElement *node1 = nodeNode->FirstChildElement( "node" );
			while( node1 != 0x0 )
			{
				DaeNode *childNode = parseNodes( node1, node );

				node1 = node1->NextSiblingElement( "node" );
			}
		}
		else
		{
			delete node; node = 0x0;
		}
		
		return node;
	}


	bool parse( TiXmlElement *visSceneNode )
	{
		id = visSceneNode->Attribute( "id" );
		
		TiXmlElement *node1 = visSceneNode->FirstChildElement( "node" );
		while( node1 != 0x0 )
		{
			parseNodes( node1, 0x0 );

			node1 = node1->NextSiblingElement( "node" );
		}
		
		return true;
	}
};


struct DaeLibVisScenes
{
	vector< DaeVisualScene * >	visScenes;


	~DaeLibVisScenes()
	{
		for( unsigned int i = 0; i < visScenes.size(); ++i ) delete visScenes[i];
	}
	

	DaeVisualScene *findVisualScene( string id )
	{
		if( id == "" ) return 0x0;
		
		for( unsigned int i = 0; i < visScenes.size(); ++i )
		{
			if( visScenes[i]->id == id ) return visScenes[i];
		}

		return 0x0;
	}
	
	
	bool parse( TiXmlElement *rootNode )
	{
		TiXmlElement *node1 = rootNode->FirstChildElement( "library_visual_scenes" );
		if( node1 == 0x0 ) return true;

		TiXmlElement *node2 = node1->FirstChildElement( "visual_scene" );
		while( node2 != 0x0 )
		{
			DaeVisualScene *visScene = new DaeVisualScene();

			if( visScene->parse( node2 ) ) visScenes.push_back( visScene );
			else delete visScene;

			node2 = node2->NextSiblingElement( "visual_scene" );
		}
		
		return true;
	}
};
