// *************************************************************************************************
//
// Horde3D
//   Next-Generation Graphics Engine
// --------------------------------------
// Copyright (C) 2006-2007 Nicolas Schulz
//
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//
// *************************************************************************************************

#include "daeMain.h"
#include "utils.h"


bool SimpleColladaFile::parseFile( string fileName )
{
	// Parse Collada file
	TiXmlDocument *doc = new TiXmlDocument( fileName.c_str() );
	doc->SetCondenseWhiteSpace( false );
	
	if( !doc->LoadFile() || doc->Error() )
	{
		log( "Error: Input file not found or invalid XML syntax" );
		return false;
	}
	
	TiXmlElement *rootNode = doc->RootElement();
	if( rootNode == 0x0 || strcmp( rootNode->Value(), "COLLADA" ) != 0 )
	{	
		log( "Error: Invalid Collada document" );
		return false;
	}

	if( strcmp( rootNode->Attribute( "version" ), "1.4.0") != 0 &&
		strcmp( rootNode->Attribute( "version" ), "1.4.1") != 0 )
	{
		log( "Error: Unsupported Collada version" );
		return false;
	}

	// Parse asset
	TiXmlElement *node1 = rootNode->FirstChildElement( "asset" );
	if( node1 != 0x0 )
	{
		node1 = node1->FirstChildElement( "up_axis" );
		if( node1 != 0x0 )
		{
			if( strcmp( node1->GetText(), "Y_UP" ) == 0 )
			{
				y_up = true;
			}
			else if( strcmp( node1->GetText(), "Z_UP" ) == 0 )
			{
				y_up = false;
			}
			else
			{
				log( "Warning: Unsupported up-axis" );
			}
		}
	}
	
	// Parse libraries
	libImages.parse( rootNode );

	libEffects.parse( rootNode );
	// Assign images
	for( unsigned int i = 0; i < libEffects.effects.size(); ++i )
	{
		DaeEffect &effect = *libEffects.effects[i];

		if( effect.diffuseMapId != "" )
		{
			effect.diffuseMap = libImages.findImage( effect.diffuseMapId );

			if( effect.diffuseMap == 0x0 )
				log(  "Warning: Image '" + effect.diffuseMapId + "' not found" );
		}
	}
	
	libMaterials.parse( rootNode );
	// Assign effects
	for( unsigned int i = 0; i < libMaterials.materials.size(); ++i )
	{
		DaeMaterial &material = *libMaterials.materials[i];

		material.effect = libEffects.findEffect( material.effectId );

		if( material.effect == 0x0 )
			log(  "Warning: Effect '" + material.effectId + "' not found" );
	}
	
	libGeometries.parse( rootNode );

	libControllers.parse( rootNode );

	libAnimations.parse( rootNode );
	
	libVisScenes.parse( rootNode );
	// Assign animations
	for( unsigned int i = 0; i < libVisScenes.visScenes.size(); ++i )
	{
		for( unsigned int j = 0; j < libVisScenes.visScenes[i]->nodes.size(); ++j )
		{
			libVisScenes.visScenes[i]->nodes[j]->animation =
				libAnimations.findAnimationForTarget( libVisScenes.visScenes[i]->nodes[j]->id );
		}
	}

	// Visual scene instance
	bool foundScene = false;
	node1 = rootNode->FirstChildElement( "scene" );
	if( node1 != 0x0 )
	{
		TiXmlElement *node2 = node1->FirstChildElement( "instance_visual_scene" );
		if( node2 != 0x0 )
		{
			string sceneId = node2->Attribute( "url" );
			removeGate( sceneId );
			scene = libVisScenes.findVisualScene( sceneId );
			if( scene != 0x0 ) foundScene = true;

			if( node2->NextSiblingElement( "instance_visual_scene" ) != 0x0 )
				log( "Warning: Found more than one scene instance; ignoring following instances" );
		}
	}
	
	if( !foundScene ) log( "Warning: No scene instance found" );
	
	return true;
}
