// *************************************************************************************************
//
// Horde3D
//   Next-Generation Graphics Engine
// --------------------------------------
// Copyright (C) 2006-2007 Nicolas Schulz
//
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//
// *************************************************************************************************

#include "egAnimation.h"
#include "egSingleton.h"

#include "utDebug.h"


AnimationResource::AnimationResource( const string &name, unsigned int flags ) :
	Resource( ResourceTypes::Animation, name, flags )
{
	initDefault();	
}


AnimationResource::~AnimationResource()
{
	release();
}


void AnimationResource::initDefault()
{
	_numFrames = 0;
}


void AnimationResource::release()
{
	_nodes.clear();
}


bool AnimationResource::load( const char *data, unsigned int size )
{
	if( !Resource::load( data, size ) ) return false;

	// Make sure header is available
	if( size < 8 )
	{
		TheEngineLog::instance().writeWarning( "Invalid animation file" );
		return false;
	}
	
	char *myData = (char *)data;
	
	// Check header and version
	char id[4];
	memcpy( &id, myData, 4 ); myData += 4;
	if( id[0] != 'H' || id[1] != '3' || id[2] != 'D' || id[3] != 'A' )
	{
		TheEngineLog::instance().writeWarning( "Invalid animation file" );
		return false;
	}
	unsigned int version;
	memcpy( &version, myData, sizeof( int ) ); myData += sizeof( int );
	if( version != 2 )
	{
		TheEngineLog::instance().writeWarning( "Unsupported version of animation file" );
		return false;
	}
	
	// Load animation data
	unsigned int numNodes;
	memcpy( &numNodes, myData, sizeof( int ) ); myData += sizeof( int );
	memcpy( &_numFrames, myData, sizeof( int ) ); myData += sizeof( int );

	_nodes.resize( numNodes );

	for( unsigned int i = 0; i < numNodes; ++i )
	{
		char name[256];
		AnimNode &node = _nodes[i];
		
		memcpy( name, myData, 256 ); myData += 256;
		node.name = name;

		node.frames.resize( _numFrames );
		for( unsigned int j = 0; j < _numFrames; ++j )
		{
			Frame &frame = node.frames[j];

			memcpy( &frame.rotQuat.v.x, myData, sizeof( float ) ); myData += sizeof( float );
			memcpy( &frame.rotQuat.v.y, myData, sizeof( float ) ); myData += sizeof( float );
			memcpy( &frame.rotQuat.v.z, myData, sizeof( float ) ); myData += sizeof( float );
			memcpy( &frame.rotQuat.w, myData, sizeof( float ) ); myData += sizeof( float );

			memcpy( &frame.transVec.x, myData, sizeof( float ) ); myData += sizeof( float );
			memcpy( &frame.transVec.y, myData, sizeof( float ) ); myData += sizeof( float );
			memcpy( &frame.transVec.z, myData, sizeof( float ) ); myData += sizeof( float );

			memcpy( &frame.scaleVec.x, myData, sizeof( float ) ); myData += sizeof( float );
			memcpy( &frame.scaleVec.y, myData, sizeof( float ) ); myData += sizeof( float );
			memcpy( &frame.scaleVec.z, myData, sizeof( float ) ); myData += sizeof( float );

			// Prebake transformation matrix for fast animation path
			frame.bakedTransMat.scale( frame.scaleVec.x, frame.scaleVec.y, frame.scaleVec.z );
			frame.bakedTransMat = Matrix4f( frame.rotQuat ) * frame.bakedTransMat;
			frame.bakedTransMat.translate( frame.transVec.x, frame.transVec.y, frame.transVec.z );
		}

		if( !node.frames.empty() )
			node.firstFrameInvTrans = node.frames[0].bakedTransMat.inverted();
	}
	
	return true;
}


const void *AnimationResource::getData( ResourceData::List param )
{
	const void *data = Resource::getData( param );
	if( data != 0x0 ) return data;

	if( param == ResourceData::AnimFrameCount )
	{
		return &_numFrames;
	}

	return 0x0;
}


AnimNode *AnimationResource::findNode( const string &name )
{
	for( unsigned int i = 0; i < _nodes.size(); ++i )
	{
		if( _nodes[i].name == name ) return &_nodes[i];
	}

	return 0x0;
}
