// *************************************************************************************************
//
// Horde3D
//   Next-Generation Graphics Engine
// --------------------------------------
// Copyright (C) 2006-2007 Nicolas Schulz
//
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//
// *************************************************************************************************

#include "egCom.h"
#include "egSingleton.h"

#include "utDebug.h"


// *************************************************************************************************
// Class EngineConfig
// *************************************************************************************************

EngineConfig::EngineConfig()
{
	trilinearFiltering = true;
	anisotropyFactor = 1;
	texCompression = false;
	loadTextures = true;
	fastAnimation = true;
	occlusionCulling = true;
	shadowMapSize = 512;
	debugViewMode = false;

	logging = true;
	worldSize = 10000;
}


float EngineConfig::getOption( EngineOptions::List param )
{
	switch( param )
	{
	case EngineOptions::TrilinearFiltering:
		return trilinearFiltering ? 1.0f : 0.0f;
	case EngineOptions::AnisotropyFactor:
		return (float)anisotropyFactor;
	case EngineOptions::TexCompression:
		return texCompression ? 1.0f : 0.0f;
	case EngineOptions::LoadTextures:
		return loadTextures ? 1.0f : 0.0f;
	case EngineOptions::FastAnimation:
		return fastAnimation ? 1.0f : 0.0f;
	case EngineOptions::OcclusionCulling:
		return occlusionCulling ? 1.0f : 0.0f;
	case EngineOptions::ShadowMapSize:
		return (float)shadowMapSize;
	case EngineOptions::DebugViewMode:
		return debugViewMode ? 1.0f : 0.0f;
	default:
		return Math::NaN;
	}
}


bool EngineConfig::setOption( EngineOptions::List param, float value )
{
	int size;
	
	switch( param )
	{
	case EngineOptions::TrilinearFiltering:
		trilinearFiltering = (value != 0);
		return true;
	case EngineOptions::AnisotropyFactor:
		anisotropyFactor = (int)value;
		return true;
	case EngineOptions::TexCompression:
		texCompression = (value != 0);
		return true;
	case EngineOptions::LoadTextures:
		loadTextures = (value != 0);
		return true;
	case EngineOptions::FastAnimation:
		fastAnimation = (value != 0);
		return true;
	case EngineOptions::OcclusionCulling:
		occlusionCulling = (value != 0);
		return true;
	case EngineOptions::ShadowMapSize:
		size = (int)value;

		if( size == shadowMapSize ) return true;
		if( size != 128 && size != 256 && size != 512 && size != 1024 && size != 2048 ) return false;

		// Update shadow map
		TheRenderer::instance().destroyShadowBuffer();
		
		if( !TheRenderer::instance().createShadowBuffer( size, size ) )
		{
			TheEngineLog::instance().writeError( "Failed to create shadow map" );
			// Restore old buffer
			TheRenderer::instance().createShadowBuffer( shadowMapSize, shadowMapSize );
			return false;
		}
		else
		{
			shadowMapSize = size;
			return true;
		}
	case EngineOptions::DebugViewMode:
		debugViewMode = (value != 0);
		return true;
	default:
		return false;
	}
}


// *************************************************************************************************
// Class EngineLog
// *************************************************************************************************

EngineLog::EngineLog()
{
	_firstTick = clock();
}


void EngineLog::pushMessage( const string &text, unsigned int level )
{
	if( TheEngineConfig::instance().logging && _messages.size() < MaxNumMessages )
	{
		_messages.push( LogMessage( text, level, (clock() - _firstTick) / (float)CLOCKS_PER_SEC ) );
	}
}


void EngineLog::writeInfo( const string &text )
{
	pushMessage( text, 0 );	
}


void EngineLog::writeWarning( const string &text )
{
	pushMessage( text, 1 );	
}


void EngineLog::writeError( const string &text )
{
	pushMessage( text, 2 );	
}


bool EngineLog::getMessage( LogMessage &msg )
{
	if( !_messages.empty() )
	{
		msg = _messages.front();
		_messages.pop();
		return true;
	}
	else
		return false;
}
