// *************************************************************************************************
//
// Horde3D
//   Next-Generation Graphics Engine
// --------------------------------------
// Copyright (C) 2006-2007 Nicolas Schulz
//
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//
// *************************************************************************************************

#include "egGeometry.h"
#include "egAnimation.h"
#include "egResource.h"
#include "egSingleton.h"

#include "utDebug.h"


unsigned int GeometryResource::defVertBuffer = 0;
unsigned int GeometryResource::defIndexBuffer = 0;


void GeometryResource::initResType()
{
	defVertBuffer = TheRenderer::instance().uploadVertices( 0x0, 0 );
	defIndexBuffer = TheRenderer::instance().uploadIndices( 0x0, 0 );
}


void GeometryResource::releaseResType()
{
	TheRenderer::instance().unloadBuffers( defVertBuffer, defIndexBuffer );
}


GeometryResource::GeometryResource( const string &name, unsigned int flags ) :
	Resource( ResourceTypes::Geometry, name, flags )
{
	initDefault();
}


GeometryResource::~GeometryResource()
{
	release();
}


void GeometryResource::initDefault()
{
	_16BitIndices = false;
	_vertBuffer = defVertBuffer;
	_indexBuffer = defIndexBuffer;
}


void GeometryResource::release()
{
	if( _vertBuffer != 0 && _vertBuffer != defVertBuffer )
	{
		TheRenderer::instance().unloadBuffers( _vertBuffer, 0 );
		_vertBuffer = 0;
	}
	
	if( _indexBuffer != 0 && _indexBuffer != defIndexBuffer )
	{
		TheRenderer::instance().unloadBuffers( 0, _indexBuffer );
		_indexBuffer = 0;
	}

	_vertData.clear();
	_indices.clear();
	_vertPositions.clear();
	_joints.clear();
	_morphTargets.clear();
}


bool GeometryResource::load( const char *data, unsigned int size )
{
	if( !Resource::load( data, size ) ) return false;

	// Make sure header is available
	if( size < 8 )
	{
		TheEngineLog::instance().writeWarning( "Invalid geometry file" );
		return false;
	}
	
	char *myData = (char *)data;
	
	// Check header and version
	char id[4];
	memcpy( &id, myData, 4 ); myData += 4;
	if( id[0] != 'H' || id[1] != '3' || id[2] != 'D' || id[3] != 'G' )
	{
		TheEngineLog::instance().writeWarning( "Invalid geometry file" );
		return false;
	}
	unsigned int version;
	memcpy( &version, myData, sizeof( int ) ); myData += sizeof( int );
	if( version != 4 )
	{
		TheEngineLog::instance().writeWarning( "Unsupported version of geometry file" );
		return false;
	}

	// Load joints
	unsigned int count;
	memcpy( &count, myData, sizeof( int ) ); myData += sizeof( int );

	if( count > 75 ) TheEngineLog::instance().writeWarning( "Model has more than 75 joints; this may cause defective behavior" );

	_joints.resize( count );
	for( unsigned int i = 0; i < count; ++i )
	{
		Joint &joint = _joints[i];
		
		// Inverse bind matrix
		for( unsigned int j = 0; j < 4; ++j )
		{
			for( unsigned int k = 0; k < 4; ++k )
			{
				float f;
				memcpy( &f, myData, sizeof( float ) ); myData += sizeof( float );
				joint.invBindMat.c[k][j] = f;
			}
		}
	}
	
	// Load geometry data
	memcpy( &count, myData, sizeof( int ) ); myData += sizeof( int );

	_vertPositions.resize( count );
	_vertData.resize( count );
	for( unsigned int i = 0; i < count; ++i )
	{
		VertexData &vertData = _vertData[i];
		char c;

		memcpy( &vertData.pos.x, myData, sizeof( float ) ); myData += sizeof( float );
		memcpy( &vertData.pos.y, myData, sizeof( float ) ); myData += sizeof( float );
		memcpy( &vertData.pos.z, myData, sizeof( float ) ); myData += sizeof( float );
		_vertPositions[i] = vertData.pos;
		
		memcpy( &c, myData, sizeof( char ) ); myData += sizeof( char ); vertData.normal.x = c / 127.0f;
		memcpy( &c, myData, sizeof( char ) ); myData += sizeof( char ); vertData.normal.y = c / 127.0f;
		memcpy( &c, myData, sizeof( char ) ); myData += sizeof( char ); vertData.normal.z = c / 127.0f;
		memcpy( &c, myData, sizeof( char ) ); myData += sizeof( char ); vertData.tangent.x = c / 127.0f;
		memcpy( &c, myData, sizeof( char ) ); myData += sizeof( char ); vertData.tangent.y = c / 127.0f;
		memcpy( &c, myData, sizeof( char ) ); myData += sizeof( char ); vertData.tangent.z = c / 127.0f;
		memcpy( &c, myData, sizeof( char ) ); myData += sizeof( char ); vertData.bitangent.x = c / 127.0f;
		memcpy( &c, myData, sizeof( char ) ); myData += sizeof( char ); vertData.bitangent.y = c / 127.0f;
		memcpy( &c, myData, sizeof( char ) ); myData += sizeof( char ); vertData.bitangent.z = c / 127.0f;

		memcpy( &vertData.u0, myData, sizeof( float ) ); myData += sizeof( float );
		memcpy( &vertData.v0, myData, sizeof( float ) ); myData += sizeof( float );
		memcpy( &vertData.u1, myData, sizeof( float ) ); myData += sizeof( float );
		memcpy( &vertData.v1, myData, sizeof( float ) ); myData += sizeof( float );
		
		// TODO
		//memcpy( &vertData.u2, myData, sizeof( float ) ); myData += sizeof( float );
		//memcpy( &vertData.v2, myData, sizeof( float ) ); myData += sizeof( float );
		//memcpy( &vertData.u3, myData, sizeof( float ) ); myData += sizeof( float );
		//memcpy( &vertData.v3, myData, sizeof( float ) ); myData += sizeof( float );
		myData += sizeof( float ) * 4;

		// Load skinning data if available
		if( !_joints.empty() )
		{
			for( unsigned int j = 0; j < 4; ++j )
			{
				unsigned char uc;
				memcpy( &uc, myData, sizeof( char ) ); myData += sizeof( char );
				vertData.boneVec[j] = (float)uc;
			}
			for( unsigned int j = 0; j < 4; ++j )
			{
				unsigned char uc;
				memcpy( &uc, myData, sizeof( char ) ); myData += sizeof( char );
				vertData.weightVec[j] = uc / 255.0f;
			}
		}
		else
		{
			// Default joint data
			for( unsigned int j = 0; j < 4; ++j )
			{
				vertData.boneVec[j] = 0;
				vertData.weightVec[j] = 0.25;
			}
		}
	}

	memcpy( &count, myData, sizeof( int ) ); myData += sizeof( int );

	_indices.resize( count );
	for( unsigned int i = 0; i < count; ++i )
	{
		memcpy( &_indices[i], myData, sizeof( int ) ); myData += sizeof( int );
	}

	// Load morph targets
	memcpy( &count, myData, sizeof( int ) ); myData += sizeof( int );

	_morphTargets.resize( count );
	for( unsigned int i = 0; i < count; ++i )
	{
		MorphTarget &mt = _morphTargets[i];
		char name[256];
	
		memcpy( name, myData, 256 ); myData += 256;
		mt.name = name;

		unsigned int numVerts;
		memcpy( &numVerts, myData, sizeof( int ) ); myData += sizeof( int );

		mt.diffs.resize( numVerts );
		for( unsigned int j = 0; j < numVerts; ++j )
		{
			MorphDiff &md = mt.diffs[j];
	
			memcpy( &md.vertIndex, myData, sizeof( int ) ); myData += sizeof( int );
			memcpy( &md.posDiff.x, myData, sizeof( float ) ); myData += sizeof( float );
			memcpy( &md.posDiff.y, myData, sizeof( float ) ); myData += sizeof( float );
			memcpy( &md.posDiff.z, myData, sizeof( float ) ); myData += sizeof( float );
			memcpy( &md.normDiff.x, myData, sizeof( float ) ); myData += sizeof( float );
			memcpy( &md.normDiff.y, myData, sizeof( float ) ); myData += sizeof( float );
			memcpy( &md.normDiff.z, myData, sizeof( float ) ); myData += sizeof( float );
			memcpy( &md.tanDiff.x, myData, sizeof( float ) ); myData += sizeof( float );
			memcpy( &md.tanDiff.y, myData, sizeof( float ) ); myData += sizeof( float );
			memcpy( &md.tanDiff.z, myData, sizeof( float ) ); myData += sizeof( float );
			memcpy( &md.bitanDiff.x, myData, sizeof( float ) ); myData += sizeof( float );
			memcpy( &md.bitanDiff.y, myData, sizeof( float ) ); myData += sizeof( float );
			memcpy( &md.bitanDiff.z, myData, sizeof( float ) ); myData += sizeof( float );
		}
	}

	// Add default joint if necessary
	if( _joints.empty() )
	{
		_joints.push_back( Joint() );
	}

	// Upload data
	if( _vertData.size() > 0 && _indices.size() > 0 )
	{
		_vertBuffer = TheRenderer::instance().uploadVertices(
			&_vertData[0], sizeof( VertexData ) * _vertData.size() );
		
		// Convert indices to 16 bit if possible
		if( _indices.size() < 65000 )
		{
			_16BitIndices = true;
			unsigned short *indices = new unsigned short[_indices.size()];
			for( unsigned int i = 0; i < _indices.size(); ++i )
				indices[i] = (unsigned short)_indices[i];
			_indexBuffer = TheRenderer::instance().uploadIndices(
				indices, _indices.size() * sizeof( short ) );
			delete[] indices;
		}
		else
		{
			_indexBuffer = TheRenderer::instance().uploadIndices(
				&_indices[0], _indices.size() * sizeof( int ) );
		}
	}
	
	return true;
}
