// *************************************************************************************************
//
// Horde3D
//   Next-Generation Graphics Engine
// --------------------------------------
// Copyright (C) 2006-2007 Nicolas Schulz
//
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//
// *************************************************************************************************

#include "egSingleton.h"
#include "egModel.h"
#include "egParticle.h"
#include "utPlatform.h"
#include <string>
using namespace std;


#ifndef DLLEXP
#	ifdef PLATFORM_WIN
#		include <windows.h>
#		define DLLEXP extern "C" __declspec( dllexport )
#	else
#		define DLLEXP
#	endif
#endif


// *************************************************************************************************
// Engine core functions
// *************************************************************************************************

namespace Horde3D
{
	// *********************************************************************************************
	// Basic functions
	// *********************************************************************************************
	
	DLLEXP const char *getVersionString()
	{
		return versionString;
	}

	
	DLLEXP bool init()
	{
		return ThePipeline::instance().init();
	}


	DLLEXP void release()
	{
		// Order of destruction is important
		TheSceneManager::destroyInstance();
		ThePipeline::destroyInstance();
		TheResourceManager::destroyInstance();
		TheRenderer::destroyInstance();
		TheEngineLog::destroyInstance();
		TheEngineConfig::destroyInstance();
	}


	DLLEXP void render()
	{
		ThePipeline::instance().render();
	}


	DLLEXP void resize( int x, int y, int width, int height )
	{
		ThePipeline::instance().resize( x, y, width, height );
	}


	DLLEXP void clear()
	{
		TheSceneManager::instance().removeNode( RootNode );
		TheResourceManager::instance().clear();
	}


	DLLEXP bool loadPipelineConfig( const char *configFilename )
	{
		return ThePipeline::instance().loadConfig( configFilename );
	}


	// *********************************************************************************************
	// General functions
	// *********************************************************************************************

	DLLEXP const char *getMessage( unsigned int *level, float *time )
	{
		static string msgText;
		static LogMessage msg;
		
		if( TheEngineLog::instance().getMessage( msg ) )
		{
			if( level != 0x0 ) *level = msg.level;
			if( time != 0x0 ) *time = msg.time;
			return msg.text.c_str();
		}
		else
			return "";
	}

	
	DLLEXP NodeHandle getActiveCamera()
	{
		return TheRenderer::instance().getActiveCamera();
	}
	

	DLLEXP bool setActiveCamera( NodeHandle camNode )
	{
		SceneNode *node = TheSceneManager::instance().resolveNodeHandle( camNode );
		if( node == 0x0 || node->getType() != SceneNodeTypes::Camera ) return false;
		
		TheRenderer::instance().setActiveCamera( (CameraNode *)node );
		return true;
	}
	
	
	DLLEXP float getOption( EngineOptions::List param )
	{
		return TheEngineConfig::instance().getOption( param );
	}
	
	
	DLLEXP bool setOption( EngineOptions::List param, float value )
	{
		return TheEngineConfig::instance().setOption( param, value );
	}


	DLLEXP void showOverlay( float x_ll, float y_ll, float u_ll, float v_ll,
							 float x_lr, float y_lr, float u_lr, float v_lr,
							 float x_ur, float y_ur, float u_ur, float v_ur,
							 float x_ul, float y_ul, float u_ul, float v_ul,
							 unsigned int layer, ResHandle material )
	{
		Resource *res = TheResourceManager::instance().resolveResHandle( material ); 

		if( res != 0x0 && res->getType() == ResourceTypes::Material )
		{
			ThePipeline::instance().showOverlay( Overlay( x_ll, y_ll, u_ll, v_ll,
														  x_lr, y_lr, u_lr, v_lr,
														  x_ur, y_ur, u_ur, v_ur,
														  x_ul, y_ul, u_ul, v_ul,
														  layer ), material );
		}
	}


	// *********************************************************************************************
	// Resource functions
	// *********************************************************************************************

	DLLEXP ResourceTypes::List getResourceType( ResHandle res )
	{
		Resource *r = TheResourceManager::instance().resolveResHandle( res );
		
		if( r != 0x0 ) return r->getType();
		else return ResourceTypes::Undefined;
	}


	DLLEXP ResHandle findResource( ResourceTypes::List type, const char *name )
	{
		Resource *res = TheResourceManager::instance().findResource( type, name );
		
		if( res != 0x0 ) return res->getHandle();
		else return 0;
	}

	
	DLLEXP ResHandle addResource( ResourceTypes::List type, const char *name, unsigned int flags )
	{
		return TheResourceManager::instance().addResource( type, name, flags, true );
	}


	DLLEXP bool removeResource( ResHandle res )
	{
		return TheResourceManager::instance().removeResource( res, true );
	}


	DLLEXP bool loadResource( const char *name, const char *data, unsigned int size )
	{
		return TheResourceManager::instance().loadResource( name, data, size );
	}


	DLLEXP bool unloadResource( ResHandle res )
	{
		Resource *resObj = TheResourceManager::instance().resolveResHandle( res );

		if( resObj != 0x0 )
		{
			resObj->unload();
			return true;
		}
		else
			return false;
	}


	DLLEXP const void *getResourceData( ResHandle res, ResourceData::List param )
	{
		return TheResourceManager::instance().getResourceData( res, param );
	}
	
	
	DLLEXP bool updateResourceData( ResHandle res, ResourceData::List param, const void *data, unsigned int size )
	{
		return TheResourceManager::instance().updateResourceData( res, param, data, size );
	}
	
	
	DLLEXP const char *queryUnloadedResource()
	{
		static string resourceName;
		
		resourceName =  TheResourceManager::instance().queryUnloadedResource();

		return resourceName.c_str();
	}


	DLLEXP const char *getResourcePath( ResourceTypes::List type )
	{
		static string resourcePath;
		
		resourcePath = TheResourceManager::instance().getResourcePath( type );
		
		return resourcePath.c_str();
	}
	
	
	DLLEXP void setResourcePath( ResourceTypes::List type, const char *path )
	{
		TheResourceManager::instance().setResourcePath( type, path );
	}


	DLLEXP void releaseUnusedResources()
	{
		TheResourceManager::instance().releaseUnusedResources();
	}


	DLLEXP bool setMaterialUniform( ResHandle matRes, const char *name, float a, float b, float c, float d )
	{
		Resource *res = TheResourceManager::instance().resolveResHandle( matRes );

		if( res != 0x0 && res->getType() == ResourceTypes::Material )
			return ((MaterialResource *)res)->setUniform( name, a, b, c, d );
		else
			return false;
	}


	// *********************************************************************************************
	// Scene graph functions
	// *********************************************************************************************
	
	DLLEXP SceneNodeTypes::List getNodeType( NodeHandle node )
	{
		SceneNode *sn = TheSceneManager::instance().resolveNodeHandle( node );
		
		if( sn != 0x0 ) return sn->getType();
		else return SceneNodeTypes::Undefined;
	}


	DLLEXP const char *getNodeName( NodeHandle node )
	{
		static string emptyString = "";
		
		SceneNode *sn = TheSceneManager::instance().resolveNodeHandle( node );
		
		if( sn != 0x0 ) return sn->getName().c_str();
		else return emptyString.c_str();
	}


	DLLEXP bool setNodeName( NodeHandle node, const char *name )
	{
		SceneNode *sn = TheSceneManager::instance().resolveNodeHandle( node );
		
		if( sn != 0x0 )
		{	
			sn->setName( name );
			return true;
		}
		else return false;
	}


	DLLEXP NodeHandle getNodeChild( NodeHandle parent, const char *name, unsigned int index, bool recursive )
	{
		SceneNode *sn = TheSceneManager::instance().resolveNodeHandle( parent );
		if( sn == 0x0 )
		{
			TheEngineLog::instance().writeError( "Invalid node handler in getNodeChild" );
			return 0;
		}

		sn = sn->findChild( name, SceneNodeTypes::Undefined, index, recursive );
		
		if( sn != 0x0 ) return sn->getHandle();
		else return 0;
	}

	DLLEXP NodeHandle getNodeParent( NodeHandle node )
	{
		SceneNode* sn = TheSceneManager::instance().resolveNodeHandle( node );
		if ( sn == 0x0 )
		{
			TheEngineLog::instance().writeError( "Invalid node handler in getNodeParent" );
			return 0;
		}
		if ( sn->getParent() != 0x0 ) return sn->getParent()->getHandle();
		else return 0;
	}

	DLLEXP NodeHandle addNodes( NodeHandle parent, ResHandle sgRes )
	{
		SceneNode *parentNode = TheSceneManager::instance().resolveNodeHandle( parent );
		if( parentNode == 0x0 )
		{	
			TheEngineLog::instance().writeError( "Invalid parent node handler in addNodes" );
			return 0;
		}
		
		Resource *res = TheResourceManager::instance().resolveResHandle( sgRes );
		if( res == 0x0 || res->getType() != ResourceTypes::SceneGraph )
		{
			TheEngineLog::instance().writeError( "Invalid resource handler in addNodes" );
			return 0;
		}
		
		TheEngineLog::instance().writeInfo( string( "Adding node from resource " ) + res->getName() );
		return TheSceneManager::instance().addNodes( *parentNode, *(SceneGraphResource *)res );
	}


	DLLEXP bool removeNode( NodeHandle node )
	{
		TheEngineLog::instance().writeInfo( string( "Removing node" ) );
		
		return TheSceneManager::instance().removeNode( node );
	}


	DLLEXP bool setNodeActivation( NodeHandle node, bool active )
	{
		SceneNode *sn = TheSceneManager::instance().resolveNodeHandle( node );
		if( sn != 0x0 )
		{
			sn->setActivation( active );
			return true;
		}
		else
			return false;
	}

	
	DLLEXP bool getNodeTransform( NodeHandle node, float *px, float *py, float *pz,
								  float *rx, float *ry, float *rz, float *sx, float *sy, float *sz )
	{
		SceneNode *sn = TheSceneManager::instance().resolveNodeHandle( node );
		
		if( sn != 0x0 )
		{
			Vec3f pos, rot, scale;
			sn->getTransform( pos, rot, scale );

			if( px != 0x0 ) *px = pos.x; if( py != 0x0 ) *py = pos.y; if( pz != 0x0 ) *pz = pos.z;
			if( rx != 0x0 ) *rx = rot.x; if( ry != 0x0 ) *ry = rot.y; if( rz != 0x0 ) *rz = rot.z;
			if( sx != 0x0 ) *sx = scale.x; if( sy != 0x0 ) *sy = scale.y; if( sz != 0x0 ) *sz = scale.z;
			return true;
		}
		else
			return false;
	}
	
	
	DLLEXP bool setNodeTransform( NodeHandle node, float px, float py, float pz,
								  float rx, float ry, float rz, float sx, float sy, float sz )
	{
		SceneNode *sn = TheSceneManager::instance().resolveNodeHandle( node );
		
		if( sn != 0x0 )
		{
			sn->setTransform( Vec3f( px, py, pz ), Vec3f( rx, ry, rz ), Vec3f( sx, sy, sz ) );
			return true;
		}
		else
			return false;
	}


	DLLEXP bool getNodeTransformMatrices( NodeHandle node, const float **relMat, const float **absMat )
	{
		SceneNode *sn = TheSceneManager::instance().resolveNodeHandle( node );
		if( sn != 0x0 )
		{
			sn->getTransMatrices( relMat, absMat );
			return true;
		}
		else
			return false;
	}


	DLLEXP bool setNodeTransformMatrix( NodeHandle node, const float *mat4x4 )
	{
		SceneNode *sn = TheSceneManager::instance().resolveNodeHandle( node );
		
		if( sn != 0x0 )
		{
			static Matrix4f mat;
			memcpy( mat.c, mat4x4, 16 * sizeof( float ) );
			sn->setTransform( mat );
			return true;
		}
		else
			return false;
	}


	DLLEXP bool getNodeAABB( NodeHandle node, float *minX, float *minY, float *minZ,
							 float *maxX, float *maxY, float *maxZ )
	{
		SceneNode *sn = TheSceneManager::instance().resolveNodeHandle( node );
		if( sn != 0x0 )
		{
			TheSceneManager::instance().updateNodes();
			
			*minX = sn->getBBox().getMinCoords().x;
			*minY = sn->getBBox().getMinCoords().y;
			*minZ = sn->getBBox().getMinCoords().z;
			*maxX = sn->getBBox().getMaxCoords().x;
			*maxY = sn->getBBox().getMaxCoords().y;
			*maxZ = sn->getBBox().getMaxCoords().z;
			return true;
		}
		else
			return false;
	}


	DLLEXP NodeHandle addGroupNode( NodeHandle parent, const char *name )
	{
		SceneNode *parentNode = TheSceneManager::instance().resolveNodeHandle( parent );
		if( parentNode == 0x0 ) return 0;
		
		TheEngineLog::instance().writeInfo( string( "Adding node " ) + name );
		return TheSceneManager::instance().addGroupNode( *parentNode, name );
	}


	DLLEXP float getGroupParam( NodeHandle node, GroupNodeParams::List param )
	{
		SceneNode *sn = TheSceneManager::instance().resolveNodeHandle( node );
		if( sn != 0x0 && sn->getType() == SceneNodeTypes::Group )
		{
			return ((GroupNode *)sn)->getParam( param );
		}
		else
			return Math::NaN;
	}


	DLLEXP bool setGroupParam( NodeHandle node, GroupNodeParams::List param, float value )
	{
		SceneNode *sn = TheSceneManager::instance().resolveNodeHandle( node );
		if( sn != 0x0 && sn->getType() == SceneNodeTypes::Group )
		{
			return ((GroupNode *)sn)->setParam( param, value );
		}
		else
			return false;
	}


	DLLEXP NodeHandle addModelNode( NodeHandle parent, const char *name, ResHandle geoRes )
	{
		SceneNode *parentNode = TheSceneManager::instance().resolveNodeHandle( parent );
		if( parentNode == 0x0 ) return 0;
		
		TheEngineLog::instance().writeInfo( string( "Adding node " ) + name );
		return TheSceneManager::instance().addModelNode( *parentNode, name, geoRes );
	}


	DLLEXP bool setupModelAnimStage( NodeHandle node, unsigned int stage, ResHandle res,
									 const char *animMask, bool additive )
	{
		SceneNode *sn = TheSceneManager::instance().resolveNodeHandle( node );
		if( sn != 0x0 && sn->getType() == SceneNodeTypes::Model )
		{
			string mask = "";
			if( animMask != 0x0 ) mask = animMask;
			return ((ModelNode *)sn)->setupAnimStage( stage, res, mask, additive );
		}
		else if( sn != 0x0 && sn->getType() == SceneNodeTypes::Group )
		{
			// Pass call to children
			bool result = false;
			for( unsigned int i = 0; i < sn->getChildren().size(); ++i )
			{
				result |= setupModelAnimStage( sn->getChildren()[i]->getHandle(), stage, res, animMask, additive );
			}
			return result;
		}
		else
			return false;
	}


	DLLEXP bool setModelAnimParams( NodeHandle node, unsigned int stage, float time, float weight )
	{
		SceneNode *sn = TheSceneManager::instance().resolveNodeHandle( node );
		if( sn != 0x0 && sn->getType() == SceneNodeTypes::Model )
		{
			return ((ModelNode *)sn)->setAnimParams( stage, time, weight );
		}
		else if( sn != 0x0 && sn->getType() == SceneNodeTypes::Group )
		{
			// Pass call to children
			bool result = false;
			for( unsigned int i = 0; i < sn->getChildren().size(); ++i )
			{
				result |= setModelAnimParams( sn->getChildren()[i]->getHandle(), stage, time, weight );
			}
			return result;
		}
		else
			return false;
	}


	DLLEXP bool setModelMorpher( NodeHandle node, const char *target, float weight )
	{
		SceneNode *sn = TheSceneManager::instance().resolveNodeHandle( node );
		if( sn != 0x0 && sn->getType() == SceneNodeTypes::Model )
		{
			return ((ModelNode *)sn)->setMorphParam( target, weight );
		}
		else if( sn != 0x0 && sn->getType() == SceneNodeTypes::Group )
		{
			// Pass call to children
			bool result = false;
			for( unsigned int i = 0; i < sn->getChildren().size(); ++i )
			{
				result |= setModelMorpher( sn->getChildren()[i]->getHandle(), target, weight );
			}
			return result;
		}
		else
			return false;
	}


	DLLEXP NodeHandle addMeshNode( NodeHandle parent, const char *name, ResHandle matRes,
								   unsigned int batchStart, unsigned int batchCount,
								   unsigned int vertRStart, unsigned int vertREnd )
	{
		SceneNode *parentNode = TheSceneManager::instance().resolveNodeHandle( parent );
		if( parentNode == 0x0 ) return 0;
		
		TheEngineLog::instance().writeInfo( string( "Adding node " ) + name );
		return TheSceneManager::instance().addMeshNode( *parentNode, name, matRes, batchStart,
														batchCount, vertRStart, vertREnd );
	}


	DLLEXP float getMeshParam( NodeHandle node, MeshNodeParams::List param )
	{
		SceneNode *sn = TheSceneManager::instance().resolveNodeHandle( node );
		if( sn != 0x0 && sn->getType() == SceneNodeTypes::Mesh )
		{
			return ((MeshNode *)sn)->getParam( param );
		}
		else
			return Math::NaN;
	}


	DLLEXP bool setMeshParam( NodeHandle node, MeshNodeParams::List param, float value )
	{
		SceneNode *sn = TheSceneManager::instance().resolveNodeHandle( node );
		if( sn != 0x0 && sn->getType() == SceneNodeTypes::Mesh )
		{
			return ((MeshNode *)sn)->setParam( param, value );
		}
		else if( sn != 0x0 && (sn->getType() == SceneNodeTypes::Group ||
							   sn->getType() == SceneNodeTypes::Model) )
		{
			// Pass call to children
			bool result = false;
			for( unsigned int i = 0; i < sn->getChildren().size(); ++i )
			{
				result |= setMeshParam( sn->getChildren()[i]->getHandle(), param, value );
			}
			return result;
		}
		else
			return false;
	}


	DLLEXP bool getMeshData( NodeHandle node, unsigned int *vertPosArrayCount, const float **vertPosArrayData,
							 unsigned int *indexArrayCount, const unsigned int **indexArayData, unsigned int *indexOffset )
	{
		SceneNode *sn = TheSceneManager::instance().resolveNodeHandle( node );
		if( sn != 0x0 && sn->getType() == SceneNodeTypes::Mesh )
		{
			((MeshNode *)sn)->getData( vertPosArrayCount, vertPosArrayData, indexArrayCount, indexArayData );
			*indexOffset = ((MeshNode *)sn)->getVertRStart();
			return true;
		}
		else
			return false;
	}


	DLLEXP bool setMeshUniform( NodeHandle node, const char *name, float a, float b, float c, float d, bool recursive )
	{
		SceneNode *sn = TheSceneManager::instance().resolveNodeHandle( node );
		if( sn != 0x0 && sn->getType() == SceneNodeTypes::Mesh )
		{
			return ((MeshNode *)sn)->setUniform( name, a, b, c, d, recursive );
		}
		else if( sn != 0x0 && (sn->getType() == SceneNodeTypes::Group ||
							   sn->getType() == SceneNodeTypes::Model) )
		{
			// Pass call to children
			bool result = false;
			for( unsigned int i = 0; i < sn->getChildren().size(); ++i )
			{
				result |= setMeshUniform( sn->getChildren()[i]->getHandle(), name, a, b, c, d, recursive );
			}
			return result;
		}
		else
			return false;
	}


	DLLEXP NodeHandle addJointNode( NodeHandle parent, const char *name, unsigned int jointIndex )
	{
		SceneNode *parentNode = TheSceneManager::instance().resolveNodeHandle( parent );
		if( parentNode == 0x0 ) return 0;
		
		TheEngineLog::instance().writeInfo( string( "Adding node " ) + name );
		return TheSceneManager::instance().addJointNode( *parentNode, name, jointIndex );
	}


	DLLEXP NodeHandle addLightNode( NodeHandle parent, const char *name, ResHandle materialRes,
									const char *lightingContext, const char *shadowContext )
	{
		SceneNode *parentNode = TheSceneManager::instance().resolveNodeHandle( parent );
		if( parentNode == 0x0 ) return 0;
		
		TheEngineLog::instance().writeInfo( string( "Adding node " ) + name );
		return TheSceneManager::instance().addLightNode( *parentNode, name, materialRes,
														 lightingContext, shadowContext );
	}


	DLLEXP float getLightParam( NodeHandle node, LightNodeParams::List param )
	{
		SceneNode *sn = TheSceneManager::instance().resolveNodeHandle( node );
		if( sn != 0x0 && sn->getType() == SceneNodeTypes::Light )
		{
			return ((LightNode *)sn)->getParam( param );
		}
		else
			return Math::NaN;
	}


	DLLEXP bool setLightParam( NodeHandle node, LightNodeParams::List param, float value )
	{
		SceneNode *sn = TheSceneManager::instance().resolveNodeHandle( node );
		if( sn != 0x0 && sn->getType() == SceneNodeTypes::Light )
		{
			return ((LightNode *)sn)->setParam( param, value );
		}
		else if( sn != 0x0 && sn->getType() == SceneNodeTypes::Group )
		{
			// Pass call to children
			bool result = false;
			for( unsigned int i = 0; i < sn->getChildren().size(); ++i )
			{
				result |= setLightParam( sn->getChildren()[i]->getHandle(), param, value );
			}
			return result;
		}
		else
			return false;
	}


	DLLEXP NodeHandle addCameraNode( NodeHandle parent, const char *name )
	{
		SceneNode *parentNode = TheSceneManager::instance().resolveNodeHandle( parent );
		if( parentNode == 0x0 ) return 0;
		
		TheEngineLog::instance().writeInfo( string( "Adding node " ) + name );
		return TheSceneManager::instance().addCameraNode( *parentNode, name );
	}


	DLLEXP float getCameraParam( NodeHandle node, CameraNodeParams::List param )
	{
		SceneNode *sn = TheSceneManager::instance().resolveNodeHandle( node );
		if( sn != 0x0 && sn->getType() == SceneNodeTypes::Camera )
		{
			return ((CameraNode *)sn)->getParam( param );
		}
		else
			return Math::NaN;
	}


	DLLEXP bool setCameraParam( NodeHandle node, CameraNodeParams::List param, float value )
	{
		SceneNode *sn = TheSceneManager::instance().resolveNodeHandle( node );
		if( sn != 0x0 && sn->getType() == SceneNodeTypes::Camera )
		{
			return ((CameraNode *)sn)->setParam( param, value );
		}
		else if( sn != 0x0 && sn->getType() == SceneNodeTypes::Group )
		{
			// Pass call to children
			bool result = false;
			for( unsigned int i = 0; i < sn->getChildren().size(); ++i )
			{
				result |= setCameraParam( sn->getChildren()[i]->getHandle(), param, value );
			}
			return result;
		}
		else
			return false;
	}


	DLLEXP bool setupCameraView( NodeHandle node, float fov, float aspect,
								 float nearDist, float farDist )
	{
		SceneNode *sn = TheSceneManager::instance().resolveNodeHandle( node );
		if( sn != 0x0 && sn->getType() == SceneNodeTypes::Camera )
		{
			((CameraNode *)sn)->setupViewParams( fov, aspect, nearDist, farDist );
			return true;
		}
		else if( sn != 0x0 && sn->getType() == SceneNodeTypes::Group )
		{
			// Pass call to children
			for( unsigned int i = 0; i < sn->getChildren().size(); ++i )
			{
				setupCameraView( sn->getChildren()[i]->getHandle(), fov, aspect, nearDist, farDist );
			}
			return true;
		}
		else
			return false;	
	}


	DLLEXP NodeHandle addEmitterNode( NodeHandle parent, const char *name,
									  ResHandle matRes, ResHandle effectRes,
									  unsigned int maxParticleCount, int respawnCount )
	{
		SceneNode *parentNode = TheSceneManager::instance().resolveNodeHandle( parent );
		if( parentNode == 0x0 ) return 0;
		
		TheEngineLog::instance().writeInfo( string( "Adding node " ) + name );
		return TheSceneManager::instance().addEmitterNode( *parentNode, name, matRes, effectRes,
														   maxParticleCount, respawnCount );
	}


	DLLEXP float getEmitterParam( NodeHandle node, EmitterNodeParams::List param )
	{
		SceneNode *sn = TheSceneManager::instance().resolveNodeHandle( node );
		if( sn != 0x0 && sn->getType() == SceneNodeTypes::Emitter )
		{
			return ((EmitterNode *)sn)->getParam( param );
		}
		else
			return Math::NaN;
	}


	DLLEXP bool setEmitterParam( NodeHandle node, EmitterNodeParams::List param, float value )
	{
		SceneNode *sn = TheSceneManager::instance().resolveNodeHandle( node );
		if( sn != 0x0 && sn->getType() == SceneNodeTypes::Emitter )
		{
			return ((EmitterNode *)sn)->setParam( param, value );
		}
		else if( sn != 0x0 && sn->getType() == SceneNodeTypes::Group )
		{
			// Pass call to children
			bool result = false;
			for( unsigned int i = 0; i < sn->getChildren().size(); ++i )
			{
				result |= setEmitterParam( sn->getChildren()[i]->getHandle(), param, value );
			}
			return result;
		}
		else
			return false;
	}


	DLLEXP bool advanceEmitterTime( NodeHandle node, float timeDelta )
	{
		SceneNode *sn = TheSceneManager::instance().resolveNodeHandle( node );
		if( sn != 0x0 && sn->getType() == SceneNodeTypes::Emitter )
		{
			((EmitterNode *)sn)->advanceTime( timeDelta );
			return true;
		}
		else if( sn != 0x0 && sn->getType() == SceneNodeTypes::Group )
		{
			// Pass call to children
			for( unsigned int i = 0; i < sn->getChildren().size(); ++i )
			{
				advanceEmitterTime( sn->getChildren()[i]->getHandle(), timeDelta );
			}
			return true;
		}
		else
			return false;	
	}
}


// *************************************************************************************************
// DLL entry point
// *************************************************************************************************


#ifdef PLATFORM_WIN
BOOL APIENTRY DllMain( HANDLE hModule, DWORD ul_reason_for_call, LPVOID lpReserved )
{
	#if defined( _MSC_VER ) && defined( _DEBUG )
	//_crtBreakAlloc = 1397;
	_CrtSetDbgFlag ( _CRTDBG_ALLOC_MEM_DF | _CRTDBG_LEAK_CHECK_DF );
	#endif
	
	/*switch( ul_reason_for_call )
	{
	case DLL_PROCESS_DETACH:
		_CrtDumpMemoryLeaks();
		break;
	}*/
	
	return TRUE;
}
#endif
