// *************************************************************************************************
//
// Horde3D
//   Next-Generation Graphics Engine
// --------------------------------------
// Copyright (C) 2006-2007 Nicolas Schulz
//
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//
// *************************************************************************************************

#include "egMaterial.h"
#include "egTextures.h"
#include "egShader.h"
#include "egCom.h"
#include "egSingleton.h"
#include "tinyxml.h"
#include "utPlatform.h"

#include "utDebug.h"


MaterialResource::MaterialResource( const string &name, unsigned int flags ) :
	Resource( ResourceTypes::Material, name, flags )
{
	initDefault();	
}


MaterialResource::~MaterialResource()
{
	release();
}


void MaterialResource::initDefault()
{
	_shaderRes = 0x0;
	_class = "";
}


void MaterialResource::release()
{
	if( _shaderRes != 0x0 )
	{	
		_shaderRes->spinRefCount( false );
		_shaderRes = 0x0;
	}
	
	for( unsigned int i = 0; i < _texUnits.size(); ++i )
	{
		if( _texUnits[i].texRes != 0x0 )
		{
			_texUnits[i].texRes->spinRefCount( false );
			_texUnits[i].texRes = 0x0;
		}
	}

	_texUnits.clear();
	_uniforms.clear();
}


bool MaterialResource::raiseError( TiXmlDocument *doc )
{
	// Reset
	release();
	initDefault();

	TheEngineLog::instance().writeWarning( "Invalid XML syntax in material file" );
	
	doc->Clear();
	delete doc;
	
	return false;
}


bool MaterialResource::load( const char *data, unsigned int size )
{
	if( !Resource::load( data, size ) ) return false;

	// Parse material
	TiXmlDocument *doc = new TiXmlDocument();
	doc->SetCondenseWhiteSpace( false );

	doc->Parse( data );
	if( doc->Error() ) return raiseError( doc );

	TiXmlElement *root = doc->RootElement();
	if( root == 0x0 || strcmp( root->Value(), "Material" ) != 0 ) return raiseError( doc );

	// Class
    if( root->Attribute( "class" ) != 0x0 ) _class = root->Attribute( "class" );
    
    // Shader
	TiXmlElement *node1 = root->FirstChildElement( "Shader" );
	if( node1 != 0x0 )
	{
		if( node1->Attribute( "name" ) == 0x0 ) return raiseError( doc );
			
		unsigned int shader = TheResourceManager::instance().addResource(
				ResourceTypes::Shader, node1->Attribute( "name" ), 0, false );
		_shaderRes = (ShaderResource *)TheResourceManager::instance().resolveResHandle( shader );
	}

	// Texture units
	node1 = root->FirstChildElement( "TexUnit" );
	while( node1 != 0x0 )
	{
		if( node1->Attribute( "unit" ) == 0x0 ) return raiseError( doc );
		if( node1->Attribute( "map" ) == 0x0 ) return raiseError( doc );
		
		TexUnit texUnit;
		texUnit.unit = atoi( node1->Attribute( "unit" ) );

		unsigned int texMap;
		unsigned int flags = 0;

		if( node1->Attribute( "allowCompression" ) != 0x0 &&
			_stricmp( node1->Attribute( "allowCompression" ), "false" ) == 0 )
		{
			flags |= ResourceFlags::NoCompression;
		}
	
		if( node1->Attribute( "type" ) != 0x0 && _stricmp( node1->Attribute( "type" ), "CUBE" ) == 0 )
		{
			texMap = TheResourceManager::instance().addResource( ResourceTypes::TextureCube, 
						node1->Attribute( "map" ), flags, false );
		}
		else
		{
			texMap = TheResourceManager::instance().addResource( ResourceTypes::Texture2D,
						node1->Attribute( "map" ), flags, false );
		}

		texUnit.texRes = TheResourceManager::instance().resolveResHandle( texMap );
		
		_texUnits.push_back( texUnit );
		
		node1 = node1->NextSiblingElement( "TexUnit" );
	}
		
	// Vector uniforms
	node1 = root->FirstChildElement( "Uniform" );
	while( node1 != 0x0 )
	{
		if( node1->Attribute( "name" ) == 0x0 ) return raiseError( doc );

		Uniform uniform;
		uniform.name = node1->Attribute( "name" );

		if( node1->Attribute( "a" ) != 0x0 ) uniform.values[0] = (float)atof( node1->Attribute( "a" ) );
		if( node1->Attribute( "b" ) != 0x0 ) uniform.values[1] = (float)atof( node1->Attribute( "b" ) );
		if( node1->Attribute( "c" ) != 0x0 ) uniform.values[2] = (float)atof( node1->Attribute( "c" ) );
		if( node1->Attribute( "d" ) != 0x0 ) uniform.values[3] = (float)atof( node1->Attribute( "d" ) );

		_uniforms.push_back( uniform );

		node1 = node1->NextSiblingElement( "Uniform" );
	}

	doc->Clear();
	delete doc;
	
	return true;
}


bool MaterialResource::setUniform( const string &name, float a, float b, float c, float d )
{
	for( unsigned int i = 0; i < _uniforms.size(); ++i )
	{
		if( _uniforms[i].name == name )
		{
			_uniforms[i].values[0] = a;
			_uniforms[i].values[1] = b;
			_uniforms[i].values[2] = c;
			_uniforms[i].values[3] = d;
			return true;
		}
	}

	return false;
}


bool MaterialResource::isOfClass( const string &theClass )
{
	static string theClass2;
	
	if( theClass != "" )
	{
		if( theClass[0]	!= '~' )
		{
			if( _class.find( theClass, 0 ) != 0 ) return false;
			if( _class.length() > theClass.length() && _class[theClass.length()] != '.' ) return false;
		}
		else	// Not operator
		{
			theClass2 = theClass.substr( 1, theClass.length() - 1);
			
			if( _class.find( theClass2, 0 ) == 0 )
			{
				if( _class.length() == theClass2.length() )
				{
					return false;
				}
				else
				{
					if( _class[theClass2.length()] == '.' ) return false;
				}
			}
		}
	}
	
	return true;
}
