// *************************************************************************************************
//
// Horde3D
//   Next-Generation Graphics Engine
// --------------------------------------
// Copyright (C) 2006-2007 Nicolas Schulz
//
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//
// *************************************************************************************************

#include "egModel.h"
#include "egMaterial.h"
#include "egSingleton.h"

#include "utDebug.h"


// *************************************************************************************************
// Class MeshNode
// *************************************************************************************************

MeshNode::MeshNode( const string &name, SceneNode &parent, unsigned int matRes,
				    unsigned int batchStart, unsigned int batchCount,
					unsigned int vertRStart, unsigned int vertREnd ) :
	SceneNode( SceneNodeTypes::Mesh, name )
{
	Resource *res = TheResourceManager::instance().resolveResHandle( matRes );
	if( res != 0x0 )
	{
		_materialRes = (MaterialResource *)res;
		_materialRes->spinRefCount( true );
	}
	else
		_materialRes = 0x0;
	
	_batchStart = batchStart;
	_batchCount = batchCount;
	_vertRStart = vertRStart;
	_vertREnd = vertREnd;

	// Find parent model node
	SceneNode *node = &parent;
	while( node->getType() != SceneNodeTypes::Model ) node = node->getParent();
	_parentModel = (ModelNode *)node;
	
	// Calculate local bounding box
	GeometryResource &geoRes = _parentModel->getGeometryResource();
	Vec3f &bBMin = _localBBox.getMinCoords();
	Vec3f &bBMax = _localBBox.getMaxCoords();
	
	if( _vertRStart < geoRes._vertPositions.size() && _vertREnd < geoRes._vertPositions.size() )
	{
		bBMin = Vec3f( Math::MaxFloat, Math::MaxFloat, Math::MaxFloat );
		bBMax = Vec3f( -Math::MaxFloat, -Math::MaxFloat, -Math::MaxFloat );
		for( unsigned int i = _vertRStart; i <= _vertREnd; ++i )
		{
			Vec3f &vertPos = geoRes._vertPositions[i];

			if( vertPos.x < bBMin.x ) bBMin.x = vertPos.x;
			if( vertPos.y < bBMin.y ) bBMin.y = vertPos.y;
			if( vertPos.z < bBMin.z ) bBMin.z = vertPos.z;
			if( vertPos.x > bBMax.x ) bBMax.x = vertPos.x;
			if( vertPos.y > bBMax.y ) bBMax.y = vertPos.y;
			if( vertPos.z > bBMax.z ) bBMax.z = vertPos.z;
		}
	}
	else
	{
		bBMin = Vec3f( 0, 0, 0 );
		bBMax = Vec3f( 0, 0, 0 );
	}

	_animatableListIndex = 0;
}


MeshNode::~MeshNode()
{
	if( _materialRes != 0x0 ) _materialRes->spinRefCount( false );
}


bool MeshNode::canAttach( SceneNode &parent )
{
	return (parent.getType() == SceneNodeTypes::Model) ||
		   (parent.getType() == SceneNodeTypes::Mesh);
}


float MeshNode::getParam( MeshNodeParams::List param )
{
	switch( param )
	{
	case MeshNodeParams::MaterialRes:
		if( _materialRes != 0x0 ) return (float)_materialRes->getHandle();
		else return 0;
	default:
		return Math::NaN;
	}
}


bool MeshNode::setParam( MeshNodeParams::List param, float value )
{
	Resource *res;
	
	switch( param )
	{
	case MeshNodeParams::MaterialRes:
		res = TheResourceManager::instance().resolveResHandle( (unsigned int)value );
		if( res == 0x0 || res->getType() != ResourceTypes::Material )
		{	
			TheEngineLog::instance().writeError( "Invalid base resource for mesh node" );
			return false;
		}
		
		if( _materialRes != 0x0 ) _materialRes->spinRefCount( false );
		_materialRes = (MaterialResource *)res;
		_materialRes->spinRefCount( true );
		_uniforms.resize( 0 );
		return true;
	default:
		return false;
	}
}


void MeshNode::getData( unsigned int *vertPosArrayCount, const float **vertPosArrayData,
						unsigned int *indexArrayCount, const unsigned int **indexArrayData )
{
	if( vertPosArrayCount != 0x0 )
	{	
		if( _batchCount > 0 )
			*vertPosArrayCount = _vertREnd - _vertRStart + 1;
		else
			*vertPosArrayCount = 0;
	}
	
	if( vertPosArrayData != 0x0 )
	{
		if( _vertRStart < _parentModel->getGeometryResource()._vertPositions.size() &&
			_vertREnd < _parentModel->getGeometryResource()._vertPositions.size() )
			*vertPosArrayData = &_parentModel->getGeometryResource()._vertPositions[_vertRStart].x;
		else
			*vertPosArrayData = 0x0;
	}
	
	if( indexArrayCount != 0x0 ) *indexArrayCount = _batchCount;

	if( indexArrayData != 0x0 )
	{
		if( _batchStart + _batchCount <= _parentModel->getGeometryResource()._indices.size() )
			*indexArrayData = &_parentModel->getGeometryResource()._indices[_batchStart];
		else
			*indexArrayData = 0x0;
	}
}


bool MeshNode::setUniform( const string &name, float a, float b, float c, float d, bool recursive )
{
	if( recursive )
	{
		// Handle children
		for( unsigned int i = 0; i < _children.size(); ++i )
		{
			if( _children[i]->getType() == SceneNodeTypes::Mesh )
			{
				((MeshNode *)_children[i])->setUniform( name, a, b, c, d, recursive );
			}
		}
	}
	
	// Check if the uniform is already in the list
	for( unsigned int i = 0; i < _uniforms.size(); ++i )
	{
		if( _uniforms[i].name == name )
		{
			_uniforms[i].values[0] = a;
			_uniforms[i].values[1] = b;
			_uniforms[i].values[2] = c;
			_uniforms[i].values[3] = d;
			return true;
		}
	}

	// Check if material defines that uniform
	if( _materialRes != 0x0 )
	{
		for( unsigned int i = 0; i < _materialRes->_uniforms.size(); ++i )
		{
			if( _materialRes->_uniforms[i].name == name )
			{
				Uniform unif;
				unif.name = name;
				unif.values[0] = a;
				unif.values[1] = b;
				unif.values[2] = c;
				unif.values[3] = d;
				_uniforms.push_back( unif );
				return true;
			}
		}
	}
	
	return false;
}


void MeshNode::onAttach( SceneNode &node )
{
	// Forward event to parent model
	_parentModel->onAttach( node );
}


void MeshNode::onDetach( SceneNode &node )
{
	// Forward event to parent model
	_parentModel->onDetach( node );
}


// *************************************************************************************************
// Class JointNode
// *************************************************************************************************

JointNode::JointNode( const string &name, SceneNode &parent, unsigned int jointIndex ) :
	SceneNode( SceneNodeTypes::Joint, name )
{
	_jointIndex = jointIndex;

	// Find parent model node
	SceneNode *node = &parent;
	while( node->getType() != SceneNodeTypes::Model ) node = node->getParent();
	_parentModel = (ModelNode *)node;

	_animatableListIndex = 0;
}


bool JointNode::canAttach( SceneNode &parent )
{
	return (parent.getType() == SceneNodeTypes::Model) ||
		   (parent.getType() == SceneNodeTypes::Joint);
}


//float JointNode::getParam( JointNodeParams::List param )
//{
//	return Math::NaN;
//}


//bool JointNode::setParam( JointNodeParams::List param, float value )
//{
//	return false;
//}


void JointNode::onUpdate()
{
	if( _parent->getType() != SceneNodeTypes::Joint )
		_relModelMat = _relTrans;
	else
		//_relModelMat = ((JointNode *)_parent)->_relModelMat * _relTrans;
		_relModelMat.fastMult( ((JointNode *)_parent)->_relModelMat, _relTrans );

	if( _parentModel->jointExists( _jointIndex ) )
	{
		Matrix4f mat = _relModelMat * _parentModel->getGeometryResource().getInvBindMat( _jointIndex );
		_parentModel->setSkinningMat( _jointIndex, mat.getRow( 0 ), mat.getRow( 1 ), mat.getRow( 2 ) );
	}
}


void JointNode::onAttach( SceneNode &node )
{
	// Forward event to parent model
	_parentModel->onAttach( node );
}


void JointNode::onDetach( SceneNode &node )
{
	// Forward event to parent model
	_parentModel->onDetach( node );
}


// *************************************************************************************************
// Class ModelNode
// *************************************************************************************************

ModelNode::ModelNode( const string &name, GeometryResource &geoRes ) :
	SceneNode( SceneNodeTypes::Model, name ), _geoRes( geoRes )
{
	for( unsigned int i = 0; i < MaxNumAnimStages; ++i ) _animStages[i] = 0x0;
	
	_morphVBO = 0;
	_animDirty = true;

	_geoRes.spinRefCount( true );
		
	// Init joint data
	_skinMatRows.resize( _geoRes._joints.size() * 3 );
	for( unsigned int i = 0; i < _skinMatRows.size() / 3; ++i )
	{
		_skinMatRows[i * 3 + 0] = Vec4f( 1, 0, 0, 0 );
		_skinMatRows[i * 3 + 1] = Vec4f( 0, 1, 0, 0 );
		_skinMatRows[i * 3 + 2] = Vec4f( 0, 0, 1, 0 );
	}

	_animatableNodes.reserve( _geoRes._joints.size() + 5 );	// Guess number of meshes with 5

	// Copy morph targets
	_morpherUsed = false;
	if( !_geoRes._morphTargets.empty() )
	{
		_morphers.resize( _geoRes._morphTargets.size() );
		for( unsigned int i = 0; i < _morphers.size(); ++i )
		{	
			Morpher &morpher = _morphers[i]; 
			
			morpher.name = _geoRes._morphTargets[i].name;
			morpher.index = i;
			morpher.weight = 0;
		}
	
		// Create VBO for morph targets
		_morphVerts.resize( _geoRes._vertData.size() );
		for( unsigned int i = 0; i < _morphVerts.size(); ++i )
		{
			MorphVertex &mv = _morphVerts[i];
			VertexData &vd = _geoRes._vertData[i];

			mv.pos = vd.pos;
			mv.normal = vd.normal;
			mv.tangent = vd.tangent;
			mv.bitangent = vd.bitangent;
		}
		_morphVBO = TheRenderer::instance().uploadVertices( (float *)&_morphVerts[0],
			sizeof( MorphVertex ) * (unsigned int)_morphVerts.size() );
	}
}


ModelNode::~ModelNode()
{
	_geoRes.spinRefCount( false );
	if( _morphVBO != 0 ) TheRenderer::instance().unloadBuffers( _morphVBO, 0 );

	for( unsigned int i = 0; i < MaxNumAnimStages; ++i )
	{	
		if( _animStages[i] != 0x0 && _animStages[i]->anim != 0x0 )
			_animStages[i]->anim->spinRefCount( false );
		delete _animStages[i];
	}
}


//float ModelNode::getParam( ModelNodeParams::List param )
//{
//	return Math::NaN;
//}


//bool ModelNode::setParam( ModelNodeParams::List param, float value )
//{
//	return false;
//}


void ModelNode::setAnimMask( SceneNode *node, AnimStage &animStage )
{
	if( node == 0x0 ) return;
	
	unsigned int index;
	
	if( node->getType() == SceneNodeTypes::Joint )
		index = ((JointNode *)node)->_animatableListIndex;
	else if( node->getType() == SceneNodeTypes::Mesh )
		index = ((MeshNode *)node)->_animatableListIndex;
	else return;

	animStage.animMask[index] = !animStage.animMask[index];

	for( unsigned int i = 0; i < node->getChildren().size(); ++i )
	{
		setAnimMask( node->getChildren()[i], animStage );
	}
}


bool ModelNode::setupAnimStage( unsigned int stage, unsigned int animRes, const string &animMask, bool additive )
{
	if( stage > MaxNumAnimStages ) return false;

	Resource *res = TheResourceManager::instance().resolveResHandle( animRes );
	
	// Resource ID 0 is allowed for erasing stage
	if( animRes != 0 )
	{
		if( res == 0x0 || res->getType() != ResourceTypes::Animation )
		{
			TheEngineLog::instance().writeError( "Invalid animation resource for model node" );
			return false;
		}
	}
	
	AnimStage *curStage = _animStages[stage];
	if( animRes == 0 )
	{
		// Erase stage
		if( curStage != 0x0 )
		{
			if( curStage->anim != 0x0 ) curStage->anim->spinRefCount( false );	
			delete _animStages[stage]; _animStages[stage] = 0x0;
		}
		
		markDirty();	// Mark scene node as dirty so that update function is called
		_animDirty = true;
		
		return true;
	}
	else if( curStage != 0x0 )
	{
		// Reset stage
		if( curStage->anim != 0x0 ) curStage->anim->spinRefCount( false );

		for( unsigned int i = 0; i < curStage->animLookup.size(); ++i )
		{
			curStage->animLookup[i] = 0x0;
			curStage->animMask[i] = true;
		}
	}
	else
	{
		// Create stage
		_animStages[stage] = new AnimStage();
		_animStages[stage]->animLookup.resize( _animatableNodes.size(), 0x0 );
		_animStages[stage]->animMask.resize( _animatableNodes.size(), true );
		curStage = _animStages[stage];
	}
	
	curStage->additive = additive;
	curStage->anim = (AnimationResource *)res;
	curStage->anim->spinRefCount( true );
	
	// Update lookup table
	for( unsigned int i = 0; i < _animatableNodes.size(); ++i )
	{
		if( _animatableNodes[i] == 0x0 ) continue;
		
		curStage->animLookup[i] = curStage->anim->findNode( _animatableNodes[i]->getName() );
	}

	// Initialize masking
	if( animMask != "" )
	{
		// Reset mask
		for( unsigned int i = 0; i < curStage->animMask.size(); ++i )
			curStage->animMask[i] = false;
		
		unsigned int start = 0;
		unsigned int end = 0;
		while( end <= animMask.size() )
		{
			end = animMask.find( "<>", start );
			string nodeName = animMask.substr( start, end - start );
			start = end + 2;

			SceneNode *node = findChild( nodeName, SceneNodeTypes::Joint, 0, true );
			if( node == 0x0 ) node = findChild( nodeName, SceneNodeTypes::Mesh, 0, true );
			
			setAnimMask( node, *curStage );
		}
	}

	return setAnimParams( stage, 0.0f, 1.0f );
}


bool ModelNode::setAnimParams( unsigned int stage, float time, float weight )
{
	if( stage > MaxNumAnimStages ) return false;

	AnimStage *curStage = _animStages[stage];
	if( curStage == 0x0 || curStage->anim == 0x0 ) return false;

	curStage->animTime = time;
	curStage->blendWeight = weight;

	markDirty();	// Mark scene node as dirty so that update function is called
	_animDirty = true;

	update();
	
	return true;
}


bool ModelNode::setMorphParam( const string &targetName, float weight )
{
	if( _morphers.size() == 0 ) return false;
	
	_morpherUsed = false;
	
	bool result = false;

	// Set specified morph target or all targets if targetName == ""
	for( unsigned int i = 0; i < _morphers.size(); ++i )
	{
		if( targetName == "" || _morphers[i].name == targetName )
		{
			_morphers[i].weight = weight;
			result = true;
		}

		if( _morphers[i].weight > 0 ) _morpherUsed = true;
	}

	if( !_morpherUsed ) return result;

	// Recalculate vertex positions
	// Set morph target vertex attributes to base mesh attributes
	for( unsigned int i = 0; i < _morphVerts.size(); ++i )
	{
		MorphVertex &mv = _morphVerts[i];
		VertexData &vd = _geoRes._vertData[i];
		
		mv.pos = vd.pos;
		mv.normal = vd.normal;
		mv.tangent = vd.tangent;
		mv.bitangent = vd.bitangent;
	}
	
	for( unsigned int i = 0; i < _morphers.size(); ++i )
	{
		MorphTarget &mt = _geoRes._morphTargets[_morphers[i].index];
		float weight = _morphers[i].weight;
		
		for( unsigned int j = 0; j < mt.diffs.size(); ++j )
		{
			MorphVertex &mv = _morphVerts[mt.diffs[j].vertIndex];
			MorphDiff &md = mt.diffs[j];
			
			// TODO: Better use normalized blending
			mv.pos += md.posDiff * weight;
			mv.normal += md.normDiff * weight;
			mv.tangent += md.tanDiff * weight;
			mv.bitangent += md.bitanDiff * weight;
			
			// Renormalize
			// TODO: Does this have to happen in this loop?
			mv.normal = mv.normal.normalized();
			mv.tangent = mv.tangent.normalized();
			mv.bitangent = mv.bitangent.normalized();
		}
	}

	TheRenderer::instance().uploadVertices( (float *)&_morphVerts[0],
			sizeof( MorphVertex ) * (unsigned int)_morphVerts.size(), _morphVBO );

	return result;
}


void ModelNode::onUpdate()
{
	if( !_animDirty ) return;
	_animDirty = false;
	
	// Find active stages
	static vector< AnimStage * > activeBlendStages, activeAddStages;
	activeBlendStages.resize( 0 );
	activeAddStages.resize( 0 );
	for( unsigned int i = 0; i < MaxNumAnimStages; ++i )
	{
		if( _animStages[i] != 0x0 && _animStages[i]->anim != 0x0 )
		{	
			if( _animStages[i]->additive )
				activeAddStages.push_back( _animStages[i] );
			else
				activeBlendStages.push_back( _animStages[i] );
		}
	}

	// Animate
	// TODO: Create more optimized path for fast animation
	if( activeBlendStages.size() == 1 && activeAddStages.empty() && TheEngineConfig::instance().fastAnimation )
	{
		// Fast animation path
		AnimStage *curStage = activeBlendStages[0];
		
		for( unsigned int i = 0; i < _animatableNodes.size(); ++i )
		{
			if( _animatableNodes[i] == 0x0 || curStage->animLookup[i] == 0x0 ) continue;
			
			AnimNode &an = *curStage->animLookup[i];
			if( an.frames.size() > 0 && curStage->animMask[i] )
			{
				unsigned int frame = (unsigned int)curStage->animTime % an.frames.size();
				Matrix4f &curAnimTrans = an.frames[frame].bakedTransMat;
				
				_animatableNodes[i]->getRelTrans() = curAnimTrans;
			}
		}

		// Mark joint nodes as dirty
		markChildrenDirty();
	}
	else
	{
		// Animation blending
		for( unsigned int i = 0; i < _animatableNodes.size(); ++i )
		{
			if( _animatableNodes[i] == 0x0 ) continue;
			
			Quaternion rot;
			Vec3f trans, scale;

			bool matGenerated = false;
			float accum = 0.0f;

			for( unsigned int j = 0; j < activeBlendStages.size(); ++j )
			{
				AnimStage *curStage = activeBlendStages[j];
				if( !curStage->animMask[i] ) continue;

				if( curStage->animLookup[i] != 0x0 )
				{
					unsigned int numFrames = (unsigned int)curStage->animLookup[i]->frames.size();
					
					if( numFrames > 0 )
					{
						matGenerated = true;

						float weight = curStage->blendWeight;
						
						if( weight == 0.0f ) weight = 0.0001f;	// Avoid division by zero
						if( accum + weight > 1.0f ) weight = 1.0f - accum;

						// Fast animation with sampled frame data
						if( TheEngineConfig::instance().fastAnimation )
						{
							unsigned int f0 = (int)curStage->animTime % numFrames;
							Frame &frame = curStage->animLookup[i]->frames[f0];
							
							if( j == 0 )
							{
								rot = frame.rotQuat;
								trans = frame.transVec;
								scale = frame.scaleVec;
							}
							else
							{
								float blend = accum / (accum + weight);
								
								rot = frame.rotQuat.slerp( rot, blend );
								trans = frame.transVec.lerp( trans, blend );
								scale = frame.scaleVec.lerp( scale, blend );
							}	
						}
						// Animation with inter-frame interpolation
						else
						{
							unsigned int f0 = (int)curStage->animTime;
							float amount = curStage->animTime - f0;
							f0 = f0 % numFrames;
							unsigned int f1 = f0 + 1;
							if( f1 > numFrames - 1 ) f1 = numFrames - 1;
							
							Frame &frame0 = curStage->animLookup[i]->frames[f0];
							Frame &frame1 = curStage->animLookup[i]->frames[f1];
							
							// Inter-frame interpolation
							Vec3f transVec = frame0.transVec.lerp( frame1.transVec, amount );
							Vec3f scaleVec = frame0.scaleVec.lerp( frame1.scaleVec, amount );
							Quaternion rotQuat = frame0.rotQuat.slerp( frame1.rotQuat, amount );

							if( j == 0 )
							{
								rot = rotQuat;
								trans = transVec;
								scale = scaleVec;
							}
							else
							{
								float blend = accum / (accum + weight);
								
								rot = rotQuat.slerp( rot, blend );
								trans = transVec.lerp( trans, blend );
								scale = scaleVec.lerp( scale, blend );
							}
						}

						accum += curStage->blendWeight;
					}
				}

				if( accum >= 1.0f ) break;
			}
			
			if( matGenerated )
			{
				Matrix4f mat;
				mat.scale( scale.x, scale.y, scale.z );
				mat = Matrix4f( rot ) * mat;
				mat.translate( trans.x, trans.y, trans.z );
				_animatableNodes[i]->getRelTrans() = mat;
			}

			// Additive animations
			Matrix4f mat;
			matGenerated = false;
			for( unsigned int j = 0; j < activeAddStages.size(); ++j )
			{
				AnimStage *curStage = activeAddStages[j];
				if( !curStage->animMask[i] ) continue; 

				if( curStage->animLookup[i] != 0x0 )
				{
					unsigned int numFrames = (unsigned int)curStage->animLookup[i]->frames.size();
					
					if( numFrames > 0 )
					{
						matGenerated = true;

						if( TheEngineConfig::instance().fastAnimation )
						{
							unsigned int f0 = (int)curStage->animTime % numFrames;
							Frame &frame = curStage->animLookup[i]->frames[f0];
							
							mat = mat * curStage->animLookup[i]->firstFrameInvTrans * frame.bakedTransMat;
						}
						else
						{
							unsigned int f0 = (int)curStage->animTime;
							float amount = curStage->animTime - f0;
							f0 = f0 % numFrames;
							unsigned int f1 = f0 + 1;
							if( f1 > numFrames - 1 ) f1 = numFrames - 1;
							Frame &frame0 = curStage->animLookup[i]->frames[f0];
							Frame &frame1 = curStage->animLookup[i]->frames[f1];
							
							// Inter-frame interpolation
							Vec3f transVec = frame0.transVec.lerp( frame1.transVec, amount );
							Vec3f scaleVec = frame0.scaleVec.lerp( frame1.scaleVec, amount );
							Quaternion rotQuat = frame0.rotQuat.slerp( frame1.rotQuat, amount );
							
							Matrix4f frameMat;
							frameMat.scale( scaleVec.x, scaleVec.y, scaleVec.z );
							frameMat = Matrix4f( rotQuat ) * frameMat;
							frameMat.translate( transVec.x, transVec.y, transVec.z );
							
							mat = mat * curStage->animLookup[i]->firstFrameInvTrans * frameMat;
						}
					}
				}
			}

			if( matGenerated )
			{
				_animatableNodes[i]->getRelTrans() = _animatableNodes[i]->getRelTrans() * mat;
			}
		}

		// Mark joint nodes as dirty
		markChildrenDirty();
	}
}


void ModelNode::onAttach( SceneNode &node )
{
	// TODO: Update animation lookup
	
	if( node.getType() == SceneNodeTypes::Joint )
		((JointNode *)&node)->_animatableListIndex = _animatableNodes.size();
	else if( node.getType() == SceneNodeTypes::Mesh )
		((MeshNode *)&node)->_animatableListIndex = _animatableNodes.size();
	else
		return;

	_animatableNodes.push_back( &node );
}


void ModelNode::onDetach( SceneNode &node )
{
	// TODO: Update animation lookup
	
	if( node.getType() == SceneNodeTypes::Joint ||
		node.getType() == SceneNodeTypes::Mesh )
	{
		for( unsigned int i = 0; i < _animatableNodes.size(); ++i )
		{
			if( _animatableNodes[i] == &node )
			{
				_animatableNodes[i] = 0x0;
				break;
			}
		}
	}
}
