// *************************************************************************************************
//
// Horde3D
//   Next-Generation Graphics Engine
// --------------------------------------
// Copyright (C) 2006-2007 Nicolas Schulz
//
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//
// *************************************************************************************************

#pragma once

#include "egScene.h"
#include "egGeometry.h"
#include "egAnimation.h"
#include "egMaterial.h"
#include "utMath.h"

class MaterialResource;
class ModelNode;


const unsigned int MaxNumAnimStages = 16;


class MeshNode : public SceneNode
{
protected:

	MaterialResource	*_materialRes;
	unsigned int		_batchStart, _batchCount;
	unsigned int		_vertRStart, _vertREnd;
	
	ModelNode			*_parentModel;
	BoundingBox			_localBBox;

	unsigned int		_animatableListIndex;

	vector< Uniform >	_uniforms;

public:

	MeshNode( const string &name, SceneNode &parent, unsigned int matRes,
			  unsigned int batchStart, unsigned int batchCount, 
			  unsigned int vertRStart, unsigned int vertREnd );
	~MeshNode();

	BoundingBox *getLocalBBox() { return &_localBBox; }
	bool canAttach( SceneNode &parent );
	float getParam( MeshNodeParams::List param );
	bool setParam( MeshNodeParams::List param, float value );
	void getData( unsigned int *vertPosArrayCount, const float **vertPosArrayData,
				  unsigned int *indexArrayCount, const unsigned int **indexArayData );
	bool setUniform( const string &name, float a, float b, float c, float d, bool recursive );

	void onAttach( SceneNode &node );
	void onDetach( SceneNode &node );

	MaterialResource *getMaterialRes() { return _materialRes; }
	unsigned int getBatchStart() { return _batchStart; }
	unsigned int getBatchCount() { return _batchCount; }
	unsigned int getVertRStart() { return _vertRStart; }
	unsigned int getVertREnd() { return _vertREnd; }
	ModelNode *getParentModel() { return _parentModel; }
	vector< Uniform > &getUniforms() { return _uniforms; }

	friend class ModelNode;
};


class JointNode : public SceneNode
{
protected:

	ModelNode		*_parentModel;
	unsigned int	_jointIndex;
	Matrix4f		_relModelMat;		// Transformation relative to parent model

	unsigned int	_animatableListIndex;

public:

	JointNode( const string &name, SceneNode &parent, unsigned int jointIndex );
	
	bool canAttach( SceneNode &parent );
	//float getParam( JointNodeParams::List param );
	//bool setParam( JointNodeParams::List param, float value );

	void onUpdate();
	void onAttach( SceneNode &node );
	void onDetach( SceneNode &node );

	friend class ModelNode;
};


struct Morpher	// Morph modifier
{
	string			name;
	unsigned int	index;		//Index of morph target
	float			weight;
};


struct MorphVertex
{
	Vec3f			pos;
	Vec3f			normal, tangent, bitangent;
};


struct AnimStage
{
	AnimationResource			*anim;
	bool						additive;
	float						animTime;
	float						blendWeight;
	vector< AnimNode * >		animLookup;
	vector< bool >				animMask;
};


class ModelNode : public SceneNode
{
protected:

	GeometryResource			&_geoRes;
	vector< Vec4f >				_skinMatRows;
	vector< SceneNode * >		_animatableNodes;
	
	bool						_morpherUsed;
	vector< Morpher >			_morphers;
	vector< MorphVertex >		_morphVerts;	
	unsigned int				_morphVBO;
	
	AnimStage					*_animStages[MaxNumAnimStages];
	bool						_animDirty;

	static void setAnimMask( SceneNode *node, AnimStage &animStage );

public:

	ModelNode( const string &name, GeometryResource &geoRes );
	~ModelNode();

	//float getParam( ModelNodeParams::List param );
	//bool setParam( ModelNodeParams::List param, float value );
	bool setupAnimStage( unsigned int stage, unsigned int animRes, const string &animMask, bool additive );
	bool setAnimParams( unsigned int stage, float time, float weight );
	bool setMorphParam( const string &targetName, float weight );

	void onUpdate();
	void onAttach( SceneNode &node );
	void onDetach( SceneNode &node );

	GeometryResource &getGeometryResource() { return _geoRes; }
	bool jointExists( unsigned int jointIndex ) { return jointIndex < _skinMatRows.size() / 3; }
	void setSkinningMat( unsigned int index, const Vec4f &row0, const Vec4f &row1, const Vec4f &row2 )
		{ _skinMatRows[index * 3 + 0] = row0; _skinMatRows[index * 3 + 1] = row1;
		  _skinMatRows[index * 3 + 2] = row2; }

	friend class Renderer;
};
