// *************************************************************************************************
//
// Horde3D
//   Next-Generation Graphics Engine
// --------------------------------------
// Copyright (C) 2006-2007 Nicolas Schulz
//
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//
// *************************************************************************************************

#include "egParticle.h"
#include "egMaterial.h"
#include "egSingleton.h"
#include "tinyxml.h"
#include "utPlatform.h"

#include "utDebug.h"


// *************************************************************************************************
// ParticleChannel
// *************************************************************************************************

ParticleChannel::ParticleChannel()
{
	reset();
}


void ParticleChannel::reset()
{
	startMin = 0; startMax = 0; endRate = 0;
}


bool ParticleChannel::parse( TiXmlElement *node )
{
	if( node->Attribute( "startMin" ) == 0x0 ) return false;
	startMin = (float)atof( node->Attribute( "startMin" ) );

	if( node->Attribute( "startMax" ) != 0x0 )
		startMax = (float)atof( node->Attribute( "startMax" ) );
	else
		startMax = startMin;
	
	if( node->Attribute( "endRate" ) != 0x0 )
		endRate = (float)atof( node->Attribute( "endRate" ) );
	else
		endRate = 1.0f;

	return true;
}


// *************************************************************************************************
// EffectResource
// *************************************************************************************************

EffectResource::EffectResource( const string &name, unsigned int flags ) :
	Resource( ResourceTypes::Effect, name, flags )
{
	initDefault();	
}


EffectResource::~EffectResource()
{
	release();
}


void EffectResource::initDefault()
{
	_lifeMin = 0; _lifeMax = 0;
	_moveVel.reset();
	_rotVel.reset();
	_size.reset();
	_colR.reset();
	_colG.reset();
	_colB.reset();
	_colA.reset();
}


void EffectResource::release()
{
}


bool EffectResource::raiseError( TiXmlDocument *doc )
{
	TheEngineLog::instance().writeWarning( "Invalid XML syntax in effect file" );
	
	doc->Clear();
	delete doc;
	
	return false;
}


bool EffectResource::load( const char *data, unsigned int size )
{
	if( !Resource::load( data, size ) ) return false;

	// Parse effect resource
	TiXmlDocument *doc = new TiXmlDocument();
	doc->SetCondenseWhiteSpace( false );

	doc->Parse( data );
	if( doc->Error() ) return raiseError( doc );

	TiXmlElement *root = doc->RootElement();
	if( root == 0x0 || strcmp( root->Value(), "ParticleConfig" ) != 0 ) return raiseError( doc );
	
	if( root->Attribute( "lifeMin" ) == 0x0 ) return raiseError( doc );
	if( root->Attribute( "lifeMax" ) == 0x0 ) return raiseError( doc );

	_lifeMin = (float)atof( root->Attribute( "lifeMin" ) );
	_lifeMax = (float)atof( root->Attribute( "lifeMax" ) );

	TiXmlElement *node1 = root->FirstChildElement( "ChannelOverLife" );
	while( node1 != 0x0 )
	{
		if( node1->Attribute( "channel" ) != 0x0 )
		{
			if( _stricmp( node1->Attribute( "channel" ), "moveVel" ) == 0 ) _moveVel.parse( node1 );
			else if( _stricmp( node1->Attribute( "channel" ), "size" ) == 0 ) _size.parse( node1 );
			else if( _stricmp( node1->Attribute( "channel" ), "rotVel" ) == 0 ) _rotVel.parse( node1 );
			else if( _stricmp( node1->Attribute( "channel" ), "colR" ) == 0 ) _colR.parse( node1 );
			else if( _stricmp( node1->Attribute( "channel" ), "colG" ) == 0 ) _colG.parse( node1 );
			else if( _stricmp( node1->Attribute( "channel" ), "colB" ) == 0 ) _colB.parse( node1 );
			else if( _stricmp( node1->Attribute( "channel" ), "colA" ) == 0 ) _colA.parse( node1 );
		}

		node1 = node1->NextSiblingElement( "ChannelOverLife" );
	}

	doc->Clear();
	delete doc;
	
	return true;
}


// *************************************************************************************************
// EmitterNode
// *************************************************************************************************

EmitterNode::EmitterNode( const string &name, unsigned int matRes, unsigned int effectRes,
						  unsigned int maxParticleCount, int respawnCount ) :
	SceneNode( SceneNodeTypes::Emitter, name )
{
	Resource *res = TheResourceManager::instance().resolveResHandle( matRes );
	if( res != 0x0 && res->getType() == ResourceTypes::Material )
	{
		_matRes = (MaterialResource *)res;
		_matRes->spinRefCount( true );
	}
	else
		_matRes = 0x0;

	res = TheResourceManager::instance().resolveResHandle( effectRes );
	if( res != 0x0 && res->getType() == ResourceTypes::Effect )
	{
		_effectRes = (EffectResource *)res;
		_effectRes->spinRefCount( true );
	}
	else
		_effectRes = 0x0;

	_particleCount = maxParticleCount;
	_respawnCount = respawnCount;
	_delay = 0; _emissionRate = 0; _spreadAngle = 0;
	_force = Vec3f( 0, 0, 0 );

	_emissionAccum = 0;
	
	// Initialize particles
	_particles = new ParticleData[_particleCount];
	_parPositions = new Vec3f[_particleCount];
	_parSizesANDRotations = new float[_particleCount * 2];
	_parColors = new float[_particleCount * 4];
	for( unsigned int i = 0; i < _particleCount; ++i )
	{
		_particles[i].life = 0;
		_particles[i].respawnCounter = 0;
		
		_parPositions[i] = Vec3f( 0, 0, 0 );
		_parSizesANDRotations[i*2+0] = 0.0f;
		_parSizesANDRotations[i*2+1] = 0.0f;
		_parColors[i*4+0] = 0.0f;
		_parColors[i*4+1] = 0.0f;
		_parColors[i*4+2] = 0.0f;
		_parColors[i*4+3] = 0.0f;
	}
}


EmitterNode::~EmitterNode()
{
	if( _matRes != 0x0 ) _matRes->spinRefCount( false );
	if( _effectRes != 0x0 ) _effectRes->spinRefCount( false );

	delete[] _particles;
	delete[] _parPositions;
	delete[] _parSizesANDRotations;
	delete[] _parColors;
}


float EmitterNode::getParam( EmitterNodeParams::List param )
{
	switch( param )
	{
	case EmitterNodeParams::Delay:
		return _delay;
	case EmitterNodeParams::EmissionRate:
		return _emissionRate;
	case EmitterNodeParams::SpreadAngle:
		return _spreadAngle;
	case EmitterNodeParams::ForceX:
		return _force.x;
	case EmitterNodeParams::ForceY:
		return _force.y;
	case EmitterNodeParams::ForceZ:
		return _force.z;
	default:
		return Math::NaN;
	}
}


bool EmitterNode::setParam( EmitterNodeParams::List param, float value )
{
	switch( param )
	{
	case EmitterNodeParams::Delay:
		_delay = value;
		return true;
	case EmitterNodeParams::EmissionRate:
		_emissionRate = value;
		return true;
	case EmitterNodeParams::SpreadAngle:
		_spreadAngle = value;
		return true;
	case EmitterNodeParams::ForceX:
		_force.x = value;
		return true;
	case EmitterNodeParams::ForceY:
		_force.y = value;
		return true;
	case EmitterNodeParams::ForceZ:
		_force.z = value;
		return true;
	default:
		return false;
	}
}


float randomF( float min, float max )
{
	if ((int)(max * 100.0f) - (int)(min * 100.0f) == 0)
	{
		return min;
	}
	else
	{
		return (rand()%((int)(max * 100.0f) - (int)(min * 100.0f))) / 100.0f + min;
	}
}


void EmitterNode::advanceTime( float timeDelta )
{
	if( _effectRes == 0x0 ) return;
	
	Vec3f bBMin( Math::MaxFloat, Math::MaxFloat, Math::MaxFloat );
	Vec3f bBMax( -Math::MaxFloat, -Math::MaxFloat, -Math::MaxFloat );
	
	if( _delay <= 0 )
		_emissionAccum += _emissionRate * timeDelta;
	else
		_delay -= timeDelta;
	
	for( unsigned int i = 0; i < _particleCount; ++i )
	{
		ParticleData &p = _particles[i];
		
		// Create particle
		if( p.life <= 0 && ((int)p.respawnCounter < _respawnCount || _respawnCount < 0) )
		{
			if( _emissionAccum >= 1.0f )
			{
				// Respawn
				p.maxLife = randomF( _effectRes->_lifeMin, _effectRes->_lifeMax );
				p.life = p.maxLife;
				float angle = degToRad( _spreadAngle / 2 );
				Matrix4f m = _absTrans;
				m.c[3][0] = 0; m.c[3][1] = 0; m.c[3][2] = 0;
				m.rotate( randomF( -angle, angle ), randomF( -angle, angle ), randomF( -angle, angle ) );
				p.dir = (m * Vec3f( 0, 0, -1 )).normalized();
				++p.respawnCounter;

				// Generate start values
				p.moveVel0 = randomF( _effectRes->_moveVel.startMin, _effectRes->_moveVel.startMax );
				p.rotVel0 = randomF( _effectRes->_rotVel.startMin, _effectRes->_rotVel.startMax );
				p.size0 = randomF( _effectRes->_size.startMin, _effectRes->_size.startMax );
				p.r0 = randomF( _effectRes->_colR.startMin, _effectRes->_colR.startMax );
				p.g0 = randomF( _effectRes->_colG.startMin, _effectRes->_colG.startMax );
				p.b0 = randomF( _effectRes->_colB.startMin, _effectRes->_colB.startMax );
				p.a0 = randomF( _effectRes->_colA.startMin, _effectRes->_colA.startMax );
				
				// Update arrays
				_parPositions[i] = Vec3f( _absTrans.c[3][0], _absTrans.c[3][1], _absTrans.c[3][2] );
				_parSizesANDRotations[i * 2 + 0] = p.size0;
				_parSizesANDRotations[i * 2 + 1] = randomF( 0, 360 );
				_parColors[i * 4 + 0] = p.r0;
				_parColors[i * 4 + 1] = p.g0;
				_parColors[i * 4 + 2] = p.b0;
				_parColors[i * 4 + 3] = p.a0;

				// Update emitter
				_emissionAccum -= 1;
				if( _emissionAccum < 0 ) _emissionAccum = 0;
			}
		}
		
		// Update particle
		if( p.life > 0 )
		{
			// Interpolate data
			float fac = 1.0f - (p.life / p.maxLife);
			
			float moveVel = p.moveVel0 * (1.0f + (_effectRes->_moveVel.endRate - 1.0f) * fac);
			float rotVel = p.rotVel0 * (1.0f + (_effectRes->_rotVel.endRate - 1.0f) * fac);
			_parSizesANDRotations[i * 2 + 0] = p.size0 * (1.0f + (_effectRes->_size.endRate - 1.0f) * fac);
			_parColors[i * 4 + 0] = p.r0 * (1.0f + (_effectRes->_colR.endRate - 1.0f) * fac);
			_parColors[i * 4 + 1] = p.g0 * (1.0f + (_effectRes->_colG.endRate - 1.0f) * fac);
			_parColors[i * 4 + 2] = p.b0 * (1.0f + (_effectRes->_colB.endRate - 1.0f) * fac);
			_parColors[i * 4 + 3] = p.a0 * (1.0f + (_effectRes->_colA.endRate - 1.0f) * fac);

			// Update particle position and rotation
			_parPositions[i] += p.dir * moveVel * timeDelta;
			_parPositions[i] += _force * timeDelta;
			_parSizesANDRotations[i * 2+ 1] +=  rotVel * timeDelta;

			// Decrease lifetime
			p.life -= timeDelta;
			
			// Check if particle is dying
			if( p.life <= 0 )
			{
				_parSizesANDRotations[i * 2 + 0] = 0.0f;
			}
		}

		// Update bounding box
		Vec3f &vertPos = _parPositions[i];
		if( vertPos.x < bBMin.x ) bBMin.x = vertPos.x;
		if( vertPos.y < bBMin.y ) bBMin.y = vertPos.y;
		if( vertPos.z < bBMin.z ) bBMin.z = vertPos.z;
		if( vertPos.x > bBMax.x ) bBMax.x = vertPos.x;
		if( vertPos.y > bBMax.y ) bBMax.y = vertPos.y;
		if( vertPos.z > bBMax.z ) bBMax.z = vertPos.z;
	}

	_localBBox.getMinCoords() = bBMin;
	_localBBox.getMaxCoords() = bBMax;
	
	markDirty();
}
