// *************************************************************************************************
//
// Horde3D
//   Next-Generation Graphics Engine
// --------------------------------------
// Copyright (C) 2006-2007 Nicolas Schulz
//
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//
// *************************************************************************************************

#include "egPipeline.h"
#include "egSingleton.h"
#include "tinyxml.h"
#include "utPlatform.h"
#include <fstream>
using namespace std;

#include "utDebug.h"


Pipeline::~Pipeline()
{
	reset();
}


void Pipeline::reset()
{
	destroyRenderTargets();

	for( unsigned int i = 0; i < _pipelineCommands.size(); ++i )
	{
		for( unsigned int j = 0; j < _pipelineCommands[i].valParams.size(); ++j )
		{
			delete _pipelineCommands[i].valParams[j];
		}

		for( unsigned int j = 0; j < _pipelineCommands[i].resParams.size(); ++j )
		{
			TheResourceManager::instance().removeResource( _pipelineCommands[i].resParams[j], false );
		}
	}

	_pipelineCommands.clear();
}


void Pipeline::addRenderTarget( const string &id, bool depthBuffer, unsigned int numBuffers,
							   RenderBufferFormats::List format, float scale )
{
	if( numBuffers > 4 ) return;
	
	RenderTarget rt;
	
	rt.id = id;
	rt.hasDepthBuffer = depthBuffer;
	rt.numTexBuffers = numBuffers;
	rt.format = format;
	rt.scale = scale;

	_renderTargets.push_back( rt );
}


RenderTarget *Pipeline::findRenderTarget( const string &name )
{
	for( unsigned int i = 0; i < _renderTargets.size(); ++i )
	{
		if( _renderTargets[i].id == name )
		{
			return &_renderTargets[i];
		}
	}
	
	return 0x0;
}


bool Pipeline::createRenderTargets()
{
	for( unsigned int i = 0; i < _renderTargets.size(); ++i )
	{
		RenderTarget &rt = _renderTargets[i];
	
		unsigned int width = (int)(TheRenderer::instance().getVPWidth() * rt.scale);
		unsigned int height = (int)(TheRenderer::instance().getVPHeight() * rt.scale);
		
		rt.rendBuf = TheRenderer::instance().createRenderBuffer(
						width, height, rt.format, rt.hasDepthBuffer, rt.numTexBuffers );
		if( rt.rendBuf.fbo == 0 ) return false;
	}
	
	return true;
}


void Pipeline::destroyRenderTargets()
{
	for( unsigned int i = 0; i < _renderTargets.size(); ++i )
	{
		RenderTarget &rt = _renderTargets[i];
		
		TheRenderer::instance().destroyRenderBuffer( rt.rendBuf );
	}
}


bool Pipeline::raiseError( TiXmlDocument *doc )
{
	TheEngineLog::instance().writeWarning( "Invalid XML syntax in pipeline config" );
	
	doc->Clear();
	delete doc;
	
	return false;
}


bool Pipeline::parsePipelineCommands( TiXmlElement *node )
{
	TiXmlElement *node1 = node->FirstChildElement();

	while( node1 != 0x0 )
	{
		if( strcmp( node1->Value(), "SetTarget" ) == 0 )
		{
			if( node1->Attribute( "target" ) == 0x0 ) return false;
			
			if( strcmp( node1->Attribute( "target" ), "" ) == 0 )
			{
				_pipelineCommands.push_back( PipelineCommand( PipelineCommands::SetRenderTarget ) );
				(_pipelineCommands.end() - 1)->refParams.push_back( 0x0 );
			}
			else
			{
				if( findRenderTarget( node1->Attribute( "target" ) ) == 0x0 ) return false;
				_pipelineCommands.push_back( PipelineCommand( PipelineCommands::SetRenderTarget ) );
				(_pipelineCommands.end() - 1)->refParams.push_back( findRenderTarget( node1->Attribute( "target" ) ) );
			}
		}
		else if( strcmp( node1->Value(), "BindBuffers" ) == 0 )
		{
			if( strcmp( node1->Attribute( "target" ), "" ) == 0 ) return false;
			if( findRenderTarget( node1->Attribute( "target" ) ) == 0x0 ) return false;
			_pipelineCommands.push_back( PipelineCommand( PipelineCommands::BindBuffers ) );
			(_pipelineCommands.end() - 1)->refParams.push_back( findRenderTarget( node1->Attribute( "target" ) ) );
		}
		else if( strcmp( node1->Value(), "Clear" ) == 0 )
		{
			_pipelineCommands.push_back( PipelineCommand( PipelineCommands::Clear ) );
			
			vector< PCParam * > &valParams = (_pipelineCommands.end() - 1)->valParams;
			valParams.resize( 9 );
			valParams[0] = new PCBoolParam( false );
			valParams[1] = new PCBoolParam( false );
			valParams[2] = new PCBoolParam( false );
			valParams[3] = new PCBoolParam( false );
			valParams[4] = new PCBoolParam( false );
			valParams[5] = new PCFloatParam( 0.0f );
			valParams[6] = new PCFloatParam( 0.0f );
			valParams[7] = new PCFloatParam( 0.0f );
			valParams[8] = new PCFloatParam( 0.0f );
			
			if( node1->Attribute( "depth" ) != 0x0 && _stricmp( node1->Attribute( "depth" ), "true" ) == 0 )
				((PCBoolParam *)valParams[0])->set( true );
			
			if( node1->Attribute( "buf0" ) != 0x0 && _stricmp( node1->Attribute( "buf0" ), "true" ) == 0 )
				((PCBoolParam *)valParams[1])->set( true );

			if( node1->Attribute( "buf1" ) != 0x0 && _stricmp( node1->Attribute( "buf1" ), "true" ) == 0 )
				((PCBoolParam *)valParams[2])->set( true );

			if( node1->Attribute( "buf2" ) != 0x0 && _stricmp( node1->Attribute( "buf2" ), "true" ) == 0 )
				((PCBoolParam *)valParams[3])->set( true );

			if( node1->Attribute( "buf3" ) != 0x0 && _stricmp( node1->Attribute( "buf3" ), "true" ) == 0 )
				((PCBoolParam *)valParams[4])->set( true );

			if( node1->Attribute( "r" ) != 0x0 )
				((PCFloatParam *)valParams[5])->set( (float)atof( node1->Attribute( "r" ) ) );

			if( node1->Attribute( "g" ) != 0x0 )
				((PCFloatParam *)valParams[6])->set( (float)atof( node1->Attribute( "g" ) ) );

			if( node1->Attribute( "b" ) != 0x0 )
				((PCFloatParam *)valParams[7])->set( (float)atof( node1->Attribute( "b" ) ) );

			if( node1->Attribute( "a" ) != 0x0 )
				((PCFloatParam *)valParams[8])->set( (float)atof( node1->Attribute( "a" ) ) );
		}
		else if( strcmp( node1->Value(), "DrawGeometry" ) == 0 )
		{
			if( node1->Attribute( "context" ) == 0x0 ) return false;
			_pipelineCommands.push_back( PipelineCommand( PipelineCommands::DrawGeometry ) );
			(_pipelineCommands.end() - 1)->valParams.push_back( new PCStringParam( node1->Attribute( "context" ) ) );
            if( node1->Attribute( "class" ) != 0x0 ) 
                (_pipelineCommands.end() - 1)->valParams.push_back( new PCStringParam( node1->Attribute( "class" ) ) );
            else
                (_pipelineCommands.end() - 1)->valParams.push_back( new PCStringParam( "" ) );
		}
		else if( strcmp( node1->Value(), "DrawOverlays" ) == 0 )
		{
			if( node1->Attribute( "context" ) == 0x0 ) return false;
			_pipelineCommands.push_back( PipelineCommand( PipelineCommands::DrawOverlays ) );
			(_pipelineCommands.end() - 1)->valParams.push_back( new PCStringParam( node1->Attribute( "context" ) ) );
		}
		else if( strcmp( node1->Value(), "DrawFSQuad" ) == 0 )
		{
			if( node1->Attribute( "material" ) == 0x0 ) return false;
			if( node1->Attribute( "context" ) == 0x0 ) return false;
			_pipelineCommands.push_back( PipelineCommand( PipelineCommands::DrawFSQuad ) );
			unsigned int matRes = TheResourceManager::instance().addResource(
				ResourceTypes::Material, node1->Attribute( "material" ), 0, false );
			(_pipelineCommands.end() - 1)->resParams.push_back( matRes );
			(_pipelineCommands.end() - 1)->valParams.push_back( new PCStringParam( node1->Attribute( "context" ) ) );
		}
		else if( strcmp( node1->Value(), "ForEachLight_DrawGeometry" ) == 0 )
		{
		    _pipelineCommands.push_back( PipelineCommand( PipelineCommands::DrawLightGeometry ) );
		    if( node1->Attribute( "class" ) != 0x0 )
                (_pipelineCommands.end() - 1)->valParams.push_back( new PCStringParam( node1->Attribute( "class" ) ) );
            else
                (_pipelineCommands.end() - 1)->valParams.push_back( new PCStringParam( "" ) );
		}
		else if( strcmp( node1->Value(), "ForEachLight_DrawShape" ) == 0 )
		{
			_pipelineCommands.push_back( PipelineCommand( PipelineCommands::DrawLightShapes ) );
		}

		node1 = node1->NextSiblingElement();
	}

	return true;
}


bool Pipeline::loadConfig( const char *configData )
{
	// TODO: Load data from memory rather than file directly

	reset();
	
// *** Load data ***
	ifstream inf( configData, ios::binary );
	if( !inf ) return false;
	// Find size of resource file
	inf.seekg( 0, ios::end );
	int size = inf.tellg();
	// Copy resource file to memory
	char *data = new char[size + 1];
	inf.seekg( 0 );
	inf.read( data, size );
	inf.close();
	// Null-terminate buffer - this is important for XML parsers 
	data[size] = '\0';
	
// *** Parse data ***
	TiXmlDocument *doc = new TiXmlDocument();
	doc->SetCondenseWhiteSpace( false );

	doc->Parse( data );
	if( doc->Error() ) return raiseError( doc );

	TiXmlElement *root = doc->RootElement();
	if( root == 0x0 || strcmp( root->Value(), "Pipeline" ) != 0 ) return raiseError( doc );

	// Engine configuration
	TiXmlElement *node1 = root->FirstChildElement( "EngineConfig" );
	if( node1 != 0x0 )
	{
		if( node1->Attribute( "trilinearFiltering" ) != 0x0 )
		{
			if( _stricmp( node1->Attribute( "trilinearFiltering" ), "false" ) == 0 ||
				strcmp( node1->Attribute( "trilinearFiltering" ), "0" ) == 0 )
			{
				TheEngineConfig::instance().setOption( EngineOptions::TrilinearFiltering, 0 );
			}
			else
				TheEngineConfig::instance().setOption( EngineOptions::TrilinearFiltering, 1 );
		}
		if( node1->Attribute( "anisotropyFactor" ) != 0x0 )
		{
			TheEngineConfig::instance().setOption( EngineOptions::AnisotropyFactor,
				(float)atof( node1->Attribute( "anisotropyFactor" ) ) );
		}
		if( node1->Attribute( "texCompression" ) != 0x0 )
		{
			if( _stricmp( node1->Attribute( "texCompression" ), "false" ) == 0 ||
				strcmp( node1->Attribute( "texCompression" ), "0" ) == 0 )
			{
				TheEngineConfig::instance().setOption( EngineOptions::TexCompression, 0 );
			}
			else
				TheEngineConfig::instance().setOption( EngineOptions::TexCompression, 1 );
		}
		if( node1->Attribute( "loadTextures" ) != 0x0 )
		{
			if( _stricmp( node1->Attribute( "loadTextures" ), "false" ) == 0 ||
				strcmp( node1->Attribute( "loadTextures" ), "0" ) == 0 )
			{
				TheEngineConfig::instance().setOption( EngineOptions::LoadTextures, 0 );
			}
			else
				TheEngineConfig::instance().setOption( EngineOptions::LoadTextures, 1 );
		}
		if( node1->Attribute( "fastAnimation" ) != 0x0 )
		{
			if( _stricmp( node1->Attribute( "fastAnimation" ), "false" ) == 0 ||
				strcmp( node1->Attribute( "fastAnimation" ), "0" ) == 0 )
			{
				TheEngineConfig::instance().setOption( EngineOptions::FastAnimation, 0 );
			}
			else
				TheEngineConfig::instance().setOption( EngineOptions::FastAnimation, 1 );
		}
		if( node1->Attribute( "occlusionCulling" ) != 0x0 )
		{
			if( _stricmp( node1->Attribute( "occlusionCulling" ), "false" ) == 0 ||
				strcmp( node1->Attribute( "occlusionCulling" ), "0" ) == 0 )
			{
				TheEngineConfig::instance().setOption( EngineOptions::OcclusionCulling, 0 );
			}
			else
				TheEngineConfig::instance().setOption( EngineOptions::OcclusionCulling, 1 );
		}
		if( node1->Attribute( "shadowMapSize" ) != 0x0 )
		{
			TheEngineConfig::instance().setOption( EngineOptions::ShadowMapSize,
				(float)atof( node1->Attribute( "shadowMapSize" ) ) );
		}
		if( node1->Attribute( "debugViewMode" ) != 0x0 )
		{
			if( _stricmp( node1->Attribute( "debugViewMode" ), "false" ) == 0 ||
				strcmp( node1->Attribute( "debugViewMode" ), "0" ) == 0 )
			{
				TheEngineConfig::instance().setOption( EngineOptions::DebugViewMode, 0 );
			}
			else
				TheEngineConfig::instance().setOption( EngineOptions::DebugViewMode, 1 );
		}
	}
	
	// RenderTarget initialization
	node1 = root->FirstChildElement( "Init" );
	if( node1 != 0x0 )
	{
		TiXmlElement *node2 = node1->FirstChildElement( "RenderTarget" );
		while( node2 != 0x0 )
		{
			if( node2->Attribute( "id" ) == 0x0 ) return raiseError( doc );
			string id = node2->Attribute( "id" );
			
			if( node2->Attribute( "format" ) == 0x0 ) return raiseError( doc );
			RenderBufferFormats::List format;
			if( _stricmp( node2->Attribute( "format" ), "RGBA8" ) == 0 )
				format = RenderBufferFormats::RGBA8;
			else if( _stricmp( node2->Attribute( "format" ), "RGBA16F" ) == 0 )
				format = RenderBufferFormats::RGBA16F;
			else if( _stricmp( node2->Attribute( "format" ), "RGBA32F" ) == 0 )
				format = RenderBufferFormats::RGBA32F;
			else return raiseError( doc );

			bool depth = false;
			if( node2->Attribute( "depth" ) != 0x0 && _stricmp( node2->Attribute( "depth" ), "true" ) == 0 )
				depth = true;
						
			float scale = 1.0f;
			if( node2->Attribute( "scale" ) != 0x0 )
				scale = (float)atof( node2->Attribute( "scale" ) );

			unsigned int numBuffers = 1;
			if( node2->Attribute( "numBuffers" ) != 0x0 )
				numBuffers = atoi( node2->Attribute( "numBuffers" ) );

			addRenderTarget( id, depth, numBuffers, format, scale );

			node2 = node2->NextSiblingElement( "RenderTarget" );
		}
	}

	// Pipeline commands
	node1 = root->FirstChildElement( "Stage" );
	while( node1 != 0x0 )
	{
		if( !parsePipelineCommands( node1 ) ) return raiseError( doc );
		
		node1 = node1->NextSiblingElement( "Stage" );
	}
	
	doc->Clear();
	delete doc;
	delete[] data;


// *** Setup data ***
	// Create render targets
	if( !createRenderTargets() )
	{
		TheEngineLog::instance().writeError( "Failed to create render target" );
		return false;
	}

	return true;
}


void Pipeline::processCommands()
{
	for( unsigned int i = 0; i < _pipelineCommands.size(); ++i )
	{
		PipelineCommand &pc = _pipelineCommands[i];
		RenderTarget *rt;

		switch( pc.command )
		{
		case PipelineCommands::SetRenderTarget:
			rt = (RenderTarget *)pc.refParams[0];
			if( rt != 0x0 ) TheRenderer::instance().setRenderBuffer( &rt->rendBuf );
			else TheRenderer::instance().setRenderBuffer( 0x0 );

			TheRenderer::instance().bindBuffers( 0x0 );
			break;

		case PipelineCommands::BindBuffers:
			rt = (RenderTarget *)pc.refParams[0];
		
			TheRenderer::instance().bindBuffers( &rt->rendBuf );
			break;

		case PipelineCommands::Clear:
			TheRenderer::instance().clear( ((PCBoolParam *)pc.valParams[0])->get(),
				((PCBoolParam *)pc.valParams[1])->get(), ((PCBoolParam *)pc.valParams[2])->get(),
				((PCBoolParam *)pc.valParams[3])->get(), ((PCBoolParam *)pc.valParams[4])->get(),
				((PCFloatParam *)pc.valParams[5])->get(), ((PCFloatParam *)pc.valParams[6])->get(),
				((PCFloatParam *)pc.valParams[7])->get(), ((PCFloatParam *)pc.valParams[8])->get() );
			break;

		case PipelineCommands::DrawGeometry:
			TheRenderer::instance().drawGeometry( ((PCStringParam *)pc.valParams[0])->get(),
												  ((PCStringParam *)pc.valParams[1])->get() );
			break;

		case PipelineCommands::DrawOverlays:
			TheRenderer::instance().drawOverlays( ((PCStringParam *)pc.valParams[0])->get(), _overlays );
			break;

		case PipelineCommands::DrawFSQuad:
			TheRenderer::instance().drawFSQuad( pc.resParams[0],
												((PCStringParam *)pc.valParams[0])->get() );
		break;

		case PipelineCommands::DrawLightGeometry:
			TheRenderer::instance().drawLightGeometry( ((PCStringParam *)pc.valParams[0])->get() );
			break;

		case PipelineCommands::DrawLightShapes:
			TheRenderer::instance().drawLightShapes();
			break;
		}
	}
}


bool Pipeline::init()
{
	if( !TheRenderer::instance().init() ) return false;
	if( !TheResourceManager::instance().init() ) return false;

	_overlays.reserve( 100 );
	
	return true;
}


void Pipeline::render()
{
	if( !TheEngineConfig::instance().debugViewMode )
		processCommands();
	else
		TheRenderer::instance().renderDebugView();

	_overlays.resize( 0 );
}


void Pipeline::resize( int x, int y, int width, int height )
{
	TheRenderer::instance().resize( x, y, width, height );

	// Recreate render targets
	destroyRenderTargets();
	createRenderTargets();
}


void Pipeline::showOverlay( const Overlay &overlay, unsigned int matRes )
{
	_overlays.push_back( overlay );

	Resource *res = TheResourceManager::instance().resolveResHandle( matRes );
	if( res != 0x0 && res->getType() == ResourceTypes::Material )
		_overlays[_overlays.size() - 1].materialResource = (MaterialResource *)res;
}
