// *************************************************************************************************
//
// Horde3D
//   Next-Generation Graphics Engine
// --------------------------------------
// Copyright (C) 2006-2007 Nicolas Schulz
//
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//
// *************************************************************************************************

#include "egRenderer.h"
#include "utOpenGL.h"
#include "egSingleton.h"
#include "egGeometry.h"
#include "egModel.h"
#include "egParticle.h"
#include "egMaterial.h"
#include "egTextures.h"
#include "egShader.h"
#include "egLight.h"

#include "tinyxml.h"

#include "utDebug.h"


Renderer::Renderer() : RendererBase()
{
	_smFBO = 0; _smTex = 0;
	_defShadowMap = 0;
	_particleVBO = 0;
	_curLight = 0x0;
	_curMatRes = 0x0;
	_curShader = 0x0;
	_curUpdateStamp = 1;

	setActiveCamera( 0x0 );
}


Renderer::~Renderer()
{
	destroyShadowBuffer();
	unloadTexture( _defShadowMap, false );
	if( _particleVBO != 0 ) unloadBuffers( _particleVBO, 0 );
}


bool Renderer::init()
{
	bool failed = false;
	
	if( !RendererBase::init() ) failed = true;

	// Check that OpenGL 2.0 is available
	if( glExt::majorVersion < 2 || glExt::minorVersion < 0 )
	{
		TheEngineLog::instance().writeError( "OpenGL 2.0 not supported" );
		failed = true;
	}
	
	// Check extensions
	if( !glExt::EXT_framebuffer_object )
	{
		TheEngineLog::instance().writeError( "Extension EXT_framebuffer_object not supported" );
		failed = true;
	}
	if( !glExt::EXT_texture_filter_anisotropic )
	{
		TheEngineLog::instance().writeError( "Extension EXT_texture_filter_anisotropic not supported" );
		failed = true;
	}
	if( !glExt::EXT_texture_compression_s3tc )
	{
		TheEngineLog::instance().writeError( "Extension EXT_texture_compression_s3tc not supported" );
		failed = true;
	}
	if( !glExt::ARB_texture_float )
	{
		TheEngineLog::instance().writeWarning( "Extension ARB_texture_float not supported" );
	}
	if( !glExt::ARB_texture_non_power_of_two )
	{
		TheEngineLog::instance().writeWarning( "Extension ARB_texture_non_power_of_two not supported" );
	}

	if( failed )
	{
		TheEngineLog::instance().writeError( "Failed to init engine, debug info following:" );
		string s = (char *)glGetString( GL_VENDOR );
		TheEngineLog::instance().writeWarning( "GL_VENDOR:<br /> '<i>" + s + "</i>'" );
		s = (char *)glGetString( GL_RENDERER );
		TheEngineLog::instance().writeWarning( "GL_RENDERER:<br /> '<i>" + s + "</i>'" );
		s = (char *)glGetString( GL_VERSION );
		TheEngineLog::instance().writeWarning( "GL_VERSION:<br /> '<i>" + s + "</i>'" );
		s = (char *)glGetString( GL_EXTENSIONS );
		TheEngineLog::instance().writeWarning( "GL_EXTENSIONS:<br /> '<i>" + s + "</i>'" );

		return false;
	}

	// Init OpenGL states
	glHint( GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST );
	glDepthFunc( GL_LEQUAL );
	glShadeModel( GL_SMOOTH );
	glDisable( GL_MULTISAMPLE );
	glEnable( GL_DEPTH_TEST );
	glEnable( GL_CULL_FACE );
	glClearDepth( 1.0f );
	glClearColor( 0.0f, 0.0f, 0.0f, 1.0f );
	
	// Create shadow map
	if( !createShadowBuffer( TheEngineConfig::instance().shadowMapSize,
							 TheEngineConfig::instance().shadowMapSize ) )
	{
		TheEngineLog::instance().writeError( "Failed to create shadow map" );
		return false;
	}

	// Create default shadow map
	float shadowTex[] = { 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
						  1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
						  1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
						  1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1 };
	glGenTextures( 1, &_defShadowMap );
	glBindTexture( GL_TEXTURE_2D, _defShadowMap );
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST );
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST );
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE );
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE );
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_COMPARE_MODE, GL_COMPARE_R_TO_TEXTURE );
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_COMPARE_FUNC, GL_LEQUAL );
	glTexImage2D( GL_TEXTURE_2D, 0, GL_DEPTH_COMPONENT16, 8, 8, 0,
				  GL_DEPTH_COMPONENT, GL_FLOAT, shadowTex );

	// Create particle geometry array
	ParticleVert v0( 0, 0, 0 );
	ParticleVert v1( 1, 0, 1 );
	ParticleVert v2( 1, 1, 2 );
	ParticleVert v3( 0, 1, 3 );
	
	ParticleVert parVerts[ParticlesPerBatch * 4];
	for( unsigned int i = 0; i < ParticlesPerBatch; ++i )
	{
		parVerts[i * 4 + 0] = v0; parVerts[i * 4 + 0].index = (float)i;
		parVerts[i * 4 + 1] = v1; parVerts[i * 4 + 1].index = (float)i;
		parVerts[i * 4 + 2] = v2; parVerts[i * 4 + 2].index = (float)i;
		parVerts[i * 4 + 3] = v3; parVerts[i * 4 + 3].index = (float)i;

	}
	_particleVBO = TheRenderer::instance().uploadVertices(
		(float *)parVerts, ParticlesPerBatch * 4 * sizeof( ParticleVert ) );
	
	return true;
}


void Renderer::resize( int x, int y, int width, int height )
{
	RendererBase::resize( x, y, width, height );
}


unsigned int Renderer::getActiveCamera()
{
	if( _curCamera == 0x0 )
		return 0;
	else
		return _curCamera->_handle;
}


void Renderer::setActiveCamera( CameraNode *cam )
{
	if( cam != 0x0 )
		_curCamera = cam;
	else
		_curCamera = (CameraNode *)&TheSceneManager::instance().getDefCamNode();
}


unsigned int Renderer::loadShader( const char *vertexShader, const char *fragmentShader, ShaderContext &sc )
{
	unsigned int shaderId = RendererBase::loadShader( vertexShader, fragmentShader );
	if( shaderId == 0 ) return 0;

	setShader( shaderId );
	
	// Set standard uniforms
	sc.uni_texs[0] = glGetUniformLocation( shaderId, "tex0" );
	if( sc.uni_texs[0] >= 0 ) glUniform1i( sc.uni_texs[0], 0 );
	sc.uni_texs[1] = glGetUniformLocation( shaderId, "tex1" );
	if( sc.uni_texs[1] >= 0 ) glUniform1i( sc.uni_texs[1], 1 );
	sc.uni_texs[2] = glGetUniformLocation( shaderId, "tex2" );
	if( sc.uni_texs[2] >= 0 ) glUniform1i( sc.uni_texs[2], 2 );
	sc.uni_texs[3] = glGetUniformLocation( shaderId, "tex3" );
	if( sc.uni_texs[3] >= 0 ) glUniform1i( sc.uni_texs[3], 3 );
	sc.uni_texs[4] = glGetUniformLocation( shaderId, "tex4" );
	if( sc.uni_texs[4] >= 0 ) glUniform1i( sc.uni_texs[4], 4 );
	sc.uni_texs[5] = glGetUniformLocation( shaderId, "tex5" );
	if( sc.uni_texs[5] >= 0 ) glUniform1i( sc.uni_texs[5], 5 );
	sc.uni_texs[6] = glGetUniformLocation( shaderId, "tex6" );
	if( sc.uni_texs[6] >= 0 ) glUniform1i( sc.uni_texs[6], 6 );
	sc.uni_texs[7] = glGetUniformLocation( shaderId, "tex7" );
	if( sc.uni_texs[7] >= 0 ) glUniform1i( sc.uni_texs[7], 7 );

	int loc = glGetUniformLocation( shaderId, "buf0" );
	if( loc >= 0 ) glUniform1i( loc, 12 );
	loc = glGetUniformLocation( shaderId, "buf1" );
	if( loc >= 0 ) glUniform1i( loc, 13 );
	loc = glGetUniformLocation( shaderId, "buf2" );
	if( loc >= 0 ) glUniform1i( loc, 14 );
	loc = glGetUniformLocation( shaderId, "buf3" );
	if( loc >= 0 ) glUniform1i( loc, 15 );

	loc = glGetUniformLocation( shaderId, "shadowMap" );
	if( loc >= 0 ) glUniform1i( loc, 8 );

	// Get uniform locations
	sc.uni_worldMat = glGetUniformLocation( shaderId, "worldMat" );
	sc.uni_worldNormalMat = glGetUniformLocation( shaderId, "worldNormalMat" );
	sc.uni_viewer = glGetUniformLocation( shaderId, "viewer" );
	sc.uni_lightPos = glGetUniformLocation( shaderId, "lightPos" );
	sc.uni_lightDir = glGetUniformLocation( shaderId, "lightDir" );
	sc.uni_lightColor = glGetUniformLocation( shaderId, "lightColor" );
	sc.uni_lightCosCutoff = glGetUniformLocation( shaderId, "lightCosCutoff" );
	sc.uni_shadowSplitDists = glGetUniformLocation( shaderId, "shadowSplitDists" );
	sc.uni_shadowMats = glGetUniformLocation( shaderId, "shadowMats" );
	sc.uni_shadowMapSize = glGetUniformLocation( shaderId, "shadowMapSize" );
	sc.uni_shadowBias = glGetUniformLocation( shaderId, "shadowBias" );
	sc.uni_skinMatRows = glGetUniformLocation( shaderId, "skinMatRows[0]" );

	// Get attribute locations
	sc.attrib_texCoords0 = glGetAttribLocation( shaderId, "texCoords0" );
	sc.attrib_texCoords1 = glGetAttribLocation( shaderId, "texCoords1" );
	sc.attrib_normal = glGetAttribLocation( shaderId, "normal" );
	sc.attrib_tangent = glGetAttribLocation( shaderId, "tangent" );
	sc.attrib_bitangent = glGetAttribLocation( shaderId, "bitangent" );
	sc.attrib_bones = glGetAttribLocation( shaderId, "bones" );
	sc.attrib_weights = glGetAttribLocation( shaderId, "weights" );

	return shaderId;
}


void Renderer::setupViewMatrices( CameraNode *cam )
{
	glMatrixMode( GL_PROJECTION );
	glLoadIdentity();
	glFrustum( _curCamera->_frustLeft, _curCamera->_frustRight, _curCamera->_frustBottom,
			   _curCamera->_frustTop, _curCamera->_frustNear, _curCamera->_frustFar );
	glMatrixMode( GL_MODELVIEW );
	glLoadMatrixf( cam->_invTrans.x );
}


bool Renderer::setMaterial( MaterialResource *matRes, const string &shaderContext )
{
	if( matRes == 0x0 ) return false;
	if( matRes == _curMatRes ) return true;
		
	// Setup shader and render config
	if( matRes->_shaderRes != 0x0 )
	{
		ShaderContext *sc = matRes->_shaderRes->findContext( shaderContext );
		if( sc == 0x0 ) return false;
		
		if( sc != _curShader && sc->shaderObject != 0 )
		{
			_curShader = sc;
			setShader( _curShader->shaderObject );	
		}

		if( sc->writeDepth ) glDepthMask( GL_TRUE );
		else glDepthMask( GL_FALSE );

		if( sc->blendMode == BlendModes::Replace )
		{
			glDisable( GL_BLEND );
		}
		else if( sc->blendMode == BlendModes::Blend )
		{
			glEnable( GL_BLEND );
			glBlendFunc( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );
		}
		else if( sc->blendMode == BlendModes::Add )
		{
			glEnable( GL_BLEND );
			glBlendFunc( GL_ONE, GL_ONE );
		}
		else if( sc->blendMode == BlendModes::AddBlended )
		{
			glEnable( GL_BLEND );
			glBlendFunc( GL_SRC_ALPHA, GL_ONE );
		}
	}
	if( _curShader == 0x0 ) return false;

	_curMatRes = matRes;
	
	// Setup standard shader uniforms
	if( _curShader->lastUpdateStamp < _curUpdateStamp )
	{
		unsigned int shaderId = _curShader->shaderObject;
		
		if( _curCamera != 0x0 )	 // Viewer params
		{
			if( _curShader->uni_viewer >= 0 )
				glUniform3fv( _curShader->uni_viewer, 1, &_curCamera->_absTrans.x[12] );
		}
		if( _curLight != 0x0 )	// Light params
		{
			if( _curShader->uni_lightPos >= 0 )
				glUniform4f( _curShader->uni_lightPos, _curLight->_absPos.x, _curLight->_absPos.y,
							 _curLight->_absPos.z, _curLight->_radius );
			
			if( _curShader->uni_lightDir >= 0 )
				glUniform3fv( _curShader->uni_lightDir, 1, &_curLight->_spotDir.x );
			
			if( _curShader->uni_lightColor >= 0 )
				glUniform3f( _curShader->uni_lightColor, _curLight->_diffCol_R,
							 _curLight->_diffCol_G, _curLight->_diffCol_B );
			
			if( _curShader->uni_lightCosCutoff >= 0 )
				glUniform1f( _curShader->uni_lightCosCutoff, cosf( degToRad( _curLight->_fov/ 2 ) ) );
			
			if( _curShader->uni_shadowSplitDists >= 0 )
				glUniform4fv( _curShader->uni_shadowSplitDists, 1, &_splitPlanes[1] );

			if( _curShader->uni_shadowMats >= 0 )
				glUniformMatrix4fv( _curShader->uni_shadowMats, 4, false, &_lightMats[0].x[0] );
			
			if( _curShader->uni_shadowMapSize >= 0 )
				glUniform1f( _curShader->uni_shadowMapSize,
							 (float)TheEngineConfig::instance().shadowMapSize );
			
			if( _curShader->uni_shadowBias >= 0 )
				glUniform1f( _curShader->uni_shadowBias, _curLight->_shadowMapBias );
		}

		_curShader->lastUpdateStamp = _curUpdateStamp;
	}
	
	// Setup material parameters
	for( unsigned int i = 0; i < _curMatRes->_texUnits.size(); ++i )
	{
		TexUnit &texUnit = _curMatRes->_texUnits[i];
		
		if( texUnit.texRes != 0x0 && texUnit.unit < 8 && _curShader->uni_texs[texUnit.unit] >= 0 )
		{
			glActiveTexture( GL_TEXTURE0 + texUnit.unit );
			
			if( texUnit.texRes->getType() == ResourceTypes::Texture2D )
			{
				Texture2DResource *texRes = (Texture2DResource *)texUnit.texRes;
				glBindTexture( GL_TEXTURE_CUBE_MAP, 0 );
				glBindTexture( GL_TEXTURE_2D, texRes->getTexObject() );
			}
			else if( texUnit.texRes->getType() == ResourceTypes::TextureCube )
			{
				TextureCubeResource *texRes = (TextureCubeResource *)texUnit.texRes;
				glBindTexture( GL_TEXTURE_CUBE_MAP, texRes->getTexObject() );
			}
		}
	}

	glActiveTexture( GL_TEXTURE0 );

	// Setup material shader uniforms
	for( unsigned int i = 0; i < _curMatRes->_uniforms.size(); ++i )
	{
		Uniform &uniform = _curMatRes->_uniforms[i];

		int loc = glGetUniformLocation( _curShader->shaderObject, uniform.name.c_str() );
		if( loc >= 0 )
		{
			glUniform4f( loc, uniform.values[0], uniform.values[1],
						 uniform.values[2], uniform.values[3] );
		}
	}

	return true;
}


void Renderer::drawMeshes( const string &shaderContext, const string &theClass )
{
	ModelNode *curModelNode = 0x0;
	GeometryResource *curGeoRes = 0x0;
	bool morpherWasUsed = false;

	_curMatRes = 0x0;
	_curShader = 0x0;

	// Enables vertex arrays
	glEnableClientState( GL_VERTEX_ARRAY );

	// Loop through mesh queue
	for( unsigned int i = 0, s = TheSceneManager::instance().getMeshQueue().size(); i < s; ++i )
	{
		MeshNode *meshNode = (MeshNode *)TheSceneManager::instance().getMeshQueue()[i];
		bool skelChanged = false;

		// Bind geometry if the the current mesh belongs to another model than the previous one
		if( curModelNode != meshNode->getParentModel() )
		{
			bool geoChanged = false;
			
			curModelNode = meshNode->getParentModel();
			skelChanged = true;
			
			// Check if different geometry is used and bind attributes
			if( curGeoRes != &curModelNode->getGeometryResource() )
			{
				curGeoRes = &curModelNode->getGeometryResource();
				geoChanged = true;
				_curMatRes = 0x0;
				_curShader = 0x0;
			
				// Indices
				glBindBuffer( GL_ELEMENT_ARRAY_BUFFER, curGeoRes->getIndexBuffer() );
			}

			// Bind variable attributes if geometry changed
			if( geoChanged || (morpherWasUsed && !curModelNode->_morpherUsed) )
			{
				morpherWasUsed = false;
				
				// Position
				glBindBuffer( GL_ARRAY_BUFFER, curGeoRes->getVertBuffer() );
				glVertexPointer( 3, GL_FLOAT, sizeof( VertexData ), (char *)0 );
			}
			
			// Bind individual geometry if morphers are used
			if( curModelNode->_morpherUsed )
			{
				morpherWasUsed = true;
				glBindBuffer( GL_ARRAY_BUFFER, curModelNode->_morphVBO );
				glVertexPointer( 3, GL_FLOAT, sizeof( MorphVertex ), (char *)0 );
				glBindBuffer( GL_ARRAY_BUFFER, curGeoRes->getVertBuffer() );
			}
		}

		// Check that mesh is valid
		if( meshNode->getBatchStart() + meshNode->getBatchCount() > curGeoRes->_indices.size() )
		{	
			continue;
		}
		
		// *** Render mesh ***
		ShaderContext *prevShader = _curShader;
		if( !setMaterial( meshNode->getMaterialRes(), shaderContext ) ) continue;
		if( !meshNode->getMaterialRes()->isOfClass( theClass ) ) continue;

		// Set individual material uniforms
		for( unsigned int i = 0; i < meshNode->getUniforms().size(); ++i )
		{
			Uniform &uniform = meshNode->getUniforms()[i];

			int loc = glGetUniformLocation( _curShader->shaderObject, uniform.name.c_str() );
			if( loc >= 0 )
			{
				glUniform4f( loc, uniform.values[0], uniform.values[1],
							 uniform.values[2], uniform.values[3] );
			}
		}

		if( skelChanged )
		{
			// Skeleton
			if( _curShader->uni_skinMatRows >= 0 && curModelNode->_skinMatRows.size() > 0 )
			{
				glUniform4fv( _curShader->uni_skinMatRows,
					(unsigned int)curModelNode->_skinMatRows.size(),
					(float *)&curModelNode->_skinMatRows[0] );
			}
		}

		if( _curShader != prevShader )
		{
			// Texture coordinates
			if( _curShader->attrib_texCoords0 >= 0 )
			{
				glVertexAttribPointer( _curShader->attrib_texCoords0,
					2, GL_FLOAT, GL_FALSE, sizeof( VertexData ), (char *)0 + sizeof( float ) * 6 );
				glEnableVertexAttribArray( _curShader->attrib_texCoords0 );
			}
			if( _curShader->attrib_texCoords1 >= 0 )
			{
				glVertexAttribPointer( _curShader->attrib_texCoords1,
					2, GL_FLOAT, GL_FALSE, sizeof( VertexData ), (char *)0 + sizeof( float ) * 22 );
				glEnableVertexAttribArray( _curShader->attrib_texCoords1 );
			}
			
			// Tangent space basis
			if( _curShader->attrib_normal >= 0 || _curShader->attrib_tangent >= 0 ||
				_curShader->attrib_bitangent >= 0 )
			{
				unsigned int size = sizeof( VertexData );
				unsigned int normPtr = sizeof( float ) * 3;
				unsigned int tanPtr = sizeof( float ) * 16;
				unsigned int bitanPtr = sizeof( float ) * 19;
				
				if( curModelNode->_morpherUsed )
				{
					size = sizeof( MorphVertex );
					normPtr = sizeof( float ) * 3;
					tanPtr = sizeof( float ) * 6;
					bitanPtr = sizeof( float ) * 9;
					
					glBindBuffer( GL_ARRAY_BUFFER, curModelNode->_morphVBO );
				}
				
				// Normal
				if( _curShader->attrib_normal >= 0 )
				{
					glVertexAttribPointer( _curShader->attrib_normal,
						3, GL_FLOAT, GL_FALSE, size, (char *)0 + normPtr );
					glEnableVertexAttribArray( _curShader->attrib_normal );
				}
				
				// Tangent
				if( _curShader->attrib_tangent >= 0 )
				{
					glVertexAttribPointer( _curShader->attrib_tangent,
						3, GL_FLOAT, GL_FALSE, size, (char *)0 + tanPtr );
					glEnableVertexAttribArray( _curShader->attrib_tangent );
				}

				// Bitangent
				if( _curShader->attrib_bitangent >= 0 )
				{
					glVertexAttribPointer( _curShader->attrib_bitangent,
						3, GL_FLOAT, GL_FALSE, size, (char *)0 + bitanPtr );
					glEnableVertexAttribArray( _curShader->attrib_bitangent );
				}

				if( curModelNode->_morpherUsed )
				{
					glBindBuffer( GL_ARRAY_BUFFER, curGeoRes->getVertBuffer() );
				}
			}
			
			// Bone indices
			if( _curShader->attrib_bones >= 0 )
			{
				glVertexAttribPointer( _curShader->attrib_bones, 4, GL_FLOAT, GL_FALSE,
									   sizeof( VertexData ), (char *)0 + sizeof( float ) * 8 );
				glEnableVertexAttribArray( _curShader->attrib_bones );
			}

			// Weights
			if( _curShader->attrib_weights >= 0 )
			{
				glVertexAttribPointer( _curShader->attrib_weights, 4, GL_FLOAT, GL_FALSE,
									   sizeof( VertexData ), (char *)0 + sizeof( float ) * 12 );
				glEnableVertexAttribArray( _curShader->attrib_weights );
			}
		}

		// World transformation
		if( _curShader->uni_worldMat >= 0 )
		{
			glUniformMatrix4fv( _curShader->uni_worldMat, 1, false, &meshNode->_absTrans.x[0] );
		}
		if( _curShader->uni_worldNormalMat >= 0 )
		{
			// TODO: Optimize this
			Matrix4f normalMat4 = meshNode->_absTrans.inverted().transposed();
			float normalMat[9] = { normalMat4.x[0], normalMat4.x[1], normalMat4.x[2],
								   normalMat4.x[4], normalMat4.x[5], normalMat4.x[6],
								   normalMat4.x[8], normalMat4.x[9], normalMat4.x[10] };
			glUniformMatrix3fv( _curShader->uni_worldNormalMat, 1, false, normalMat );
		}

		// Render
		glDrawRangeElements( GL_TRIANGLES, meshNode->getVertRStart(), meshNode->getVertREnd(),
							 meshNode->getBatchCount(), 
							 curGeoRes->_16BitIndices ? GL_UNSIGNED_SHORT : GL_UNSIGNED_INT,
							 (char *)0 + meshNode->getBatchStart() *
							 (curGeoRes->_16BitIndices ? sizeof( short ) : sizeof( int )) );

		// Reset
		if( !meshNode->getUniforms().empty() )
		{
			_curMatRes = 0x0;
			_curShader = 0x0;
		}
	}
}


void Renderer::drawParticles( const string &shaderContext, const string &theClass )
{
	_curMatRes = 0x0;
	_curShader = 0x0;

	// Calculate right and up vectors for camera alignment
	float mat[16];
	glGetFloatv( GL_MODELVIEW_MATRIX, mat );
	Vec3f right = Vec3f( mat[0], mat[4], mat[8] );
	Vec3f up = Vec3f (mat[1], mat[5], mat[9] );
	Vec3f corners[4] = { -right - up, right - up, right + up, -right + up };

	// Bind particle geometry array
	glBindBuffer( GL_ARRAY_BUFFER, _particleVBO );
	glVertexPointer( 3, GL_FLOAT, sizeof( ParticleVert ), (char *)0 );
	glEnableClientState( GL_VERTEX_ARRAY );
	glTexCoordPointer( 2, GL_FLOAT, sizeof( ParticleVert ), (char *)0 + sizeof( float ) * 3 );
	glEnableClientState( GL_TEXTURE_COORD_ARRAY );

	// Loop through emitter queue
	for( unsigned int i = 0; i < TheSceneManager::instance().getEmitterQueue().size(); ++i )
	{
		EmitterNode *emitter = (EmitterNode *)TheSceneManager::instance().getEmitterQueue()[i];
		
		if( emitter->_particleCount == 0 ) continue;
		if( !setMaterial( emitter->_matRes, shaderContext ) ) continue;
		if( !emitter->_matRes->isOfClass( theClass ) ) continue;
		
		// Shader uniforms
		int loc = glGetUniformLocation( _curShader->shaderObject, "parCorners" );
		if( loc >= 0 ) glUniform3fv( loc, 4, (float *)corners );
		
		// Shader vertex attributes
		loc = glGetAttribLocation( _curShader->shaderObject, "parIdx" );
		if( loc >= 0 )
		{
			glVertexAttribPointer( loc, 1, GL_FLOAT, GL_FALSE, sizeof( ParticleVert ),
								   (char *)0 + sizeof( float ) * 6 );
			glEnableVertexAttribArray( loc );
		}
		loc = glGetAttribLocation( _curShader->shaderObject, "parCornerIdx" );
		if( loc >= 0 )
		{
			glVertexAttribPointer( loc, 1, GL_FLOAT, GL_FALSE, sizeof( ParticleVert ),
								   (char *)0 + sizeof( float ) * 5 );
			glEnableVertexAttribArray( loc );
		}

		// Divide particles in batches and render them
		for( unsigned int j = 0; j < emitter->_particleCount / ParticlesPerBatch; ++j )
		{
			// Check if batch needs to be rendered
			bool allDead = true;
			for( unsigned int k = 0; k < ParticlesPerBatch; ++k )
			{
				if( emitter->_particles[j*ParticlesPerBatch + k].life > 0 )
				{
					allDead = false;
					break;
				}
			}
			if( allDead ) continue;

			// Render batch
			loc = glGetUniformLocation( _curShader->shaderObject, "parPosArray" );
			if( loc >= 0 ) glUniform3fv( loc, ParticlesPerBatch, (float *)emitter->_parPositions + j*ParticlesPerBatch*3 );
			loc = glGetUniformLocation( _curShader->shaderObject, "parSizeAndRotArray" );
			if( loc >= 0 ) glUniform2fv( loc, ParticlesPerBatch, (float *)emitter->_parSizesANDRotations + j*ParticlesPerBatch*2 );
			loc = glGetUniformLocation( _curShader->shaderObject, "parColorArray" );
			if( loc >= 0 ) glUniform4fv( loc, ParticlesPerBatch, (float *)emitter->_parColors + j*ParticlesPerBatch*4 );

			glDrawArrays( GL_QUADS, 0, ParticlesPerBatch * 4 );
		}

		unsigned int count = emitter->_particleCount % ParticlesPerBatch;
		if( count > 0 )
		{
			unsigned int offset = (int)(emitter->_particleCount / ParticlesPerBatch) * ParticlesPerBatch;
			
			// Check if batch needs to be rendered
			bool allDead = true;
			for( unsigned int k = 0; k < count; ++k )
			{
				if( emitter->_particles[offset + k].life > 0 )
				{
					allDead = false;
					break;
				}
			}
			if( allDead ) continue;
			
			// Render batch
			loc = glGetUniformLocation( _curShader->shaderObject, "parPosArray" );
			if( loc >= 0 ) glUniform3fv( loc, count, (float *)emitter->_parPositions + offset*3 );
			loc = glGetUniformLocation( _curShader->shaderObject, "parSizeAndRotArray" );
			if( loc >= 0 ) glUniform2fv( loc, count, (float *)emitter->_parSizesANDRotations + offset*2 );
			loc = glGetUniformLocation( _curShader->shaderObject, "parColorArray" );
			if( loc >= 0 ) glUniform4fv( loc, count, (float *)emitter->_parColors + offset*4 );
			
			glDrawArrays( GL_QUADS, 0, count * 4 );
		}
	}
	
	glDisableClientState( GL_TEXTURE_COORD_ARRAY );
}


bool Renderer::createShadowBuffer( unsigned int width, unsigned int height )
{
	int curFBO;
	glGetIntegerv( GL_FRAMEBUFFER_BINDING_EXT, &curFBO );
	
	// Create framebuffer
	glGenFramebuffersEXT( 1, &_smFBO );
	glBindFramebufferEXT( GL_FRAMEBUFFER_EXT, _smFBO );
	glDrawBuffer( GL_NONE );
	glReadBuffer( GL_NONE );

	// Attach renderable textures
	glGenTextures( 1, &_smTex );
	glBindTexture( GL_TEXTURE_2D, _smTex );
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_BORDER );
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_BORDER );
	float borderColor[] = {1.0f, 1.0f, 1.0f, 1.0f};
	glTexParameterfv(GL_TEXTURE_2D, GL_TEXTURE_BORDER_COLOR, borderColor);
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_COMPARE_MODE, GL_COMPARE_R_TO_TEXTURE );
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_COMPARE_FUNC, GL_LEQUAL );
	glTexImage2D( GL_TEXTURE_2D, 0, GL_DEPTH_COMPONENT, width, height, 0, GL_DEPTH_COMPONENT, GL_UNSIGNED_BYTE, 0x0 );
	glFramebufferTexture2DEXT( GL_FRAMEBUFFER_EXT, GL_DEPTH_ATTACHMENT_EXT, GL_TEXTURE_2D, _smTex, 0 );

	int depthBits;
	glGetTexLevelParameteriv( GL_TEXTURE_2D, 0, GL_TEXTURE_DEPTH_SIZE, &depthBits );
	if( depthBits == 16 )
	{
		TheEngineLog::instance().writeWarning( "Shadow map precision is limited to 16 bit" );
	}

	// Check if successful
	unsigned int status = glCheckFramebufferStatusEXT( GL_FRAMEBUFFER_EXT );
	glBindFramebufferEXT( GL_FRAMEBUFFER_EXT, curFBO );
	
	if( status != GL_FRAMEBUFFER_COMPLETE_EXT ) return false;
	
	return true;
}


void Renderer::destroyShadowBuffer()
{
	if( _smTex != 0 ) glDeleteTextures( 1, &_smTex );
	if( _smFBO != 0 ) glDeleteFramebuffersEXT( 1, &_smFBO );

	_smTex = 0; _smFBO = 0;
}


Matrix4f Renderer::calcLightMat( const Frustum &frustum, bool allowCropping )
{
	// TODO: This routine still has problems with back projections when v1.w < 0
	
	float projMat[16];
	
	glMatrixMode( GL_PROJECTION );
	glLoadIdentity();
	myPerspective( _curLight->_fov, 1, _curCamera->_frustNear, _curLight->_radius );
	glGetFloatv( GL_PROJECTION_MATRIX, projMat );
	glMatrixMode( GL_MODELVIEW );

	if( !allowCropping ) return Matrix4f( projMat );

	Matrix4f lightMat = Matrix4f( projMat ) * _curLight->_invTrans;

	float minX =  Math::MaxFloat;
    float minY =  Math::MaxFloat;
	float minZ =  Math::MaxFloat;
	float maxX = -Math::MaxFloat;
    float maxY = -Math::MaxFloat;
	float maxZ = -Math::MaxFloat;
	
	for( unsigned int i = 0; i < 8; ++i )
	{
		Vec4f v1 = lightMat * Vec4f( frustum.getCorner( i ) );

		v1.x /= v1.w;
		v1.y /= v1.w;
		v1.z /= v1.w;

		if( v1.x < minX ) minX = v1.x;
		if( v1.y < minY ) minY = v1.y;
		if( v1.z < minZ ) minZ = v1.z;
		if( v1.x > maxX ) maxX = v1.x;
		if( v1.y > maxY ) maxY = v1.y;
		if( v1.z > maxZ ) maxZ = v1.z;
	}

	// Clamp the min and max values to post projection range [-1, 1]
	minX = clamp( minX, -1, 1 );
	minY = clamp( minY, -1, 1 );
	minZ = clamp( minZ, -1, 1 );
	maxX = clamp( maxX, -1, 1 );
	maxY = clamp( maxY, -1, 1 );
	maxZ = clamp( maxZ, -1, 1 );

	// Don't change near plane so that shadow casters don't get culled
	minZ = 0.0f;

	// Zoom in current split and create appropriate matrix
	Matrix4f cropView;
	float scaleX = 2.0f / (maxX - minX);
	float scaleY = 2.0f / (maxY - minY);
	float offsetX = -0.5f * (maxX + minX) * scaleX;
	float offsetY = -0.5f * (maxY + minY) * scaleY;
	float scaleZ = 1.0f / (maxZ - minZ);
	float offsetZ = -minZ * scaleZ;

	cropView.x[0] = scaleX;
	cropView.x[5] = scaleY;
	cropView.x[10] = scaleZ;
	cropView.x[12] = offsetX;
	cropView.x[13] = offsetY;
	cropView.x[14] = offsetZ;

	return cropView * Matrix4f( projMat );
}


void Renderer::updateShadowMap()
{
	if( _curLight == 0x0 || _curCamera == 0x0 ) return;
	
	int curFBO;
	glGetIntegerv( GL_FRAMEBUFFER_BINDING_EXT, &curFBO );
	glBindFramebufferEXT( GL_FRAMEBUFFER_EXT, _smFBO );
	glViewport( 0, 0, TheEngineConfig::instance().shadowMapSize, TheEngineConfig::instance().shadowMapSize );
	
	glDepthMask( GL_TRUE );
	glClearDepth( 1.0f );
    glClear( GL_DEPTH_BUFFER_BIT );

	// ********************************************************************************************
	// Cascaded Shadow Maps
	// ********************************************************************************************

	Frustum splitFrust;
	
	// Adjust camera planes
	// TODO: This can be improved to include solely lit and visible gemometry
	BoundingBox &bb = TheSceneManager::instance().getRootNode().getBBox();
	float minDist = Math::MaxFloat, maxDist = 0.0f;
	for( unsigned int i = 0; i < 8; ++i )
	{
		float dist = -(_curCamera->_invTrans * bb.getCorner( i )).z;
		if( dist < minDist ) minDist = dist;
		if( dist > maxDist ) maxDist = dist;
	}

	// Don't adjust near plane; this means less precision if scene is far away from viewer
	// but that shouldn't be too noticeable and brings better performance since the nearer
	// split volumes are empty
	minDist = _curCamera->_frustNear;
	
	// Caluclate split distances using PSSM scheme
	const float nearDist = maxf( minDist, _curCamera->_frustNear );
	const float farDist = maxf( maxDist, minDist + 1 );
	const unsigned int numMaps = _curLight->_shadowMapCount;
	const float t = _curLight->_shadowSplitLambda;
	
	_splitPlanes[0] = nearDist;
	_splitPlanes[numMaps] = farDist;

	for( unsigned int i = 1; i < numMaps; i++ )
	{
		float f = i / (float)numMaps;
		float log = nearDist * powf( farDist / nearDist, f );
		float uni = nearDist + (farDist - nearDist) * f;
		
		_splitPlanes[i] = t * log + (1 - t) * uni;
	}
	
	// Prepare shadow map rendering
	glEnable( GL_DEPTH_TEST );
	//glCullFace( GL_FRONT );	// Front face culling reduces artefacts but produces more "peter-panning"
	glMatrixMode( GL_MODELVIEW );
	glLoadMatrixf( _curLight->_invTrans.x );
	
	// Build split frustums and render shadow maps
	for( unsigned int i = 0; i < numMaps; ++i )
	{
		// Build split frustum
		float newLeft = _curCamera->_frustLeft * _splitPlanes[i] / _curCamera->_frustNear;
		float newRight = _curCamera->_frustRight * _splitPlanes[i] / _curCamera->_frustNear;
		float newBottom = _curCamera->_frustBottom * _splitPlanes[i] / _curCamera->_frustNear;
		float newTop = _curCamera->_frustTop * _splitPlanes[i] / _curCamera->_frustNear;
		splitFrust.buildViewFrustum( _curCamera->_absTrans, newLeft, newRight, newBottom, newTop,
									 _splitPlanes[i], _splitPlanes[i + 1] );
	
		// Build light projection matrix (no cropping for single map)
		_lightMats[i] = calcLightMat( splitFrust, numMaps > 1 );
		glMatrixMode( GL_PROJECTION );
		glLoadMatrixf( &_lightMats[i].x[0] );
		
		// Frustum Culling
		_lightFrustum.buildViewFrustum( _curLight->_invTrans, _lightMats[i] );
		TheSceneManager::instance().updateQueues( _lightFrustum, 0x0, true, true, false );
		
		_lightMats[i] = _lightMats[i] * _curLight->_invTrans;

		if( numMaps > 1 )
		{
			// Select quadrant of shadow map atlas
			const int sh = TheEngineConfig::instance().shadowMapSize / 2;
			if( i == 0 ) glViewport( 0, 0, sh, sh );
			else if( i == 1 ) glViewport( sh, 0, sh, sh );
			else if( i == 2 ) glViewport( sh, sh, sh, sh );
			else if( i == 3 ) glViewport( 0, sh, sh, sh );
		}
		else
		{
			const int sh = TheEngineConfig::instance().shadowMapSize;
			glViewport( 0, 0, sh, sh );
		}
	
		// Render
		drawMeshes( _curLight->_shadowContext, "" );
		drawParticles( _curLight->_shadowContext, "" );
	}

	// Setup light matrices for rendering:
	if( numMaps > 1 )
	{
		// Map from [-1,1] to [0,1] and select appropriate quadrant of shadow map
		_lightMats[0].scale( 0.25f, 0.25f, 0 );
		_lightMats[0].translate( 0.25f, 0.25f, 0 );
		_lightMats[1].scale( 0.25f, 0.25f, 0 );
		_lightMats[1].translate( 0.75f, 0.25f, 0 );
		_lightMats[2].scale( 0.25f, 0.25f, 0 );
		_lightMats[2].translate( 0.75f, 0.75f, 0 );
		_lightMats[3].scale( 0.25f, 0.25f, 0 );
		_lightMats[3].translate( 0.25f, 0.75f, 0 );
	}
	else
	{
		_lightMats[0].scale( 0.5f, 0.5f, 0 );
		_lightMats[0].translate( 0.5f, 0.5f, 0 );
	}

	// ********************************************************************************************

	glMatrixMode( GL_MODELVIEW );
	glCullFace( GL_BACK );
		
	glBindFramebufferEXT( GL_FRAMEBUFFER_EXT, curFBO );
	glViewport( _vpX, _vpY, _vpWidth, _vpHeight );
}


void Renderer::setupShadowMap()
{
	// Bind shadow map
	glActiveTexture( GL_TEXTURE8 );
	if( _curLight->_shadowMapCount > 0 )glBindTexture( GL_TEXTURE_2D, _smTex );
	else glBindTexture( GL_TEXTURE_2D, _defShadowMap );
	
	glActiveTexture( GL_TEXTURE0 );
}


void Renderer::bindBuffers( RenderBuffer *rb )
{
	if( rb == 0x0 )
	{
		glActiveTexture( GL_TEXTURE12 );
		glBindTexture( GL_TEXTURE_2D, 0 );
		glActiveTexture( GL_TEXTURE13 );
		glBindTexture( GL_TEXTURE_2D, 0 );
		glActiveTexture( GL_TEXTURE14 );
		glBindTexture( GL_TEXTURE_2D, 0 );
		glActiveTexture( GL_TEXTURE15 );
		glBindTexture( GL_TEXTURE_2D, 0 );
	}
	else
	{
		if( rb->texBufs[0] != 0 )
		{
			glActiveTexture( GL_TEXTURE12 );
			glBindTexture( GL_TEXTURE_2D, rb->texBufs[0] );
		}
		
		if( rb->texBufs[1] != 0 )
		{
			glActiveTexture( GL_TEXTURE13 );
			glBindTexture( GL_TEXTURE_2D, rb->texBufs[1] );
		}

		if( rb->texBufs[2] != 0 )
		{
			glActiveTexture( GL_TEXTURE14 );
			glBindTexture( GL_TEXTURE_2D, rb->texBufs[2] );
		}

		if( rb->texBufs[3] != 0 )
		{
			glActiveTexture( GL_TEXTURE15 );
			glBindTexture( GL_TEXTURE_2D, rb->texBufs[3] );
		}
	}

	glActiveTexture( GL_TEXTURE0 );
}


void Renderer::clear( bool depth, bool buf0, bool buf1, bool buf2, bool buf3,
					  float r, float g, float b, float a )
{
	int mask = 0;
	if( depth ) mask |= GL_DEPTH_BUFFER_BIT;
	if( buf0 ) mask |= GL_COLOR_BUFFER_BIT;

	glDisable( GL_BLEND );	// Clearing floating point buffers causes problems when blending is enabled on Radeon 9600
	glDepthMask( GL_TRUE );
	glClearColor( r, g, b, a );
	
	if( _curRendBuf == 0x0 )
	{
		glClear( mask );
	}
	else
	{
		if( buf0 || buf1 || buf2 || buf3 )
		{
			if( buf0 )
			{
				glDrawBuffer( GL_COLOR_ATTACHMENT0_EXT );
				glClear( mask );
				mask = GL_COLOR_BUFFER_BIT;
			}
			if( buf1 && _curRendBuf->texBufs[1] != 0 )
			{
				glDrawBuffer( GL_COLOR_ATTACHMENT1_EXT );
				glClear( mask );
				mask = GL_COLOR_BUFFER_BIT;
			}
			if( buf2 && _curRendBuf->texBufs[2] != 0 )
			{
				glDrawBuffer( GL_COLOR_ATTACHMENT2_EXT );
				glClear( mask );
				mask = GL_COLOR_BUFFER_BIT;
			}
			if( buf3 && _curRendBuf->texBufs[3] != 0 )
			{
				glDrawBuffer( GL_COLOR_ATTACHMENT3_EXT );
				glClear( mask );
				mask = GL_COLOR_BUFFER_BIT;
			}
		}
		else if( depth )
		{
			// Clear depth only
			glClear( GL_DEPTH_BUFFER_BIT );
		}
	}

	// Set draw buffers
	setRenderBuffer( _curRendBuf );
}


void Renderer::drawOverlays( const string &shaderContext, vector< Overlay > &overlays )
{
	_curMatRes = 0x0;
	_curShader = 0x0;
	
	glMatrixMode( GL_PROJECTION );
	glLoadIdentity();
	glOrtho( 0, 1, 0, 1, -1, 1 );
	glMatrixMode( GL_MODELVIEW );
	glLoadIdentity();
	
	glDisable( GL_DEPTH_TEST );
	
	for( int i = 0; i < 8; ++i )
	{
		for( unsigned int j = 0; j < overlays.size(); ++j )
		{
			if( overlays[j].layer != i ) continue;

			Overlay &overlay = overlays[j];

			if( !setMaterial( overlay.materialResource, shaderContext ) ) continue;
			
			glBegin(GL_QUADS);
			glTexCoord2fv( &overlay.u_ll ); glVertex2fv( &overlay.x_ll );
			glTexCoord2fv( &overlay.u_lr ); glVertex2fv( &overlay.x_lr );
			glTexCoord2fv( &overlay.u_ur ); glVertex2fv( &overlay.x_ur );
			glTexCoord2fv( &overlay.u_ul ); glVertex2fv( &overlay.x_ul );
			glEnd();
		}
	}

	glEnable( GL_DEPTH_TEST );
}


void Renderer::drawFSQuad( unsigned int materialResource, const string &shaderContext )
{
	_curMatRes = 0x0;
	_curShader = 0x0;

	Resource *res = TheResourceManager::instance().resolveResHandle( materialResource );
	if( !setMaterial( (MaterialResource *)res, shaderContext ) ) return;
	
	glMatrixMode( GL_PROJECTION );
	glLoadIdentity();
	glOrtho( 0, 1, 0, 1, -1, 1 );
	glMatrixMode( GL_MODELVIEW );
	glLoadIdentity();
	
	glDisable( GL_DEPTH_TEST );
	
	glBegin(GL_QUADS);
	glTexCoord2f( 0, 0 ); glVertex2f( 0, 0 );
	glTexCoord2f( 1, 0 ); glVertex2f( 1, 0 );
	glTexCoord2f( 1, 1 ); glVertex2f( 1, 1 );
	glTexCoord2f( 0, 1 ); glVertex2f( 0, 1 );
	glEnd();

	glEnable( GL_DEPTH_TEST );
}


void Renderer::drawGeometry( const string &shaderContext, const string &theClass )
{
	if( _curCamera == 0x0 ) return;
	
	_curCamera->genFrustum( _camFrustum );
	TheSceneManager::instance().updateQueues( _camFrustum, 0x0, true, true, false );
	
	setupViewMatrices( _curCamera );
	
	drawMeshes( shaderContext, theClass );
	drawParticles( shaderContext, theClass );
}


void Renderer::drawLightGeometry( const string &theClass )
{
	if( _curCamera == 0x0 ) return;
	
	_curCamera->genFrustum( _camFrustum );
	TheSceneManager::instance().updateQueues( _camFrustum, 0x0, false, false, true );
	
	for( unsigned int i = 0; i < TheSceneManager::instance().getLightQueue().size(); ++i )
	{
		++_curUpdateStamp;
		_curLight = (LightNode *)TheSceneManager::instance().getLightQueue()[i];
		_curLight->genFrustum( _lightFrustum );

		// Check if light is not visible
		if( _camFrustum.cullFrustum( _lightFrustum ) ) continue;
		
		TheSceneManager::instance().updateQueues( _camFrustum, &_lightFrustum, true, false, false );
	
		// Calculate light screen space position
		setupViewMatrices( _curCamera );
		float mat[16];
		glGetFloatv( GL_MODELVIEW_MATRIX, mat );
		Matrix4f m1( mat );
		glGetFloatv( GL_PROJECTION_MATRIX, mat );
		Matrix4f m2( mat );
		float bbx, bby, bbw, bbh;
		_curLight->calcScreenSpaceAABB( m2 * m1, bbx, bby, bbw, bbh );

		// Update shadow map
		if( _curLight->_shadowMapCount > 0 ) updateShadowMap();

		// Set scissor rectangle
		if( bbx != 0 || bby != 0 || bbw != 1 || bbh != 1 )
		{
			glScissor( (int)(bbx * _vpWidth), (int)(bby * _vpHeight),
					   (int)(bbw * _vpWidth), (int)(bbh * _vpHeight) );
			glEnable( GL_SCISSOR_TEST );
		}
		
		setupShadowMap();

		// Setup light attributes
		setMaterial( _curLight->_materialRes, _curLight->_lightingContext );
		
		// Render
		drawGeometry( _curLight->_lightingContext, theClass );

		// Reset
		glDisable( GL_SCISSOR_TEST );
		glActiveTexture( GL_TEXTURE8 );
		glBindTexture( GL_TEXTURE_2D, 0 );
		glActiveTexture( GL_TEXTURE0 );
	}

	_curLight = 0x0;
}


void Renderer::drawLightShapes()
{
	if( _curCamera == 0x0 ) return;
	
	_curCamera->genFrustum( _camFrustum );
	TheSceneManager::instance().updateQueues( _camFrustum, 0x0, false, false, true );
	
	for( unsigned int i = 0; i < TheSceneManager::instance().getLightQueue().size(); ++i )
	{
		++_curUpdateStamp;
		_curLight = (LightNode *)TheSceneManager::instance().getLightQueue()[i];
		_curLight->genFrustum( _lightFrustum );
		
		// Check if light is not visible
		if( _camFrustum.cullFrustum( _lightFrustum ) ) continue;
		
		// Calculate light screen space position
		setupViewMatrices( _curCamera );
		float mat[16];
		glGetFloatv( GL_MODELVIEW_MATRIX, mat );
		Matrix4f m1( mat );
		glGetFloatv( GL_PROJECTION_MATRIX, mat );
		Matrix4f m2( mat );
		float bbx, bby, bbw, bbh;
		_curLight->calcScreenSpaceAABB( m2 * m1, bbx, bby, bbw, bbh );
		
		// Update shadow map
		if( _curLight->_shadowMapCount > 0 ) updateShadowMap();

		// Prepare postprocessing step (set the camera transformation in MV matrix)
		glDisable( GL_DEPTH_TEST );
		glMatrixMode( GL_PROJECTION );
		glLoadIdentity();
		glOrtho( 0, 1, 0, 1, -1, 1 );
		glMatrixMode( GL_MODELVIEW );
		glLoadMatrixf( _curCamera->_invTrans.x );
		
		setupShadowMap();

		_curMatRes = 0x0;
		_curShader = 0x0;
		if( !setMaterial( _curLight->_materialRes, _curLight->_lightingContext ) ) continue;
		
		// Draw quad
		glBegin( GL_QUADS );
		glTexCoord2f( bbx, bby ); glVertex2f( bbx, bby );
		glTexCoord2f( bbx + bbw, bby ); glVertex2f( bbx + bbw, bby );
		glTexCoord2f( bbx + bbw, bby + bbh ); glVertex2f( bbx + bbw, bby + bbh );
		glTexCoord2f( bbx, bby + bbh ); glVertex2f( bbx, bby + bbh );
		glEnd();

		// Reset
		glEnable( GL_DEPTH_TEST );
		glActiveTexture( GL_TEXTURE8 );
		glBindTexture( GL_TEXTURE_2D, 0 );
		glActiveTexture( GL_TEXTURE0 );
	}

	_curLight = 0x0;
}


void Renderer::renderDebugView()
{
	if( _curCamera == 0x0 ) return;
	
	setRenderBuffer( 0x0 );
	setShader( 0 );
	glColor4f( 0.5f, 0.75f, 1, 1 );
	glDepthMask( GL_TRUE );
	glPolygonMode( GL_FRONT_AND_BACK, GL_LINE );
	for( unsigned int i = 0; i < 16; ++i )
	{
		glActiveTexture( GL_TEXTURE0 + i );
		glDisable( GL_TEXTURE_2D );
		glDisable( GL_TEXTURE_CUBE_MAP );
	}

	glClearColor( 0, 0, 0, 1 );
	glClear( GL_DEPTH_BUFFER_BIT | GL_COLOR_BUFFER_BIT );

	// Geometry
	_curCamera->genFrustum( _camFrustum );
	TheSceneManager::instance().updateQueues( _camFrustum, 0x0, true, false, true );
	setupViewMatrices( _curCamera );
	
	for( unsigned int i = 0, s = TheSceneManager::instance().getMeshQueue().size(); i < s; ++i )
	{
		MeshNode *meshNode = (MeshNode *)TheSceneManager::instance().getMeshQueue()[i];
		GeometryResource &geoRes = meshNode->getParentModel()->getGeometryResource();

		if( meshNode->getBatchStart() + meshNode->getBatchCount() > geoRes._indices.size() )
		{	
			continue;
		}
		
		glBindBuffer( GL_ELEMENT_ARRAY_BUFFER, geoRes.getIndexBuffer() );
		glBindBuffer( GL_ARRAY_BUFFER, geoRes.getVertBuffer() );
		glVertexPointer( 3, GL_FLOAT, sizeof( VertexData ), (char *)0 );
		glEnableClientState( GL_VERTEX_ARRAY );
		glDisableClientState( GL_TEXTURE_COORD_ARRAY );
		
		glPushMatrix();
		glMultMatrixf( &meshNode->_absTrans.x[0] );
		glDrawRangeElements( GL_TRIANGLES, meshNode->getVertRStart(), meshNode->getVertREnd(),
							 meshNode->getBatchCount(), 
							 geoRes._16BitIndices ? GL_UNSIGNED_SHORT : GL_UNSIGNED_INT,
							 (char *)0 + meshNode->getBatchStart() *
							 (geoRes._16BitIndices ? sizeof( short ) : sizeof( int )) );
		glPopMatrix();
	}
	
	// Light sources in screen space
	glPolygonMode( GL_FRONT_AND_BACK, GL_FILL );
	glDisable( GL_DEPTH_TEST );
	glEnable( GL_BLEND );
	glBlendFunc( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );
	glColor4f( 1, 1, 1, 0.25f );
	
	for( unsigned int i = 0; i < TheSceneManager::instance().getLightQueue().size(); ++i )
	{
		LightNode *lightNode = (LightNode *)TheSceneManager::instance().getLightQueue()[i];
		lightNode->genFrustum( _lightFrustum );
		
		if( _camFrustum.cullFrustum( _lightFrustum ) ) continue;

		// Calculate light screen space position
		setupViewMatrices( _curCamera );
		float mat[16];
		glGetFloatv( GL_MODELVIEW_MATRIX, mat );
		Matrix4f m1( mat );
		glGetFloatv( GL_PROJECTION_MATRIX, mat );
		Matrix4f m2( mat );
		float bbx, bby, bbw, bbh;
		lightNode->calcScreenSpaceAABB( m2 * m1, bbx, bby, bbw, bbh );

		glMatrixMode( GL_PROJECTION );
		glLoadIdentity();
		glOrtho( 0, 1, 0, 1, -1, 1 );
		glMatrixMode( GL_MODELVIEW );
		glLoadIdentity();
		
		glBegin( GL_QUADS );
		glVertex2f( bbx, bby );
		glVertex2f( bbx + bbw, bby );
		glVertex2f( bbx + bbw, bby + bbh );
		glVertex2f( bbx, bby + bbh );
		glEnd();
	}

	glEnable( GL_DEPTH_TEST );
	glDisable( GL_BLEND );
}
