// *************************************************************************************************
//
// Horde3D
//   Next-Generation Graphics Engine
// --------------------------------------
// Copyright (C) 2006-2007 Nicolas Schulz
//
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//
// *************************************************************************************************

#include "egRendererBase.h"
#include "egSingleton.h"
#include "utOpenGL.h"

#include "utDebug.h"


RendererBase::RendererBase()
{
	_initalized = false;
	_vpWidth = 320; _vpHeight = 240;
	_curRendBuf = 0x0;
}


bool RendererBase::init()
{
	_initalized = true;
	
	return initOpenGLExtensions();
}


void RendererBase::resize( int x, int y, int width, int height )
{
	_vpX = x; _vpY = y; _vpWidth = width; _vpHeight = height;
}


void RendererBase::myPerspective( float fovy, float aspect, float zNear, float zFar ) 
{    
	double ymax = zNear * tan( degToRad( fovy / 2 ) );
	double xmax = ymax * aspect;
	glFrustum( -xmax, xmax, -ymax, ymax, zNear, zFar );
}


unsigned int RendererBase::uploadVertices( void *data, unsigned int size, unsigned int bufId )
{
	// Note:
	// Interleaved arrays are generally faster than sperated arrays
	// Vertex size should be a multiple of 32 bytes (gives good performance increase)
	
	if( bufId == 0 )
	{	
		glGenBuffers( 1, &bufId );
		glBindBuffer( GL_ARRAY_BUFFER, bufId );
		glBufferData( GL_ARRAY_BUFFER, size, data, GL_DYNAMIC_DRAW );
	}
	else
	{
		glBindBuffer( GL_ARRAY_BUFFER, bufId );
		//glBufferSubData( GL_ARRAY_BUFFER, 0, numVerts * sizeof(float) * 16, data );
		glBufferData( GL_ARRAY_BUFFER, size, data, GL_DYNAMIC_DRAW );
	}
	
	return bufId;
}


unsigned int RendererBase::uploadIndices( void *indices, unsigned int size, unsigned int bufId )
{
	if( bufId == 0 ) 
	{	
		glGenBuffers( 1, &bufId );
		glBindBuffer( GL_ELEMENT_ARRAY_BUFFER, bufId );
		glBufferData( GL_ELEMENT_ARRAY_BUFFER, size, indices, GL_STATIC_DRAW );
	}
	else
	{
		glBindBuffer( GL_ELEMENT_ARRAY_BUFFER, bufId );
		//glBufferSubData( GL_ELEMENT_ARRAY_BUFFER, 0, size, indices );
		glBufferData( GL_ELEMENT_ARRAY_BUFFER, size, indices, GL_STATIC_DRAW );
	}
	
	return bufId;
}


void RendererBase::unloadBuffers( unsigned int vertBufId, unsigned int idxBufId )
{
	if( vertBufId != 0 )
	{
		glBindBuffer( GL_ARRAY_BUFFER, vertBufId );
		glBufferData( GL_ARRAY_BUFFER, 0, 0x0, GL_STATIC_DRAW );
		glBindBuffer( GL_ARRAY_BUFFER, 0 );
		
		// TODO:	NVidia driver sometimes crashes when deleting array buffer object
		//			and creating new one
		//glDeleteBuffers( 1, &vertBufId );
	}

	if( idxBufId != 0 )
	{
		glBindBuffer( GL_ELEMENT_ARRAY_BUFFER, idxBufId );
		glBufferData( GL_ELEMENT_ARRAY_BUFFER, 0, 0x0, GL_STATIC_DRAW );
		glBindBuffer( GL_ELEMENT_ARRAY_BUFFER, 0 );
		glDeleteBuffers( 1, &idxBufId );
	}
}


bool RendererBase::supportsNPOTTextures()
{
	return glExt::ARB_texture_non_power_of_two;
}


unsigned int RendererBase::uploadTexture2D( unsigned char *buffer, int width, int height, 
											int bpp, bool allowCompression, unsigned int texId )
{
	if( texId == 0 ) glGenTextures( 1, &texId );
	glBindTexture( GL_TEXTURE_2D, texId );
	
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT );
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT );
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAX_ANISOTROPY_EXT, TheEngineConfig::instance().anisotropyFactor );
	
	glTexParameteri( GL_TEXTURE_2D, GL_GENERATE_MIPMAP, GL_TRUE );
		
	if( TheEngineConfig::instance().trilinearFiltering )
		glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR );
	else
		glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_NEAREST );
	
	if( TheEngineConfig::instance().texCompression && allowCompression )
	{
		// Texture compression
		if (bpp == 24)
			glTexImage2D( GL_TEXTURE_2D, 0, GL_COMPRESSED_RGB_S3TC_DXT1_EXT, width, height, 0, GL_BGR, GL_UNSIGNED_BYTE, buffer );
		else if (bpp == 32)
			glTexImage2D( GL_TEXTURE_2D, 0, GL_COMPRESSED_RGBA_S3TC_DXT5_EXT, width, height, 0, GL_BGRA, GL_UNSIGNED_BYTE, buffer );
	}
	else
	{
		// No texture compression
		if( bpp == 24 )
			glTexImage2D( GL_TEXTURE_2D, 0, GL_RGB8, width, height, 0, GL_BGR, GL_UNSIGNED_BYTE, buffer );
		else if( bpp == 32 )
			glTexImage2D( GL_TEXTURE_2D, 0, GL_RGBA8, width, height, 0, GL_BGRA, GL_UNSIGNED_BYTE, buffer );
	}
	
	return texId;
}


void RendererBase::updateTexture2D( unsigned char *buffer, int width, int height,
								    int bpp, unsigned int texId )
{
	glBindTexture( GL_TEXTURE_2D, texId );

	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
	glTexParameteri( GL_TEXTURE_2D, GL_GENERATE_MIPMAP, GL_FALSE );
	
	if( bpp == 24 ) 
		glTexSubImage2D( GL_TEXTURE_2D, 0, 0, 0, width, height, GL_BGR, GL_UNSIGNED_BYTE, buffer );
	else if( bpp == 32 )
		glTexSubImage2D( GL_TEXTURE_2D, 0, 0, 0, width, height, GL_BGRA, GL_UNSIGNED_BYTE, buffer );
}


unsigned int RendererBase::uploadTextureCube( unsigned char *buffer, int width, int height, int bpp,
											  unsigned int cubeFace, bool allowCompression, unsigned int texId )
{
	if( texId == 0 ) glGenTextures( 1, &texId );
	glBindTexture( GL_TEXTURE_CUBE_MAP, texId );
	
	glTexParameteri( GL_TEXTURE_CUBE_MAP, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
	glTexParameteri( GL_TEXTURE_CUBE_MAP, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE );
	glTexParameteri( GL_TEXTURE_CUBE_MAP, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE );
	glTexParameteri( GL_TEXTURE_CUBE_MAP, GL_TEXTURE_WRAP_R, GL_CLAMP_TO_EDGE );
	glTexParameteri( GL_TEXTURE_CUBE_MAP, GL_TEXTURE_MAX_ANISOTROPY_EXT, TheEngineConfig::instance().anisotropyFactor );
	
	glTexParameteri( GL_TEXTURE_CUBE_MAP, GL_GENERATE_MIPMAP, GL_TRUE );
		
	if( TheEngineConfig::instance().trilinearFiltering )
		glTexParameteri( GL_TEXTURE_CUBE_MAP, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR );
	else
		glTexParameteri( GL_TEXTURE_CUBE_MAP, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_NEAREST );
	
	if( TheEngineConfig::instance().texCompression && allowCompression )
	{
		// Texture compression
		if (bpp == 24)
			glTexImage2D( GL_TEXTURE_CUBE_MAP_POSITIVE_X + cubeFace, 0, GL_COMPRESSED_RGB_S3TC_DXT1_EXT, width, height, 0, GL_BGR, GL_UNSIGNED_BYTE, buffer );
		else if (bpp == 32)
			glTexImage2D( GL_TEXTURE_CUBE_MAP_POSITIVE_X + cubeFace, 0, GL_COMPRESSED_RGBA_S3TC_DXT5_EXT, width, height, 0, GL_BGRA, GL_UNSIGNED_BYTE, buffer );
	}
	else
	{
		// No texture compression
		if( bpp == 24 )
			glTexImage2D( GL_TEXTURE_CUBE_MAP_POSITIVE_X + cubeFace, 0, GL_RGB8, width, height, 0, GL_BGR, GL_UNSIGNED_BYTE, buffer );
		else if( bpp == 32 )
			glTexImage2D( GL_TEXTURE_CUBE_MAP_POSITIVE_X + cubeFace, 0, GL_RGBA8, width, height, 0, GL_BGRA, GL_UNSIGNED_BYTE, buffer );
	}
	
	return texId;
}


void RendererBase::unloadTexture( unsigned int texId, bool cubeMap )
{
	if( !cubeMap )
	{	
		glBindTexture( GL_TEXTURE_2D, texId );
		glTexImage2D( GL_TEXTURE_2D, 0, GL_RGB8, 0, 0, 0, GL_BGR, GL_UNSIGNED_BYTE, 0x0 );
	}
	else 
	{	
		glBindTexture( GL_TEXTURE_CUBE_MAP, texId );
		glTexImage2D( GL_TEXTURE_CUBE_MAP_POSITIVE_X, 0, GL_RGB8, 0, 0, 0, GL_BGR, GL_UNSIGNED_BYTE, 0x0 );
		glTexImage2D( GL_TEXTURE_CUBE_MAP_NEGATIVE_X, 0, GL_RGB8, 0, 0, 0, GL_BGR, GL_UNSIGNED_BYTE, 0x0 );
		glTexImage2D( GL_TEXTURE_CUBE_MAP_POSITIVE_Y, 0, GL_RGB8, 0, 0, 0, GL_BGR, GL_UNSIGNED_BYTE, 0x0 );
		glTexImage2D( GL_TEXTURE_CUBE_MAP_NEGATIVE_Y, 0, GL_RGB8, 0, 0, 0, GL_BGR, GL_UNSIGNED_BYTE, 0x0 );
		glTexImage2D( GL_TEXTURE_CUBE_MAP_POSITIVE_Z, 0, GL_RGB8, 0, 0, 0, GL_BGR, GL_UNSIGNED_BYTE, 0x0 );
		glTexImage2D( GL_TEXTURE_CUBE_MAP_NEGATIVE_Z, 0, GL_RGB8, 0, 0, 0, GL_BGR, GL_UNSIGNED_BYTE, 0x0 );
	}
	
	glDeleteTextures( 1, &texId );
}


unsigned int RendererBase::loadShader( const char *vertexShader, const char *fragmentShader )
{
	int infologLength = 0;
	int charsWritten = 0;
	char *infoLog = 0x0;
	int status;

	_shaderLog = "";

	// Vertex shader
	unsigned int vs = glCreateShader( GL_VERTEX_SHADER );
	glShaderSource( vs, 1, &vertexShader, 0x0 );
	glCompileShader( vs );
	glGetShaderiv( vs, GL_COMPILE_STATUS, &status );
	if( !status )
	{	
		// Get info
		glGetShaderiv( vs, GL_INFO_LOG_LENGTH, &infologLength );
		if( infologLength > 1 )
		{
			infoLog = new char[infologLength];
			glGetShaderInfoLog( vs, infologLength, &charsWritten, infoLog );
			_shaderLog = _shaderLog + "[Vertex Shader]<br /><pre>" + infoLog + "</pre><br />";
			delete[] infoLog; infoLog = 0x0;
		}

		glDeleteShader( vs );
		return 0;
	}

	// Fragment shader
	unsigned int fs = glCreateShader( GL_FRAGMENT_SHADER );
	glShaderSource( fs, 1, &fragmentShader, 0x0 );
	glCompileShader( fs );
	glGetShaderiv( fs, GL_COMPILE_STATUS, &status );
	if( !status )
	{	
		glGetShaderiv( fs, GL_INFO_LOG_LENGTH, &infologLength );
		if( infologLength > 1 )
		{
			infoLog = new char[infologLength];
			glGetShaderInfoLog( fs, infologLength, &charsWritten, infoLog );
			_shaderLog = _shaderLog + "[Fragment Shader]<br /><pre>" + infoLog + "</pre><br />";
			delete[] infoLog; infoLog = 0x0;
		}

		glDeleteShader( vs );
		glDeleteShader( fs );
		return 0;
	}

	// Shader program
	unsigned int program = glCreateProgram();
	glAttachShader( program, vs );
	glAttachShader( program, fs );
	glDeleteShader( vs );
	glDeleteShader( fs );
	glLinkProgram( program );
	glGetProgramiv( program, GL_INFO_LOG_LENGTH, &infologLength );
	if( infologLength > 1 )
	{
		infoLog = new char[infologLength];
		glGetProgramInfoLog( program, infologLength, &charsWritten, infoLog );
		_shaderLog = _shaderLog + "[Linking]<br /><pre>" + infoLog + "</pre><br />";
		delete[] infoLog; infoLog = 0x0;
	}
	glGetProgramiv( program, GL_LINK_STATUS, &status );
	if( !status )
	{	
		glDeleteProgram( program );
		return 0;
	}

	return program;
}


void RendererBase::unloadShader( unsigned int shaderId )
{
	// TODO:	ATI driver sometimes crashes when deleting shader object
	//			and creating new one
	//glDeleteProgram( shaderId );
}


void RendererBase::setShader( unsigned int shaderId )
{
	glUseProgram( shaderId );	
}


RenderBuffer RendererBase::createRenderBuffer( unsigned int width, unsigned int height,
											   RenderBufferFormats::List format,
											   bool depth, unsigned int numTexBuffers )
{
	if( (format == RenderBufferFormats::RGBA16F || format == RenderBufferFormats::RGBA32F) &&
		!glExt::ARB_texture_float )
	{
		return RenderBuffer();
	}
	
	RenderBuffer rb;
	
	rb.width = width; rb.height = height;
	
	// Create framebuffer
	glGenFramebuffersEXT( 1, &rb.fbo );
	glBindFramebufferEXT( GL_FRAMEBUFFER_EXT, rb.fbo );

	// Attach renderable textures
	unsigned int glFormat = 0;
	if( format == RenderBufferFormats::RGBA8 ) glFormat = GL_RGBA8;
	else if( format == RenderBufferFormats::RGBA16F ) glFormat = GL_RGBA16F_ARB;
	else if( format == RenderBufferFormats::RGBA32F ) glFormat = GL_RGBA32F_ARB;

	for( unsigned int j = 0; j < numTexBuffers; ++j )
	{
		glGenTextures( 1, &rb.texBufs[j] );
		glBindTexture( GL_TEXTURE_2D, rb.texBufs[j] );
		glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST );
		glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST );
		glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE );
		glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE );
		glTexImage2D( GL_TEXTURE_2D, 0, glFormat, rb.width, rb.height, 0, GL_RGBA, GL_FLOAT, 0x0 );
		glFramebufferTexture2DEXT( GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT + j, GL_TEXTURE_2D, rb.texBufs[j], 0 );
	}

	// Attach depth buffer
	if( depth )
	{
		glGenRenderbuffersEXT( 1, &rb.depthBuf );
		glBindRenderbufferEXT( GL_RENDERBUFFER_EXT, rb.depthBuf );
		glRenderbufferStorageEXT( GL_RENDERBUFFER_EXT, GL_DEPTH_COMPONENT24, rb.width, rb.height );
		glFramebufferRenderbufferEXT( GL_FRAMEBUFFER_EXT, GL_DEPTH_ATTACHMENT_EXT, GL_RENDERBUFFER_EXT, rb.depthBuf );
	}

	// Check if successful
	unsigned int status = glCheckFramebufferStatusEXT( GL_FRAMEBUFFER_EXT );
	glBindFramebufferEXT( GL_FRAMEBUFFER_EXT, 0 );
	
	if( status != GL_FRAMEBUFFER_COMPLETE_EXT )
	{	
		return RenderBuffer();
	}

	return rb;
}


void RendererBase::destroyRenderBuffer( RenderBuffer &rb )
{
	glBindFramebufferEXT( GL_FRAMEBUFFER_EXT, 0 );
	
	if( rb.depthBuf != 0 )
	{
		glDeleteRenderbuffersEXT( 1, &rb.depthBuf );
		rb.depthBuf = 0;
	}
		
	for( unsigned int i = 0; i < 4; ++i )
	{
		if( rb.texBufs[i] != 0 )glDeleteTextures( 1, &rb.texBufs[i] );
		rb.texBufs[i] = 0;
	}

	if( rb.fbo != 0 )
	{
		glDeleteFramebuffersEXT( 1, &rb.fbo );
		rb.fbo = 0;
	}
}


void RendererBase::setRenderBuffer( RenderBuffer *rb )
{
	_curRendBuf = rb;
	
	if( rb == 0x0 || rb->fbo == 0 )
	{
		glBindFramebufferEXT( GL_FRAMEBUFFER_EXT, 0 );
		glDrawBuffer( GL_BACK_LEFT );
		glViewport( _vpX, _vpY, _vpWidth, _vpHeight );
	}
	else
	{
		unsigned int buffers[] = {	GL_COLOR_ATTACHMENT0_EXT, GL_COLOR_ATTACHMENT1_EXT,
									GL_COLOR_ATTACHMENT2_EXT, GL_COLOR_ATTACHMENT3_EXT };
		
		// Count buffers
		unsigned int numBuffers = 0;
		for( unsigned int i = 0; i < 4; ++i )
		{
			if( rb->texBufs[i] == 0 ) break;
			++numBuffers;
		}
		
		if( numBuffers > 0 )
		{
			glBindFramebufferEXT( GL_FRAMEBUFFER_EXT, rb->fbo );
			glDrawBuffers( numBuffers, buffers );
			glViewport( 0, 0, rb->width, rb->height );
		}
	}
}
