// *************************************************************************************************
//
// Horde3D
//   Next-Generation Graphics Engine
// --------------------------------------
// Copyright (C) 2006-2007 Nicolas Schulz
//
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//
// *************************************************************************************************

#include "egResource.h"
#include "egSingleton.h"
#include "egGeometry.h"
#include "egSceneGraphRes.h"
#include "egTextures.h"
#include "egMaterial.h"
#include "egAnimation.h"
#include "egShader.h"

#include "utDebug.h"


// **********************************************************************************
// Class Resource
// **********************************************************************************

Resource::Resource( ResourceTypes::List type, const string &name, unsigned int flags )
{
	_type = type;
	_name = name;
	_handle = 0;
	_loaded = false;
	_refCount = 0;
	_userRef = false;

	if( (flags & ResourceFlags::NoQuery) == ResourceFlags::NoQuery ) _noQuery = true;
	else _noQuery = false;
}


Resource::~Resource()
{
	// Remove all references
	// Nothing  to to here
}


void Resource::initDefault()
{
}


void Resource::release()
{
}


bool Resource::load( const char *data, unsigned int size )
{	
	// Resources can only be loaded once
	if( _loaded ) return false;
	
	// A NULL pointer can be used if the file could not be loaded
	if( data == 0x0 || size == 0 )
	{	
		TheEngineLog::instance().writeWarning( 
			string( "No data loaded (file not found?) for resource " ) + _name );
		_noQuery = true;
		return false;
	}

	_loaded = true;
	
	return true;
}


void Resource::unload()
{
	release();
	initDefault();
	_loaded = false;
}


const void *Resource::getData( ResourceData::List param )
{
	return 0x0;
}


bool Resource::updateData( ResourceData::List param, const void *data, unsigned int size )
{
	return false;
}


// **********************************************************************************
// Class ResourceManager
// **********************************************************************************

ResourceManager::ResourceManager()
{
	_resources.reserve( 100 );
	_resourcePaths.resize( ResourceTypes::Count, "" );
}


ResourceManager::~ResourceManager()
{
	clear();

	// Release resource types
	SceneGraphResource::releaseResType();
	GeometryResource::releaseResType();
	AnimationResource::releaseResType();
	MaterialResource::releaseResType();
	CodeResource::releaseResType();
	ShaderResource::releaseResType();
	Texture2DResource::releaseResType();
	TextureCubeResource::releaseResType();
	EffectResource::releaseResType();
}


bool ResourceManager::init()
{
	// Init resource types
	// Warning: This must usually be called after the renderer was initalized
	SceneGraphResource::initResType();
	GeometryResource::initResType();
	AnimationResource::initResType();
	MaterialResource::initResType();
	CodeResource::initResType();
	ShaderResource::initResType();
	Texture2DResource::initResType();
	TextureCubeResource::initResType();
	EffectResource::initResType();

	return true;
}


const string ResourceManager::getResourcePath( ResourceTypes::List type )
{
	if( type < ResourceTypes::Count )
		return _resourcePaths[type];
	else
		return "";
}


void ResourceManager::setResourcePath( ResourceTypes::List type, string path )
{
	if( type < ResourceTypes::Count )
	{
		if( path != "" && path[path.size() - 1] != '/' && path[path.size() - 1] != '\\' )
			path = path + "/";
	
		_resourcePaths[type] = path;
	}
}


Resource *ResourceManager::findResource( ResourceTypes::List type, const string &name )
{
	string fName = _resourcePaths[type] + name;
	
	for( unsigned int i = 0; i < _resources.size(); ++i )
	{
		if( _resources[i] == 0x0 ) continue;

		if( fName == _resources[i]->_name )
		{
			return _resources[i];
		}
	}
	
	return 0x0;
}


unsigned int ResourceManager::addResource( ResourceTypes::List type, const string &name,
										   unsigned int flags, bool userCall )
{
	if( name == "" )
	{	
		TheEngineLog::instance().writeError( "Missing name for added resource" );
		return 0;
	}
	
	
	string fName = _resourcePaths[type] + name;
	
	// Check if resource is already in list and return index
	for( unsigned int i = 0; i < _resources.size(); ++i )
	{
		if( _resources[i] == 0x0 ) continue;

		if( fName == _resources[i]->_name )
		{
			if( _resources[i]->_type == type )
			{
				if( userCall ) _resources[i]->_userRef = true;
				else ++_resources[i]->_refCount;
				return i + 1;
			}
			else
			{
				TheEngineLog::instance().writeError(
					"There is another resource with same name but different type" );
				return 0;
			}
		}
	}
	
	TheEngineLog::instance().writeInfo( string ("Adding resource " ) + fName );
	
	// Create resource
	Resource *resource = 0x0;
	switch( type ) 
	{
	case ResourceTypes::SceneGraph:
		resource = new SceneGraphResource( fName, flags );
		break;
	case ResourceTypes::Geometry:
		resource = new GeometryResource( fName, flags );
		break;
	case ResourceTypes::Animation:
		resource = new AnimationResource( fName, flags );
		break;
	case ResourceTypes::Material:
		resource = new MaterialResource( fName, flags );
		break;
	case ResourceTypes::Code:
		resource = new CodeResource( fName, flags );
		break;
	case ResourceTypes::Shader:
		resource = new ShaderResource( fName, flags );
		break;
	case ResourceTypes::Texture2D:
		resource = new Texture2DResource( fName, flags );
		break;
	case ResourceTypes::TextureCube:
		resource = new TextureCubeResource( fName, flags );
		break;
	case ResourceTypes::Effect:
		resource = new EffectResource( fName, flags );
		break;
	default:
		return 0;
	}		
	
	if( userCall ) resource->_userRef = true;
	else ++resource->_refCount;
	
	// Try to insert resource in free slot
	for( unsigned int i = 0; i < _resources.size(); ++i )
	{
		if( _resources[i] == 0x0 )
		{
			resource->_handle = i + 1;
			_resources[i] = resource;
			return i + 1;
		}
	}
	
	// If there is no free slot, add resource
	resource->_handle = _resources.size() + 1;
	_resources.push_back( resource );
	return (unsigned int)_resources.size();
}


bool ResourceManager::removeResource( unsigned int handle, bool userCall )
{
	Resource *res = resolveResHandle( handle );

	if( res == 0x0 )
	{	
		TheEngineLog::instance().writeError( "Invalid resource handler in removeResource" );
		return false;
	}

	// Decrease reference count
	if( !userCall && res->_refCount > 0 ) --res->_refCount;
	else if( userCall ) res->_userRef = false;

	return true;
}


void ResourceManager::clear()
{
	// Note:	Resources have to be removed in the right order so that references
	//			don't become invalid
	
	// Remove scene graph resources
	for( unsigned int i = 0; i < _resources.size(); ++i )
	{
		if( _resources[i] != 0x0 && _resources[i]->_type == ResourceTypes::SceneGraph )
		{
			delete _resources[i]; _resources[i] = 0x0;
		}
	}
	// Remove materials
	for( unsigned int i = 0; i < _resources.size(); ++i )
	{
		if( _resources[i] != 0x0 && _resources[i]->_type == ResourceTypes::Material )
		{
			delete _resources[i]; _resources[i] = 0x0;
		}
	}
	// Remove shaders
	for( unsigned int i = 0; i < _resources.size(); ++i )
	{
		if( _resources[i] != 0x0 && _resources[i]->_type == ResourceTypes::Shader )
		{
			delete _resources[i]; _resources[i] = 0x0;
		}
	}
	// Remove all other resources
	for( unsigned int i = 0; i < _resources.size(); ++i )
	{
		if( _resources[i] != 0x0 )
		{
			delete _resources[i]; _resources[i] = 0x0;
		}
	}
}


string ResourceManager::queryUnloadedResource()
{
	for( unsigned int i = 0; i < _resources.size(); ++i )
	{
		if( _resources[i] == 0x0 ) continue;
		
		if( !_resources[i]->_loaded && !_resources[i]->_noQuery )
			return _resources[i]->_name;
	}

	return "";
}


bool ResourceManager::loadResource( const string &name, const char *data, unsigned int size )
{
	for( unsigned int i = 0; i < _resources.size(); ++i )
	{
		if( _resources[i] == 0x0 ) continue;
		
		if( _resources[i]->_name == name )
		{
			TheEngineLog::instance().writeInfo( string( "Loading resource <b>" ) + name + "</b>" );
			return _resources[i]->load( data, size );
		}
	}

	return false;
}


const void *ResourceManager::getResourceData( unsigned int handle, ResourceData::List param )
{
	Resource *res = resolveResHandle( handle );
	
	if( res == 0x0 )
	{	
		TheEngineLog::instance().writeError( "Invalid resource handler in getResourceData" );
		return 0x0;
	}

	return res->getData( param );
}


bool ResourceManager::updateResourceData( unsigned int handle, ResourceData::List param,
										  const void *data, unsigned int size)
{
	if( data == 0x0 ) return false;
	
	Resource *res = resolveResHandle( handle );
	
	if( res == 0x0 )
	{	
		TheEngineLog::instance().writeError( "Invalid resource handler in updateResourceData" );
		return false;
	}

	if( !res->_loaded ) return false;

	return res->updateData( param, data, size );
}


void ResourceManager::releaseUnusedResources()
{
	bool removedRes = false;
	
	for( unsigned int i = 0; i < _resources.size(); ++i )
	{
		if( _resources[i] == 0x0 ) continue;
		
		if( !_resources[i]->_userRef && _resources[i]->_refCount == 0 )
		{
			TheEngineLog::instance().writeInfo( string( "Removed resource " ) + _resources[i]->_name );
			delete _resources[i]; _resources[i] = 0x0;
			removedRes = true;
		}
	}

	// Releasing a resource can remove dependencies from other resources which can also be released
	if( removedRes ) releaseUnusedResources();
}
