// *************************************************************************************************
//
// Horde3D
//   Next-Generation Graphics Engine
// --------------------------------------
// Copyright (C) 2006-2007 Nicolas Schulz
//
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//
// *************************************************************************************************

#include "egScene.h"
#include "egSingleton.h"
#include "egSceneGraphRes.h"
#include "egModel.h"
#include "egMaterial.h"
#include "egLight.h"
#include "egCamera.h"
#include "egParticle.h"
#include "tinyxml.h"

#include "utDebug.h"


// *************************************************************************************************
// Class SceneNode
// *************************************************************************************************

SceneNode::SceneNode( SceneNodeTypes::List type, const string &name )
{
	_type = type;
	_name = name;
	_active = true;
	_handle = 0;
	_dirty = true;
	_parent = 0x0;
	_attachment = 0x0;
}


SceneNode::~SceneNode()
{
	if( _attachment != 0x0 ) _attachment->_node = 0;
}


SceneNode *SceneNode::findChild( const string &name, SceneNodeTypes::List searchChildType, unsigned int index, bool recursive )
{
	SceneNode *node = 0x0;
	
	// Check if child is in children
	for( unsigned int i = 0; i < _children.size(); ++i )
	{
		if( name == "" || name == _children[i]->_name )
		{
			if( index == 0 )
			{
				node = _children[i];
				break;
			}
			else
			{
				--index;
			}
		}
	}

	if( !recursive || node != 0x0 ) return node;

	// Recurse over children
	for( unsigned int i = 0; i < _children.size(); ++i )
	{
		if( searchChildType != SceneNodeTypes::Undefined && _children[i]->_type != searchChildType )
		{	
			continue;
		}
		
		node = _children[i]->findChild( name, searchChildType, index, recursive );
		if( node != 0x0 ) return node;
	}

	return 0x0;
}


void SceneNode::getTransform( Vec3f &pos, Vec3f &rot, Vec3f &scale )
{
	_relTrans.decompose( pos, rot, scale );
	rot.x = radToDeg( rot.x );
	rot.y = radToDeg( rot.y );
	rot.z = radToDeg( rot.z );
}


void SceneNode::setTransform( Vec3f pos, Vec3f rot, Vec3f scale )
{
	_relTrans = Matrix4f();
	_relTrans.scale( scale.x, scale.y, scale.z );
	_relTrans.rotate( degToRad( rot.x ), degToRad( rot.y ), degToRad( rot.z ) );
	_relTrans.translate( pos.x, pos.y, pos.z );
	
	markDirty();
}


void SceneNode::setTransform( const Matrix4f &mat )
{
	_relTrans = mat;
	
	markDirty();
}


const void SceneNode::getTransMatrices( const float **relMat, const float **absMat )
{
	if( relMat != 0x0 )
	{
		*relMat = &_relTrans.x[0];
	}
	
	if( absMat != 0x0 )
	{
		TheSceneManager::instance().updateNodes();
		*absMat = &_absTrans.x[0];
	}
}


bool SceneNode::canAttach( SceneNode &parent )
{
	return true;
}


void SceneNode::markChildrenDirty()
{	
	for( unsigned int i = 0, s = _children.size(); i < s; ++i )
	{
		if( !_children[i]->_dirty )
		{	
			_children[i]->_dirty = true;
			_children[i]->markChildrenDirty();
		}
	}
}


void SceneNode::markDirty()
{
	_dirty = true;
	
	SceneNode *node = _parent;
	while( node != 0x0 )
	{
		node->_dirty = true;
		node = node->_parent;
	}

	markChildrenDirty();
}


bool SceneNode::update()
{
	if( !_dirty ) return false;
	
	bool bBoxChanged = false;
	
	// Calculate absolute matrix
	if( _parent != 0x0 )
		//_absTrans = _parent->_absTrans * _relTrans;
		_absTrans.fastMult( _parent->_absTrans, _relTrans );
	else
		_absTrans = _relTrans;
	
	// If there is a local bounding box, transform it to world space
	BoundingBox *locBBox = getLocalBBox();
	if( locBBox != 0x0 )
	{
		_bBox = *locBBox;
		_bBox.transform( _absTrans );
		bBoxChanged = true;
	}

	_dirty = false;
	
	onUpdate();
	if( _attachment != 0x0 ) _attachment->onUpdate();

	// Visit children
	for( unsigned int i = 0, s = _children.size(); i < s; ++i )
	{
		bBoxChanged |= _children[i]->update();
	}	

	// Recalculate bounding box if necessary
	if( bBoxChanged )
	{
		if( locBBox == 0x0 )
		{	
			bBoxChanged = false;
			_bBox.clear();
		}

		for( unsigned int i = 0, s = _children.size(); i < s; ++i )
		{
			bBoxChanged |= _bBox.makeUnion( _children[i]->_bBox ); 
		}
	}
	
	return bBoxChanged;
}


void SceneNode::onUpdate()
{
}


void SceneNode::onAttach( SceneNode &node )
{
}


void SceneNode::onDetach( SceneNode &node )
{
}



// *************************************************************************************************
// Class GroupNode
// *************************************************************************************************

GroupNode::GroupNode( const string &name ) : SceneNode( SceneNodeTypes::Group, name )
{
	_minDist = 0;
	_maxDist = Math::MaxFloat;
}



float GroupNode::getParam( GroupNodeParams::List param )
{
	switch( param )
	{
	case GroupNodeParams::MinDist:
		return _minDist;
	case GroupNodeParams::MaxDist:
		return _maxDist;
	default:
		return Math::NaN;
	}
}


bool GroupNode::setParam( GroupNodeParams::List param, float value )
{
	switch( param )
	{
	case GroupNodeParams::MinDist:
		_minDist = value;
		return true;
	case GroupNodeParams::MaxDist:
		_maxDist = value;
		return true;
	default:
		return false;
	}
}


// *************************************************************************************************
// Class SceneManager
// *************************************************************************************************

SceneManager::SceneManager()
{
	_attFactory = 0x0;
	_attCollector = 0x0;
	
	SceneNode *rootNode = new GroupNode( "RootNode" );
	rootNode->_handle = RootNode;
	_nodes.push_back( rootNode );
	SceneNode *camNode = new CameraNode( "PrimeTimeCam" );
	camNode->_handle = PrimeTimeCam;
	addNode( camNode, *rootNode );
	
	_meshQueue.reserve( 10000 );
	_lightQueue.reserve( 20 );
}


SceneManager::~SceneManager()
{
	for( unsigned int i = 0; i < _nodes.size(); ++i )
	{
		delete _nodes[i]; _nodes[i] = 0x0;
	}
}


void SceneManager::updateNodes()
{
	getRootNode().update();
}


void SceneManager::updateQueuesRec( const Frustum &frustum1, const Frustum *frustum2, SceneNode &node,
								    bool meshQueue, bool emitterQueue, bool lightQueue )
{
	if( !node._active ) return;
	
	if( node._type == SceneNodeTypes::Group )
	{
		// LOD
		Vec3f nodePos( node._absTrans.c[3][0], node._absTrans.c[3][1], node._absTrans.c[3][2] );
		float dist = (nodePos - frustum1.getOrigin()).length();
		
		GroupNode *gn = (GroupNode *)&node;
		if( dist < gn->_minDist || dist >= gn->_maxDist ) return;
	}
	else if( meshQueue && node._type == SceneNodeTypes::Mesh &&
			 !frustum1.cullBox( node._bBox ) )
	{
		if( frustum2 == 0x0 || !frustum2->cullBox( node._bBox ) )
			_meshQueue.push_back( &node );
	}
	else if( emitterQueue && node._type == SceneNodeTypes::Emitter &&
			 !frustum1.cullBox( *node.getLocalBBox() ) )	// Use local BB since particle positions are absolute
	{
		if( frustum2 == 0x0 || !frustum2->cullBox( *node.getLocalBBox() ) )
			_emitterQueue.push_back( &node );
	}
	else if( lightQueue && node._type == SceneNodeTypes::Light )
	{
		_lightQueue.push_back( &node );
	}

	// Recurse over children
	for( unsigned int i = 0, s = node._children.size(); i < s; ++i )
	{
		updateQueuesRec( frustum1, frustum2, *node._children[i], meshQueue, emitterQueue, lightQueue );
	}
}


void SceneManager::updateQueues( const Frustum &frustum1, const Frustum *frustum2,
								 bool meshQueue, bool emitterQueue, bool lightQueue )
{
	// Clear without affecting capacity
	if( meshQueue ) _meshQueue.resize( 0 );
	if( emitterQueue ) _emitterQueue.resize( 0 );
	if( lightQueue ) _lightQueue.resize( 0 );
	
	updateNodes();
	
	updateQueuesRec( frustum1, frustum2, getRootNode(), meshQueue, emitterQueue, lightQueue );
}


unsigned int SceneManager::parseNode( SGRNode &node, SceneNode *parent )
{
	if( parent == 0x0 ) return 0;
	
	SceneNode *newParent = 0x0;
	unsigned int handle = 0;
	
	// Parse node
	if( node.type == SGRNodeTypes::Group )
	{
		SGRGroupNode *node1 = (SGRGroupNode *)&node;
		handle = addGroupNode( *parent, node1->name );
		SceneNode *sn = resolveNodeHandle( handle );
		if( sn != 0x0 )
		{	
			newParent = sn;
			((GroupNode *)sn)->setParam( GroupNodeParams::MinDist, node1->minDist );
			((GroupNode *)sn)->setParam( GroupNodeParams::MaxDist, node1->maxDist );
		}
	}
	else if( node.type == SGRNodeTypes::Reference )
	{
		SGRReferenceNode *node1 = (SGRReferenceNode *)&node;
		
		Resource *res = TheResourceManager::instance().resolveResHandle( node1->sceneGraphRes );
		if( res != 0x0 )
		{
			handle = parseNode( *((SceneGraphResource *)res)->getRootNode(), parent );
			SceneNode *sn = resolveNodeHandle( handle );
			if( sn != 0x0 )
			{
				newParent = sn;
				if( node1->name != "" ) sn->_name = node1->name;
			}
		}
	}
	else if( node.type == SGRNodeTypes::Model )
	{
		SGRModelNode *node1 = (SGRModelNode *)&node;
		handle = addModelNode( *parent, node1->name, node1->geoRes );
		SceneNode *sn = resolveNodeHandle( handle );
		if( sn != 0x0 ) newParent = sn;
	}
	else if( node.type == SGRNodeTypes::Mesh )
	{
		SGRMeshNode *node1 = (SGRMeshNode *)&node;
		handle = addMeshNode( *parent, node1->name, node1->matRes,
							  node1->batchStart, node1->batchCount,
							  node1->vertRStart, node1->vertREnd );
		SceneNode *sn = resolveNodeHandle( handle );
		if( sn != 0x0 ) newParent = sn;
	}
	else if( node.type == SGRNodeTypes::Joint )
	{
		SGRJointNode *node1 = (SGRJointNode *)&node;
		handle = addJointNode( *parent, node1->name, node1->jointIndex );
		SceneNode *sn = resolveNodeHandle( handle );
		if( sn != 0x0 ) newParent = sn;
	}
	else if( node.type == SGRNodeTypes::Light )
	{
		SGRLightNode *node1 = (SGRLightNode *)&node;
		handle = addLightNode( *parent, node1->name, node1->matRes,
							   node1->lightingContext, node1->shadowContext );
		SceneNode *sn = resolveNodeHandle( handle );
		if( sn != 0x0 )
		{	
			newParent = &getNode( handle );
			((LightNode *)sn)->setParam( LightNodeParams::Radius, node1->radius );
			((LightNode *)sn)->setParam( LightNodeParams::FOV, node1->fov );
			((LightNode *)sn)->setParam( LightNodeParams::Col_R, node1->col_R );
			((LightNode *)sn)->setParam( LightNodeParams::Col_G, node1->col_G );
			((LightNode *)sn)->setParam( LightNodeParams::Col_B, node1->col_B );
			((LightNode *)sn)->setParam( LightNodeParams::ShadowMapCount, (float)node1->shadowMapCount );
			((LightNode *)sn)->setParam( LightNodeParams::ShadowSplitLambda, node1->shadowSplitLambda );
			((LightNode *)sn)->setParam( LightNodeParams::ShadowMapBias, node1->shadowMapBias );
			sn->setTransform( node1->pos, node1->rot, node1->scale );
		}
	}
	else if( node.type == SGRNodeTypes::Camera )
	{
		SGRCameraNode *node1 = (SGRCameraNode *)&node;
		handle = addCameraNode( *parent, node1->name );
		SceneNode *sn = resolveNodeHandle( handle );
		if( sn != 0x0 )
		{	
			newParent = &getNode( handle );
			((CameraNode *)sn)->setParam( CameraNodeParams::LeftPlane, node1->leftPlane );
			((CameraNode *)sn)->setParam( CameraNodeParams::RightPlane, node1->rightPlane );
			((CameraNode *)sn)->setParam( CameraNodeParams::BottomPlane, node1->bottomPlane );
			((CameraNode *)sn)->setParam( CameraNodeParams::TopPlane, node1->topPlane );
			((CameraNode *)sn)->setParam( CameraNodeParams::NearPlane, node1->nearPlane );
			((CameraNode *)sn)->setParam( CameraNodeParams::FarPlane, node1->farPlane );
			sn->setTransform( node1->pos, node1->rot, node1->scale );
		}
	}
	else if( node.type == SGRNodeTypes::Emitter )
	{
		SGREmitterNode *node1 = (SGREmitterNode *)&node;
		handle = addEmitterNode( *parent, node1->name, node1->matRes, node1->effectRes,
								 node1->maxParticleCount, node1->respawnCount );
		SceneNode *sn = resolveNodeHandle( handle );
		if( sn != 0x0 )
		{	
			newParent = &getNode( handle );
			((EmitterNode *)sn)->setParam( EmitterNodeParams::Delay, node1->delay );
			((EmitterNode *)sn)->setParam( EmitterNodeParams::EmissionRate, node1->emissionRate );
			((EmitterNode *)sn)->setParam( EmitterNodeParams::SpreadAngle, node1->spreadAngle );
			((EmitterNode *)sn)->setParam( EmitterNodeParams::ForceX, node1->fx );
			((EmitterNode *)sn)->setParam( EmitterNodeParams::ForceY, node1->fy );
			((EmitterNode *)sn)->setParam( EmitterNodeParams::ForceZ, node1->fz );
			sn->setTransform( node1->pos, node1->rot, node1->scale );
		}
	}

	if( newParent != 0x0 ) newParent->setTransform( node.pos, node.rot, node.scale );

	// Raise integration library event
	if( newParent != 0x0 && TheSceneManager::instance().getAttFactoryPtr() != 0x0 &&
		node.attachmentParams != "" )
	{
		(*TheSceneManager::instance().getAttFactoryPtr())(
			newParent->getHandle(), node.attachmentParams.c_str() );
	}
	
	// Parse children
	if( newParent != 0x0 )
	{
		for( unsigned int i = 0; i < node.children.size(); ++i )
		{
			parseNode( *node.children[i], newParent );
		}
	}

	return handle ;
}


unsigned int SceneManager::addNode( SceneNode *node, SceneNode &parent )
{
	//TheEngineLog::instance().writeInfo( string( "Adding node " ) + name );
	
	// Check if node can be attached to parent
	if( !node->canAttach( parent ) )
	{
		TheEngineLog::instance().writeError( "Invalid node addition" );
		delete node; node = 0x0;
		return 0;
	}
	
	node->_parent = &parent;
	
	// Attach to parent
	parent._children.push_back( node );

	// Raise event
	parent.onAttach( *node );

	// Mark tree as dirty
	node->markDirty();
	
	// Try to insert node in free slot
	for( unsigned int i = 0; i < _nodes.size(); ++i )
	{
		if( _nodes[i] == 0x0 )
		{
			node->_handle = i + 1;
			_nodes[i] = node;
			return i + 1;
		}
	}
	
	// If there is no free slot, add node
	_nodes.push_back( node );
	node->_handle = _nodes.size();
	return (unsigned int)_nodes.size();
}


unsigned int SceneManager::addGroupNode( SceneNode &parent, string name )
{
	GroupNode *node = new GroupNode( name );

	return addNode( node, parent );
}


unsigned int SceneManager::addModelNode( SceneNode &parent, string name, unsigned int geoRes )
{
	Resource *res = TheResourceManager::instance().resolveResHandle( geoRes );
	
	if( res != 0x0 && res->getType() == ResourceTypes::Geometry )
	{
		ModelNode *node = new ModelNode( name, *(GeometryResource *)res );

		return addNode( node, parent );
	}
	else
	{
		// TODO: Use default resource?
		TheEngineLog::instance().writeError( "Invalid geometry resource for model node" );
		return 0;
	}
}


unsigned int SceneManager::addMeshNode( SceneNode &parent, string name, unsigned int matRes,
									    unsigned int batchStart, unsigned int batchCount,
										unsigned int vertRStart, unsigned int vertREnd )
{
	// TODO: Check that batch and range parameters are valid
	MeshNode *node = new MeshNode( name, parent, matRes, batchStart, batchCount, vertRStart, vertREnd );

	return addNode( node, parent );
}


unsigned int SceneManager::addJointNode( SceneNode &parent, string name, unsigned int jointIndex )
{
	// TODO: Check that joint index is valid
	JointNode *node = new JointNode( name, parent, jointIndex );

	return addNode( node, parent );
}


unsigned int SceneManager::addLightNode( SceneNode &parent, string name, unsigned int materialRes,
										 const string &lightingContext, const string &shadowContext )
{
	LightNode *node = new LightNode( name, materialRes, lightingContext, shadowContext );

	return addNode( node, parent );
}


unsigned int SceneManager::addCameraNode( SceneNode &parent, string name )
{
	CameraNode *node = new CameraNode( name );

	return addNode( node, parent );
}


unsigned int SceneManager::addEmitterNode( SceneNode &parent, string name, unsigned int matRes,
										   unsigned int effectRes, unsigned int maxParticleCount,
										   int respawnCount )
{
	EmitterNode *node = new EmitterNode( name, matRes, effectRes, maxParticleCount, respawnCount );

	return addNode( node, parent );
}


unsigned int SceneManager::addNodes( SceneNode &parent, SceneGraphResource &sgRes )
{
	// Parse root node
	return parseNode( *sgRes.getRootNode(), &parent );
}


void SceneManager::removeNodeRec( SceneNode *node )
{
	unsigned int handle = node->_handle;
	
	// Raise event
	if( handle != RootNode && handle != PrimeTimeCam )
		node->_parent->onDetach( *node );

	// Raise integration library event
	if( TheSceneManager::instance().getAttCollectorPtr() != 0x0 && node->_attachment != 0x0 )
	{
		(*TheSceneManager::instance().getAttCollectorPtr())( handle );
	}

	// Remove children
	for( unsigned int i = 0; i < node->_children.size(); ++i )
	{
		removeNodeRec( node->_children[i] );
	}
	
	// Delete node
	if( handle != RootNode && handle != PrimeTimeCam )
	{
		delete _nodes[handle - 1]; _nodes[handle - 1] = 0x0;
	}
}


bool SceneManager::removeNode( unsigned int handle )
{
	SceneNode *sn = resolveNodeHandle( handle );
	if( sn == 0x0 )		// Call allowed for rootnode
	{
		TheEngineLog::instance().writeError( "Invalid node handler in removeNode" );
		return false;
	}

	// Reset queues
	_meshQueue.resize( 0 );
	_lightQueue.resize( 0 );

	SceneNode *parent = sn->_parent;

	removeNodeRec( sn );
	
	// Remove node from parent
	if( handle != RootNode && handle != PrimeTimeCam )
	{
		// Find child
		for( unsigned int i = 0; i < parent->_children.size(); ++i )
		{
			if( parent->_children[i] == sn )
			{
				parent->_children.erase( parent->_children.begin() + i );
				break;
			}
		}
	}
	else
	{
		sn->_children.clear();
	}
	
	// Mark dirty
	if( parent != 0x0 ) parent->markDirty();
	
	return true;
}
