// *************************************************************************************************
//
// Horde3D
//   Next-Generation Graphics Engine
// --------------------------------------
// Copyright (C) 2006-2007 Nicolas Schulz
//
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//
// *************************************************************************************************

#pragma once

#include "utMath.h"
#include "egPrimitives.h"
#include "egResource.h"
#include "egIntegrationLib.h"

class TiXmlElement;
struct SGRNode;
class SceneGraphResource;


class SceneNode
{
protected:
	
	SceneNodeTypes::List	_type;
	string					_name;
	bool					_active;
	unsigned int			_handle;
	bool					_dirty;
	Matrix4f				_relTrans, _absTrans;			// Transformation matrices
	BoundingBox				_bBox;							// AABB in world space

	SceneNode				*_parent;		// Parent node
	vector< SceneNode * >	_children;		// Child nodes
	NodeAttachment			*_attachment;	// User defined attachment
	
	void markChildrenDirty();

public:

	SceneNode( SceneNodeTypes::List type, const string &name );
	virtual ~SceneNode();

	SceneNodeTypes::List getType() { return _type; };
	SceneNode *findChild( const string &name, SceneNodeTypes::List searchChildType,
						  unsigned int index, bool recursive );

	void setActivation( bool active ) { _active = active; }
	void getTransform( Vec3f &pos, Vec3f &rot, Vec3f &scale );	// Not virtual for performance
	void setTransform( Vec3f pos, Vec3f rot, Vec3f scale );	// Not virtual for performance
	void setTransform( const Matrix4f &mat );
	const void getTransMatrices( const float **relMat, const float **absMat );

	virtual BoundingBox *getLocalBBox() { return 0x0; }
	virtual bool canAttach( SceneNode &parent );
	void markDirty();
	bool update();

	virtual void onUpdate();
	virtual void onAttach( SceneNode &node );
	virtual void onDetach( SceneNode &node );

	unsigned int getHandle() { return _handle; }
	SceneNode *getParent() { return _parent; }
	const string &getName() { return _name; }
	void setName( const string &name ) { _name = name; }
	vector< SceneNode * > &getChildren() { return _children; }
	Matrix4f &getRelTrans() { return _relTrans; }
	BoundingBox &getBBox() { return _bBox; }
	NodeAttachment *getAttachment() { return _attachment; }
	void setAttachment( NodeAttachment *attachment ) { _attachment = attachment; }

	friend class SceneManager;
	friend class Renderer;
};


class GroupNode : public SceneNode
{
protected:

	float	_minDist, _maxDist;

public:

	GroupNode( const string &name );

	float getParam( GroupNodeParams::List param );
	bool setParam( GroupNodeParams::List param, float value );

	friend class Renderer;
	friend class SceneManager;
};


class SceneManager
{
protected:

	vector< SceneNode *>	_nodes;		// _nodes[0] is root node,
										// _nodes[1] is default camera
	vector< SceneNode *>	_meshQueue, _emitterQueue, _lightQueue;
	NodeAttFactoryFunc		_attFactory;
	NodeAttCollectorFunc	_attCollector;

	void updateQueuesRec( const Frustum &frustum1, const Frustum *frustum2, SceneNode &node,
						  bool meshQueue, bool emitterQueue, bool lightQueue );
	unsigned int addNode( SceneNode *node, SceneNode &parent );
	unsigned int parseNode( SGRNode &node, SceneNode *parent );
	void removeNodeRec( SceneNode *node );

public:

	SceneManager();
	~SceneManager();
	
	unsigned int addNodes( SceneNode &parent, SceneGraphResource &sgRes );
	bool removeNode( unsigned int handle );

	unsigned int addGroupNode( SceneNode &parent, string name );
	unsigned int addModelNode( SceneNode &parent, string name, unsigned int geoRes );
	unsigned int addMeshNode( SceneNode &parent, string name, unsigned int matRes,
		unsigned int batchStart, unsigned int batchCount, unsigned int vertRStart, unsigned int vertREnd );
	unsigned int addJointNode( SceneNode &parent, string name, unsigned int jointIndex );
	unsigned int addLightNode( SceneNode &parent, string name, unsigned int materialRes,
		const string &lightingContext, const string &shadowContext );
	unsigned int addCameraNode( SceneNode &parent, string name );
	unsigned int addEmitterNode( SceneNode &parent, string name, unsigned int matRes,
		unsigned int effectRes, unsigned int maxParticleCount, int respawnCount );

	void updateNodes();
	void updateQueues( const Frustum &frustum1, const Frustum *frustum2,
					   bool meshQueue, bool emitterQueue, bool lightQueue );

	SceneNode &getNode( unsigned int handle ) { return *_nodes[handle - 1]; }
	SceneNode &getRootNode() { return *_nodes[0]; }
	SceneNode &getDefCamNode() { return *_nodes[1]; }
	vector< SceneNode * > &getMeshQueue() { return _meshQueue; }
	vector< SceneNode * > &getEmitterQueue() { return _emitterQueue; }
	vector< SceneNode * > &getLightQueue() { return _lightQueue; }
	NodeAttFactoryFunc &getAttFactoryPtr() { return _attFactory; }
	NodeAttCollectorFunc &getAttCollectorPtr() { return _attCollector; }
	
	SceneNode *resolveNodeHandle( unsigned int handle )
		{ return (handle != 0 && handle - 1 < _nodes.size()) ? _nodes[handle - 1] : 0x0; }

	friend class Renderer;
};
