// *************************************************************************************************
//
// Horde3D
//   Next-Generation Graphics Engine
// --------------------------------------
// Copyright (C) 2006-2007 Nicolas Schulz
//
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//
// *************************************************************************************************

// TODO: Error checking, cancel loading when required attribute is not found

#include "egSceneGraphRes.h"
#include "egSingleton.h"
#include "tinyxml.h"
#include "utPlatform.h"

#include "utDebug.h"


SceneGraphResource::SceneGraphResource( const string &name, unsigned int flags ) :
	Resource( ResourceTypes::SceneGraph, name, flags )
{
	initDefault();
}


SceneGraphResource::~SceneGraphResource()
{
	release();
}


void SceneGraphResource::initDefault()
{
	// Create default root node
	_rootNode = new SGRGroupNode();
	_rootNode->name = _name;
}


void SceneGraphResource::release()
{
	delete _rootNode; _rootNode = 0x0;
}


void SceneGraphResource::parseBaseAttributes( TiXmlElement *xmlNode, SGRNode &node )
{
	if( xmlNode->Attribute( "name" ) != 0x0 )
		node.name = xmlNode->Attribute( "name" );
	if( xmlNode->Attribute( "px" ) != 0x0 )
		node.pos.x = (float)atof( xmlNode->Attribute( "px" ) );
	if( xmlNode->Attribute( "py" ) != 0x0 )
		node.pos.y = (float)atof( xmlNode->Attribute( "py" ) );
	if( xmlNode->Attribute( "pz" ) != 0x0 )
		node.pos.z = (float)atof( xmlNode->Attribute( "pz" ) );
	if( xmlNode->Attribute( "rx" ) != 0x0 )
		node.rot.x = (float)atof( xmlNode->Attribute( "rx" ) );
	if( xmlNode->Attribute( "ry" ) != 0x0 )
		node.rot.y = (float)atof( xmlNode->Attribute( "ry" ) );
	if( xmlNode->Attribute( "rz" ) != 0x0 )
		node.rot.z = (float)atof( xmlNode->Attribute( "rz" ) );
	if( xmlNode->Attribute( "sx" ) != 0x0 )
		node.scale.x = (float)atof( xmlNode->Attribute( "sx" ) );
	if( xmlNode->Attribute( "sy" ) != 0x0 )
		node.scale.y = (float)atof( xmlNode->Attribute( "sy" ) );
	if( xmlNode->Attribute( "sz" ) != 0x0 )
		node.scale.z = (float)atof( xmlNode->Attribute( "sz" ) );

	TiXmlElement *node1 = xmlNode->FirstChildElement( "Attachment" );
	if( node1 != 0x0 )
	{
		TiXmlPrinter printer;
		printer.SetIndent( "" );
		printer.SetLineBreak( "" );

		node1->Accept( &printer );
		node.attachmentParams = printer.CStr();
	}
}


void SceneGraphResource::parseNode( TiXmlElement *xmlNode, SGRNode *parentNode )
{
	SGRNode *baseNode = 0x0;
	
	// Parse node
	if( strcmp( xmlNode->Value(), "Group" ) == 0 )
	{
		SGRGroupNode *node = new SGRGroupNode();
		baseNode = node;
		parseBaseAttributes( xmlNode, *node );

		if( xmlNode->Attribute( "minDist" ) != 0x0 )
		{
			node->minDist = (float)atof( xmlNode->Attribute( "minDist" ) );
		}
		if( xmlNode->Attribute( "maxDist" ) != 0x0 )
		{
			node->maxDist = (float)atof( xmlNode->Attribute( "maxDist" ) );
		}
	}
	else if( strcmp( xmlNode->Value(), "Reference" ) == 0 )
	{
		SGRReferenceNode *node = new SGRReferenceNode();
		baseNode = node;
		parseBaseAttributes( xmlNode, *node );

		if( xmlNode->Attribute( "sceneGraph" ) != 0x0 )
		{
			node->sceneGraphRes = TheResourceManager::instance().addResource(
				ResourceTypes::SceneGraph, xmlNode->Attribute( "sceneGraph" ), 0, false );
		}
	}
	else if( strcmp( xmlNode->Value(), "Model" ) == 0 )
	{
		SGRModelNode *node = new SGRModelNode();
		baseNode = node;
		parseBaseAttributes( xmlNode, *node );

		if( xmlNode->Attribute( "geometry" ) != 0x0 )
		{
			node->geoRes = TheResourceManager::instance().addResource(
				ResourceTypes::Geometry, xmlNode->Attribute( "geometry" ), 0, false );
		}
	}
	else if( strcmp( xmlNode->Value(), "Mesh" ) == 0 )
	{
		SGRMeshNode *node = new SGRMeshNode();
		baseNode = node;
		parseBaseAttributes( xmlNode, *node );

		if( xmlNode->Attribute( "batchStart" ) != 0x0 )
		{
			node->batchStart = atoi( xmlNode->Attribute( "batchStart" ) );
		}
		if( xmlNode->Attribute( "batchCount" ) != 0x0 )
		{
			node->batchCount = atoi( xmlNode->Attribute( "batchCount" ) );
		}
		if( xmlNode->Attribute( "vertRStart" ) != 0x0 )
		{
			node->vertRStart = atoi( xmlNode->Attribute( "vertRStart" ) );
		}
		if( xmlNode->Attribute( "vertREnd" ) != 0x0 )
		{
			node->vertREnd = atoi( xmlNode->Attribute( "vertREnd" ) );
		}
		if( xmlNode->Attribute( "material" ) != 0x0 )
		{
			node->matRes = TheResourceManager::instance().addResource(
				ResourceTypes::Material, xmlNode->Attribute( "material" ), 0, false );
		}
	}
	else if( strcmp( xmlNode->Value(), "Joint" ) == 0 )
	{
		SGRJointNode *node = new SGRJointNode();
		baseNode = node;
		parseBaseAttributes( xmlNode, *node );

		if( xmlNode->Attribute( "jointIndex" ) != 0x0 )
		{
			node->jointIndex = atoi( xmlNode->Attribute( "jointIndex" ) );
		}
	}
	else if( strcmp( xmlNode->Value(), "Light" ) == 0 )
	{
		SGRLightNode *node = new SGRLightNode();
		baseNode = node;
		parseBaseAttributes( xmlNode, *node );

		if( xmlNode->Attribute( "material" ) != 0x0 )
		{
			node->matRes = TheResourceManager::instance().addResource(
				ResourceTypes::Material, xmlNode->Attribute( "material" ), 0, false );
		}
		if( xmlNode->Attribute( "lightingContext" ) != 0x0 )
		{
			node->lightingContext = xmlNode->Attribute( "lightingContext" );
		}
		if( xmlNode->Attribute( "shadowContext" ) != 0x0 )
		{
			node->shadowContext = xmlNode->Attribute( "shadowContext" );
		}
		if( xmlNode->Attribute( "radius" ) != 0x0 )
		{
			node->radius = (float)atof( xmlNode->Attribute( "radius" ) );
		}
		if( xmlNode->Attribute( "fov" ) != 0x0 )
		{
			node->fov = (float)atof( xmlNode->Attribute( "fov" ) );
		}
		if( xmlNode->Attribute( "col_R" ) != 0x0 )
		{
			node->col_R = (float)atof( xmlNode->Attribute( "col_R" ) );
		}
		if( xmlNode->Attribute( "col_G" ) != 0x0 )
		{
			node->col_G = (float)atof( xmlNode->Attribute( "col_G" ) );
		}
		if( xmlNode->Attribute( "col_B" ) != 0x0 )
		{
			node->col_B = (float)atof( xmlNode->Attribute( "col_B" ) );
		}
		if( xmlNode->Attribute( "shadowMapCount" ) != 0x0 )
		{
			node->shadowMapCount = (unsigned int)atoi( xmlNode->Attribute( "shadowMapCount" ) );
		}
		if( xmlNode->Attribute( "shadowSplitLambda" ) != 0x0 )
		{
			node->shadowSplitLambda = (float)atof( xmlNode->Attribute( "shadowSplitLambda" ) );
		}
		if( xmlNode->Attribute( "shadowMapBias" ) != 0x0 )
		{
			node->shadowMapBias = (float)atof( xmlNode->Attribute( "shadowMapBias" ) );
		}
	}
	else if( strcmp( xmlNode->Value(), "Camera" ) == 0 )
	{
		SGRCameraNode *node = new SGRCameraNode();
		baseNode = node;
		parseBaseAttributes( xmlNode, *node );

		if( xmlNode->Attribute( "leftPlane" ) != 0x0 )
		{
			node->leftPlane = (float)atof( xmlNode->Attribute( "leftPlane" ) );
		}
		if( xmlNode->Attribute( "rightPlane" ) != 0x0 )
		{
			node->rightPlane = (float)atof( xmlNode->Attribute( "rightPlane" ) );
		}
		if( xmlNode->Attribute( "bottomPlane" ) != 0x0 )
		{
			node->bottomPlane = (float)atof( xmlNode->Attribute( "bottomPlane" ) );
		}
		if( xmlNode->Attribute( "topPlane" ) != 0x0 )
		{
			node->topPlane = (float)atof( xmlNode->Attribute( "topPlane" ) );
		}
		if( xmlNode->Attribute( "nearPlane" ) != 0x0 )
		{
			node->nearPlane = (float)atof( xmlNode->Attribute( "nearPlane" ) );
		}
		if( xmlNode->Attribute( "farPlane" ) != 0x0 )
		{
			node->farPlane = (float)atof( xmlNode->Attribute( "farPlane" ) );
		}
	}
	else if( strcmp( xmlNode->Value(), "Emitter" ) == 0 )
	{
		SGREmitterNode *node = new SGREmitterNode();
		baseNode = node;
		parseBaseAttributes( xmlNode, *node );

		if( xmlNode->Attribute( "material" ) != 0x0 )
		{
			node->matRes = TheResourceManager::instance().addResource(
				ResourceTypes::Material, xmlNode->Attribute( "material" ), 0, false );
		}
		if( xmlNode->Attribute( "effect" ) != 0x0 )
		{
			node->effectRes = TheResourceManager::instance().addResource(
				ResourceTypes::Effect, xmlNode->Attribute( "effect" ), 0, false );
		}

		if( xmlNode->Attribute( "maxCount" ) != 0x0 )
		{
			node->maxParticleCount = atoi( xmlNode->Attribute( "maxCount" ) );
		}
		if( xmlNode->Attribute( "respawnCount" ) != 0x0 )
		{
			node->respawnCount = atoi( xmlNode->Attribute( "respawnCount" ) );
		}
		if( xmlNode->Attribute( "delay" ) != 0x0 )
		{
			node->delay = (float)atof( xmlNode->Attribute( "delay" ) );
		}
		if( xmlNode->Attribute( "emissionRate" ) != 0x0 )
		{
			node->emissionRate = (float)atof( xmlNode->Attribute( "emissionRate" ) );
		}
		if( xmlNode->Attribute( "spreadAngle" ) != 0x0 )
		{
			node->spreadAngle = (float)atof( xmlNode->Attribute( "spreadAngle" ) );
		}
		if( xmlNode->Attribute( "forceX" ) != 0x0 )
		{
			node->fx = (float)atof( xmlNode->Attribute( "forceX" ) );
		}
		if( xmlNode->Attribute( "forceY" ) != 0x0 )
		{
			node->fy = (float)atof( xmlNode->Attribute( "forceY" ) );
		}
		if( xmlNode->Attribute( "forceZ" ) != 0x0 )
		{
			node->fz = (float)atof( xmlNode->Attribute( "forceZ" ) );
		}
	}

	if( baseNode != 0x0 )
	{
		if( parentNode != 0x0 )
		{
			parentNode->children.push_back( baseNode );
		}
		else
		{	
			delete _rootNode;	// Delete default root node
			_rootNode = baseNode;
		}
	
		// Parse children
		TiXmlElement *xmlNode1 = xmlNode->FirstChildElement();
		while( xmlNode1 != 0x0 )
		{	
			parseNode( xmlNode1, baseNode );

			xmlNode1 = xmlNode1->NextSiblingElement();
		}
	}
}


bool SceneGraphResource::load( const char *data, unsigned int size )
{
	if( !Resource::load( data, size ) ) return false;

	// Load XML file
	TiXmlDocument *xmlDoc = new TiXmlDocument();
	xmlDoc->SetCondenseWhiteSpace( false );
	
	xmlDoc->Parse( data );
	
	if( xmlDoc->Error() )
	{	
		TheEngineLog::instance().writeWarning( "Invalid XML syntax in scene graph file" );
		delete xmlDoc; xmlDoc = 0x0;
		return false;
	}

	// Parse scene nodes and load resources
	TiXmlElement *xmlRoot = xmlDoc->RootElement();
	if( xmlRoot != 0x0 )
	{	
		parseNode( xmlRoot, 0x0 );
	}

	xmlDoc->Clear();
	delete xmlDoc;

	return true;
}
