// *************************************************************************************************
//
// Horde3D
//   Next-Generation Graphics Engine
// --------------------------------------
// Copyright (C) 2006-2007 Nicolas Schulz
//
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//
// *************************************************************************************************

#pragma once

#include "egSingleton.h"
#include "utMath.h"
#include "egParticle.h"

class TiXmlElement;


// *************************************************************************************************
// Node Types
// *************************************************************************************************

namespace SGRNodeTypes
{
	enum List
	{
		Group,
		Reference,
		Model,
		Mesh,
		Joint,
		Light,
		Camera,
		Emitter
	};
}


struct SGRNode
{
	SGRNodeTypes::List	type;
	string				name;
	Vec3f				pos, rot, scale;
	string				attachmentParams;
	vector< SGRNode * >	children;


	SGRNode( SGRNodeTypes::List type ) :
		type( type ), scale( Vec3f( 1, 1, 1 ) )
	{
	}

	virtual ~SGRNode()
	{
		for( unsigned int i = 0; i < children.size(); ++i ) delete children[i];
	}
};


struct SGRGroupNode : public SGRNode
{
	float	minDist, maxDist;

	
	SGRGroupNode() : SGRNode( SGRNodeTypes::Group ),
		minDist( 0 ), maxDist( Math::MaxFloat )
	{
	}
};


struct SGRReferenceNode : public SGRNode
{
	unsigned int	sceneGraphRes;


	SGRReferenceNode() : SGRNode( SGRNodeTypes::Reference ),
		sceneGraphRes( 0 )
	{
	}

	~SGRReferenceNode()
	{
		if( sceneGraphRes != 0 )
			TheResourceManager::instance().removeResource( sceneGraphRes, false );
	}
};


struct SGRModelNode : public SGRNode
{
	unsigned int	geoRes;

	
	SGRModelNode() : SGRNode( SGRNodeTypes::Model ),
		geoRes( 0 )
	{
	}

	~SGRModelNode()
	{
		if( geoRes != 0 ) TheResourceManager::instance().removeResource( geoRes, false );
	}
};


struct SGRMeshNode : public SGRNode
{
	unsigned int	matRes;
	unsigned int	batchStart, batchCount;
	unsigned int	vertRStart, vertREnd;


	SGRMeshNode() : SGRNode( SGRNodeTypes::Mesh ),
		matRes( 0 ), batchStart( 0 ), batchCount( 0 ), vertRStart( 0 ), vertREnd( 0 )
	{
	}

	~SGRMeshNode()
	{
		if( matRes != 0 ) TheResourceManager::instance().removeResource( matRes, false );
	}
};


struct SGRJointNode : public SGRNode
{
	unsigned int	jointIndex;


	SGRJointNode() : SGRNode( SGRNodeTypes::Joint ),
		jointIndex( 0 )
	{
	}
};


struct SGRLightNode : public SGRNode
{
	unsigned int	matRes;
	string			lightingContext, shadowContext;
	float			radius, fov;
	float			col_R, col_G, col_B;
	unsigned int	shadowMapCount;
	float			shadowSplitLambda;
	float			shadowMapBias;


	SGRLightNode() : SGRNode( SGRNodeTypes::Light ),
		matRes( 0 ), radius( 0 ), fov( 90 ), col_R( 1 ), col_G( 1 ), col_B( 1 ),
		shadowMapCount( 0 ), shadowSplitLambda( 0.5f ), shadowMapBias( 0.005f )
	{
	}

	~SGRLightNode()
	{
		if( matRes != 0 ) TheResourceManager::instance().removeResource( matRes, false );
	}
};


struct SGRCameraNode : public SGRNode
{
	float	leftPlane, rightPlane;
	float	bottomPlane, topPlane;
	float	nearPlane, farPlane;

	SGRCameraNode() : SGRNode( SGRNodeTypes::Camera ),
		leftPlane( -0.055228457f ), rightPlane( 0.055228457f ),
		bottomPlane( -0.041421354f ), topPlane( 0.041421354f ),
		nearPlane( 0.1f ), farPlane( 1000.0f )
	{
		// Default params: fov=45, aspect=4/3
	}
};


struct SGREmitterNode : public SGRNode
{
	unsigned int	matRes;
	unsigned int	effectRes;
	unsigned int	maxParticleCount;
	int				respawnCount;
	float			delay, emissionRate, spreadAngle;
	float			fx, fy, fz;


	SGREmitterNode() : SGRNode( SGRNodeTypes::Emitter ),
		matRes( 0 ), effectRes( 0 ), maxParticleCount( 0 ), respawnCount( -1 ), delay( 0 ),
		emissionRate( 0 ), spreadAngle( 0 ), fx( 0 ), fy( 0 ), fz( 0 )
	{
	}
};


// *************************************************************************************************
// SceneGraphResource
// *************************************************************************************************

class SceneGraphResource : public Resource
{
private:

	SGRNode		*_rootNode;

	void parseBaseAttributes( TiXmlElement *xmlNode, SGRNode &node );
	void parseNode( TiXmlElement *node, SGRNode *parentNode );

public:

	static void initResType() {};
	static void releaseResType() {};
	
	SceneGraphResource( const string &name, unsigned int flags );
	~SceneGraphResource();
	
	void initDefault();
	void release();
	bool load( const char *data, unsigned int size );

	SGRNode *getRootNode() { return _rootNode; }

	friend class SceneManager;
};
