// *************************************************************************************************
//
// Horde3D
//   Next-Generation Graphics Engine
// --------------------------------------
// Copyright (C) 2006-2007 Nicolas Schulz
//
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//
// *************************************************************************************************

#include "egShader.h"
#include "egSingleton.h"
#include "tinyxml.h"
#include "utPlatform.h"

#include "utDebug.h"


// *************************************************************************************************
// CodeResource
// *************************************************************************************************

CodeResource::CodeResource( const string &name, unsigned int flags ) :
	Resource( ResourceTypes::Code, name, flags )
{
	initDefault();
}


CodeResource::~CodeResource()
{
	release();
}


void CodeResource::initDefault()
{
	_code = "";
}


void CodeResource::release()
{
}


bool CodeResource::load( const char *data, unsigned int size )
{
	if( !Resource::load( data, size ) ) return false;

	_code.reserve( size );
	_code = data;

	// Compile shaders that require this code block
	updateShaders();

	return true;
}


void CodeResource::updateShaders()
{
	for( unsigned int i = 0; i < TheResourceManager::instance().getResources().size(); ++i )
	{
		Resource *res = TheResourceManager::instance().getResources()[i];

		if( res != 0x0 && res->getType() == ResourceTypes::Shader )
		{
			((ShaderResource *)res)->compileShaders();
		}
	}
}


// *************************************************************************************************
// ShaderResource
// *************************************************************************************************

ShaderResource::ShaderResource( const string &name, unsigned int flags ) :
	Resource( ResourceTypes::Shader, name, flags )
{
	initDefault();
}


ShaderResource::~ShaderResource()
{
	release();
}


void ShaderResource::initDefault()
{
}


void ShaderResource::release()
{
	for( unsigned int i = 0; i < _contexts.size(); ++i )
	{
		TheRenderer::instance().unloadShader( _contexts[i].shaderObject );
	}

	_contexts.clear();
}


bool ShaderResource::raiseError( TiXmlDocument *doc )
{
	// Reset
	release();
	initDefault();

	TheEngineLog::instance().writeWarning( "Invalid XML syntax in shader file" );
	
	doc->Clear();
	delete doc;
	
	return false;
}


bool ShaderResource::parseCode( TiXmlElement *node, vector< ShaderCodeFract > &codeFracts )
{
	TiXmlElement *node1 = node->FirstChildElement();
	while( node1 != 0x0 )
	{
		if( strcmp( node1->Value(), "DefCode" ) == 0 )
		{
			ShaderCodeFract codeFract;
			codeFract.code = node1->GetText();
			codeFracts.push_back( codeFract );
		}
		else if( strcmp( node1->Value(), "InsCode" ) == 0 )
		{
			if( node1->Attribute( "code" ) == 0x0 ) return false;
		
			ShaderCodeFract codeFract;
			
			unsigned int id = TheResourceManager::instance().addResource(
				ResourceTypes::Code, node1->Attribute( "code" ), 0, false );

			Resource *res = TheResourceManager::instance().resolveResHandle( id );
			if( res != 0x0 )
			{	
				codeFract.refCodeRes = (CodeResource *)res;
				codeFracts.push_back( codeFract );
			}
		}

		node1 = node1->NextSiblingElement();
	}

	return true;
}


bool ShaderResource::load( const char *data, unsigned int size )
{
	if( !Resource::load( data, size ) ) return false;
	
	// Parse shader
	TiXmlDocument *doc = new TiXmlDocument();
	doc->SetCondenseWhiteSpace( false );

	doc->Parse( data );
	if( doc->Error() ) return raiseError( doc );

	TiXmlElement *root = doc->RootElement();
	if( root == 0x0 || strcmp( root->Value(), "Shader" ) != 0 ) return raiseError( doc );
	
	TiXmlElement *node1 = root->FirstChildElement( "Context" );
	while( node1 != 0x0 )
	{
		if( node1->Attribute( "id" ) == 0x0 ) return raiseError( doc );
		
		ShaderContext sc;

		sc.id = node1->Attribute( "id" );
		
		// Config
		TiXmlElement *node2 = node1->FirstChildElement( "RenderConfig" );
		if( node2 != 0x0 )
		{
			if( node2->Attribute( "writeDepth" ) != 0x0 )
			{
				if( _stricmp( node2->Attribute( "writeDepth" ), "false" ) == 0 )
					sc.writeDepth = false;
				else if( _stricmp( node2->Attribute( "writeDepth" ), "true" ) == 0 )
					sc.writeDepth = true;
			}
			if( node2->Attribute( "blendMode" ) != 0x0 )
			{
				if( _stricmp( node2->Attribute( "blendMode" ), "REPLACE" ) == 0 )
					sc.blendMode = BlendModes::Replace;
				else if( _stricmp( node2->Attribute( "blendMode" ), "BLEND" ) == 0 )
					sc.blendMode = BlendModes::Blend;
				else if( _stricmp( node2->Attribute( "blendMode" ), "ADD" ) == 0 )
					sc.blendMode = BlendModes::Add;
				else if( _stricmp( node2->Attribute( "blendMode" ), "ADD_BLENDED" ) == 0 )
					sc.blendMode = BlendModes::AddBlended;
			}
		}
		
		// Code 
		node2 = node1->FirstChildElement( "VertexShader" );
		if( node2 == 0x0 ) return raiseError( doc );
		if( !parseCode( node2, sc.vertShaderFracts ) ) return raiseError( doc );

		node2 = node1->FirstChildElement( "FragmentShader" );
		if( node2 == 0x0 ) return raiseError( doc );
		if( !parseCode( node2, sc.fragShaderFracts ) ) return raiseError( doc );

		_contexts.push_back( sc );
		
		node1 = node1->NextSiblingElement( "Context" );
	}

	doc->Clear();
	delete doc;

	compileShaders();
	
	return true;
}


void ShaderResource::compileShaders()
{
	string vertCode, fragCode;
	
	for( unsigned int i = 0; i < _contexts.size(); ++i )
	{
		ShaderContext &sc = _contexts[i];

		if( !sc.compiled )
		{
			// Check if all referenced code blocks are available
			bool allAvailbale = true;
			for( unsigned int j = 0; j < sc.vertShaderFracts.size(); ++j )
			{
				if( sc.vertShaderFracts[j].refCodeRes != 0x0 && !sc.vertShaderFracts[j].refCodeRes->isLoaded() )
				{
					allAvailbale = false;
					break;
				}
			}
			if( !allAvailbale ) continue;
			
			for( unsigned int j = 0; j < sc.fragShaderFracts.size(); ++j )
			{
				if( sc.fragShaderFracts[j].refCodeRes != 0x0 && !sc.fragShaderFracts[j].refCodeRes->isLoaded() )
				{
					allAvailbale = false;
					break;
				}
			}
			if( !allAvailbale ) continue;
			
			// Compile shader
			vertCode = ""; fragCode = "";
			
			for( unsigned int j = 0; j < sc.vertShaderFracts.size(); ++j )
			{
				if( !sc.vertShaderFracts[j].refCodeRes != 0x0 )
					vertCode += "\n\n" + sc.vertShaderFracts[j].code;
				else
					vertCode += "\n\n" + sc.vertShaderFracts[j].refCodeRes->getCode();
			}
			
			for( unsigned int j = 0; j < sc.fragShaderFracts.size(); ++j )
			{
				if( !sc.fragShaderFracts[j].refCodeRes != 0x0 )
					fragCode += "\n\n" + sc.fragShaderFracts[j].code;
				else
					fragCode += "\n\n" + sc.fragShaderFracts[j].refCodeRes->getCode();
			}
			
			TheEngineLog::instance().writeInfo(
				"Compiling shader context <b>" + _name + "--->" + sc.id + "</b>" );
			
			sc.shaderObject = TheRenderer::instance().loadShader( vertCode.c_str(), fragCode.c_str(), sc );

			if( TheRenderer::instance().getShaderLog() != "" )
				TheEngineLog::instance().writeWarning( TheRenderer::instance().getShaderLog() );

			if( sc.shaderObject == 0 )
				TheEngineLog::instance().writeError( "Failed compiling shader" );

			sc.compiled = true;
		}
	}	
}
