// *************************************************************************************************
//
// Horde3D
//   Next-Generation Graphics Engine
// --------------------------------------
// Copyright (C) 2006-2007 Nicolas Schulz
//
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//
// *************************************************************************************************

#include "egTextures.h"
#include "egSingleton.h"
#include "corona.h"

#include "utDebug.h"


// *************************************************************************************************
// Class ImageLoader
// *************************************************************************************************

unsigned int ImageLoader::nextSmallerPowerOfTwo( unsigned int n )
{
	while( (n & (n - 1)) != 0 ) --n;

	return n;
}


bool ImageLoader::loadImage( const char *data, unsigned int size, bool makePOT,
							 unsigned char *&pixels, int &width, int &height, int &bpp )
{
	// Uses Corona Image Library
	// Corona has support for BMP, GIF, JPEG, PCX, PNG, TGA

	corona::File *file = corona::CreateMemoryFile( data, size );
	corona::Image *image = corona::OpenImage( file, corona::PF_B8G8R8A8 );
	corona::FlipImage( image, corona::CA_X );
	
	if( !image )
	{
		delete file;
		delete image;
		
		return false;
	}

	width  = image->getWidth();
	height = image->getHeight();
	bpp = 32;

	bool scaleImage = false;

	if( makePOT )
	{
		// Convert to a POT image
		int newWidth = nextSmallerPowerOfTwo( width );
		int newHeight = nextSmallerPowerOfTwo( height );

		if( newWidth != width || newHeight != height )
		{
			scaleImage = true;
			
			pixels = new unsigned char[newWidth * newHeight * 4];
			unsigned char *p = (unsigned char *)image->getPixels();

			for( int y = 0; y < newHeight; ++y )
			{
				int yy = (int)(y * ((float)height / newHeight) + 0.5f);
				
				for( int x = 0; x < newWidth; ++x )
				{
					int xx = (int)(x * ((float)width / newWidth) + 0.5f);
					
					pixels[(y * newWidth + x) * 4 + 0] = p[(yy * width + xx) * 4 + 0];
					pixels[(y * newWidth + x) * 4 + 1] = p[(yy * width + xx) * 4 + 1];
					pixels[(y * newWidth + x) * 4 + 2] = p[(yy * width + xx) * 4 + 2];
					pixels[(y * newWidth + x) * 4 + 3] = p[(yy * width + xx) * 4 + 3];
				}
			}

			width = newWidth;
			height = newHeight;
		}
	}

	if( !scaleImage )
	{
		pixels = new unsigned char[width * height * 4];
		memcpy( pixels, image->getPixels(), width * height * 4 );
	}

	delete file;
	delete image;

	return true;
}


// *************************************************************************************************
// Class Texture2DResource
// *************************************************************************************************

unsigned int Texture2DResource::defTexObject = 0;


void Texture2DResource::initResType()
{
	unsigned char defaultTexture[] = 
		{	128,192,255, 128,192,255, 128,192,255, 128,192,255,
			128,192,255, 128,192,255, 128,192,255, 128,192,255,
			128,192,255, 128,192,255, 128,192,255, 128,192,255,
			128,192,255, 128,192,255, 128,192,255, 128,192,255	};

	// Upload default texture
	defTexObject = TheRenderer::instance().uploadTexture2D( defaultTexture, 4, 4, 24, false );
}


void Texture2DResource::releaseResType()
{
	TheRenderer::instance().unloadTexture( defTexObject, false );
}


Texture2DResource::Texture2DResource( const string &name, unsigned int flags ) :
	Resource( ResourceTypes::Texture2D, name, flags )
{
	initDefault();

	if( (flags & ResourceFlags::NoCompression) == ResourceFlags::NoCompression )
		_allowCompression = false;
	else
		_allowCompression = true;
}


Texture2DResource::~Texture2DResource()
{
	release();
}


void Texture2DResource::initDefault()
{
	_pixels = 0x0;
	_width = 0; _height = 0; _bpp = 0;
	_texObject = defTexObject;
}


void Texture2DResource::release()
{
	delete _pixels; _pixels = 0x0;
	
	if( _texObject != 0 && _texObject != defTexObject )
		TheRenderer::instance().unloadTexture( _texObject, false );

	_texObject = 0;
}


bool Texture2DResource::load( const char *data, unsigned int size )
{
	if( !Resource::load( data, size ) ) return false;
	
	if( !TheEngineConfig::instance().loadTextures )
		return true;

	bool makePOT = !TheRenderer::instance().supportsNPOTTextures();
	
	if( !ImageLoader::loadImage( data, size, makePOT, _pixels, _width, _height, _bpp ) )
	{	
		// Reset
		release();
		initDefault();

		TheEngineLog::instance().writeWarning( "Invalid image format" );
		return false;
	}

	// Check color depth
	if( _bpp != 24 && _bpp != 32 )
	{	
		// Reset
		release();
		initDefault();

		TheEngineLog::instance().writeWarning( "Invalid color depth" );
		return false;
	}
	
	// Upload texture
	_texObject = TheRenderer::instance().uploadTexture2D( _pixels, _width, _height, _bpp, _allowCompression );
	if( _texObject == 0 )
	{
		// Reset
		release();
		initDefault();

		TheEngineLog::instance().writeWarning( "Couldn't upload texture map" );
		return false;
	}

	delete[] _pixels; _pixels = 0x0;

	return true;
}


bool Texture2DResource::updateData( ResourceData::List param, const void *data, unsigned int size )
{
	if( Resource::updateData( param, data, size ) ) return true;
	
	if( param == ResourceData::Tex2DPixelData )
	{
		if( _texObject == 0 ) return false;
		if( size != _width * _height * (_bpp / 8) ) return false;

		TheRenderer::instance().updateTexture2D( (unsigned char *)data, _width, _height, _bpp, _texObject );

		return true;
	}

	return false;
}


// *************************************************************************************************
// Class TextureCubeResource
// *************************************************************************************************

unsigned int TextureCubeResource::defTexObject = 0;


void TextureCubeResource::initResType()
{
	unsigned char defaultTexture[] = 
		{	128,192,255, 128,192,255, 128,192,255, 128,192,255,
			128,192,255, 128,192,255, 128,192,255, 128,192,255,
			128,192,255, 128,192,255, 128,192,255, 128,192,255,
			128,192,255, 128,192,255, 128,192,255, 128,192,255	};

	// Upload default texture
	defTexObject = TheRenderer::instance().uploadTextureCube( defaultTexture, 4, 4, 24, 0, false );
	for( unsigned int j = 1; j < 6; j++ ) 
	{
		TheRenderer::instance().uploadTextureCube( defaultTexture, 4, 4, 24, j, false, defTexObject );
	}
}


void TextureCubeResource::releaseResType()
{
	TheRenderer::instance().unloadTexture( defTexObject, true );
}


TextureCubeResource::TextureCubeResource( const string &name, unsigned int flags ) :
	Resource( ResourceTypes::TextureCube, name, flags )
{
	initDefault();

	if( (flags & ResourceFlags::NoCompression) == ResourceFlags::NoCompression )
		_allowCompression = false;
	else
		_allowCompression = true;
}


TextureCubeResource::~TextureCubeResource()
{
	release();
}


void TextureCubeResource::initDefault()
{
	_pixels = 0x0;
	_width = 0; _height = 0; _bpp = 0;
	_texObject = defTexObject;
}


void TextureCubeResource::release()
{
	delete _pixels; _pixels = 0x0;
	
	if( _texObject != 0 && _texObject != defTexObject )
		TheRenderer::instance().unloadTexture( _texObject, true );

	_texObject = 0;
}


bool TextureCubeResource::load( const char *data, unsigned int size )
{
	if( !Resource::load( data, size ) ) return false;
	
	if( !TheEngineConfig::instance().loadTextures )
		return true;
	
	if( !ImageLoader::loadImage( data, size, false, _pixels, _width, _height, _bpp ) )
	{	
		// Reset
		release();
		initDefault();

		TheEngineLog::instance().writeWarning( "Invalid image format" );
		return false;
	}

	// Check color depth
	if( _bpp != 24 && _bpp != 32 )
	{	
		// Reset
		release();
		initDefault();

		TheEngineLog::instance().writeWarning( "Invalid color depth" );
		return false;
	}

	// Check image size
	if( (_width != 16	&& _width != 32		&& _width != 64		&& _width != 128	&&
		 _width != 256	&& _width != 512	&& _width != 1024	&& _width != 2048 &&
		 _width != 4096	&& _width != 8192)	|| (_height * 4 != _width * 3) ) 
	{
		// Reset
		release();
		initDefault();

		TheEngineLog::instance().writeWarning( "Invalid cubemap texture size" );
		return false;
	}

	unsigned int	faceWidth = _width / 4, faceHeight = _height / 3, faceBytes = _bpp / 8;
	unsigned char	*cubeFace = new unsigned char[faceWidth * faceHeight * faceBytes];
	unsigned int	x0, y0, x1, y1, c;

	const unsigned int beginY[6] = {faceHeight * 1, faceHeight * 1, faceHeight * 0,
									faceHeight * 2, faceHeight * 1, faceHeight * 1};
	const unsigned int endY[6] = {faceHeight * 2, faceHeight * 2, faceHeight * 1,
								  faceHeight * 3, faceHeight * 2, faceHeight * 2};
	const unsigned int beginX[6] = {faceWidth * 2, faceWidth * 0, faceWidth * 1,
									faceWidth * 1, faceWidth * 1, faceWidth * 3};
	const unsigned int endX[6] = {faceWidth * 3, faceWidth * 1, faceWidth * 2,
								  faceWidth * 2, faceWidth * 2, faceWidth * 4};
	
	// Extract and upload cube faces
	for( unsigned int i = 0; i < 6; ++i ) 
	{
		for( y0 = beginY[i], y1 = 0; y0 < endY[i]; ++y0, ++y1 ) 
		{
			for( x0 = beginX[i], x1 = 0; x0 < endX[i]; ++x0, ++x1 ) 
			{
				for( c = 0; c < faceBytes; ++c ) 
				{
					cubeFace[(y1 * faceWidth + x1) * faceBytes + c] = 
						_pixels[(y0 * _width + x0) * faceBytes + c];
				}
			}
		}
		
		if( i == 0 )
		{
			_texObject = TheRenderer::instance().uploadTextureCube(
					cubeFace, faceWidth, faceHeight, _bpp, i, _allowCompression );
		}
		else
		{
			TheRenderer::instance().uploadTextureCube(
					cubeFace, faceWidth, faceHeight, _bpp, i, _allowCompression, _texObject );
		}
	}
	
	delete[] cubeFace;
	delete[] _pixels; _pixels = 0x0;

	return true;
}
