// *************************************************************************************************
//
// Horde3D
//   Next-Generation Graphics Engine
// --------------------------------------
// Copyright (C) 2006-2007 Nicolas Schulz
//
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//
// *************************************************************************************************

// TODO: Make OpenGL functions platform compatible

#include "Horde3D.h"
#include "utPlatform.h"
#include <math.h>
#include <string>
#include <fstream>
using namespace std;

#ifndef DLLEXP
#	ifdef PLATFORM_WIN
#		include <windows.h>
#		define DLLEXP extern "C" __declspec( dllexport )
#	else
#		define DLLEXP
#	endif
#endif

#include <GL/gl.h>


// *************************************************************************************************
// Utility functions
// *************************************************************************************************

namespace Horde3DUtils
{
	int			_hDC = 0, _hRC = 0;
	ofstream	_outf;


	DLLEXP bool initOpenGL( int hDC )
	{
		#ifdef PLATFORM_WIN
		_hDC = hDC;
		
		// Init OpenGL rendering context
		int pixelFormat;

		static PIXELFORMATDESCRIPTOR pfd = 
		{
			sizeof( PIXELFORMATDESCRIPTOR ),			// Size of this pixel format descriptor
			1,											// Version number
			PFD_DRAW_TO_WINDOW |						// Format must support window
			PFD_SUPPORT_OPENGL |						// Format must support OpenGL
			PFD_DOUBLEBUFFER,							// Must support double buffering
			PFD_TYPE_RGBA,								// Request a RGBA format
			32,											// Select our color depth
			0, 0, 0, 0, 0, 0,							// Color bits ignored
			8,											// 8Bit alpha buffer
			0,											// Shift bit ignored
			0,											// No accumulation buffer
			0, 0, 0, 0,									// Accumulation bits ignored
			32,											// 32Bit z-buffer (depth buffer)  
			8,											// 8Bit stencil buffer
			0,											// No auxiliary buffer
			PFD_MAIN_PLANE,								// Main drawing layer
			0,											// Reserved
			0, 0, 0										// Layer masks ignored
		};

		if( !(pixelFormat = ChoosePixelFormat( (HDC)_hDC, &pfd )) ) 
		{
			return false;
		}

		if( !SetPixelFormat( (HDC)_hDC, pixelFormat, &pfd ) ) 
		{
			return false;
		}

		if( !(_hRC = (int)wglCreateContext( (HDC)_hDC )) ) 
		{
			return false;
		}

		if( !wglMakeCurrent( (HDC)_hDC, (HGLRC)_hRC) ) 
		{
			wglDeleteContext( (HGLRC)_hRC );
			return false;
		}

		return true;
		
		#else
		return false;
		#endif
	}


	DLLEXP void releaseOpenGL()
	{
		#ifdef PLATFORM_WIN
		if( _hDC == 0 || _hRC == 0 ) return;

		if( !wglMakeCurrent( 0x0, 0x0 ) ) 
		{
			return;
		}
		if( !wglDeleteContext( (HGLRC)_hRC ) ) 
		{
			return;
		}
		_hRC = 0;
		#endif
	}


	DLLEXP void swapBuffers()
	{
		#ifdef PLATFORM_WIN
		if( _hDC == 0 || _hRC == 0 ) return;

		SwapBuffers( (HDC)_hDC );
		#endif
	}


	DLLEXP bool loadResourcesFromDisk( const char *contentDir )
	{
		bool result = true;
		
		string name;
		string fileName;
		string cDir;

		cDir = contentDir;
		if( cDir != "" ) cDir = cDir + "/";
		
		// Get the first resource that needs to be loaded
		name = Horde3D::queryUnloadedResource();
		
		while( name != "" )
		{
			// Build filename of resource by adding the directory
			fileName = cDir + name;

			// Open resource file
			ifstream inf( fileName.c_str(), ios::binary );
			if( inf ) // Resource file found
			{
				// Find size of resource file
				inf.seekg( 0, ios::end );
				int size = inf.tellg();
				// Copy resource file to memory
				char *data = new char[size + 1];
				inf.seekg( 0 );
				inf.read( data, size );
				inf.close();
				// Null-terminate buffer - this is important for XML parsers 
				data[size] = '\0';
				// Send resource data to engine
				Horde3D::loadResource( name.c_str(), data, size );
				delete[] data;
			}
			else // Resource file not found
			{
				// Tell engine to use the dafault resource by using NULL as data pointer
				Horde3D::loadResource( name.c_str(), 0x0, 0 );
				result = false;
			}
			// Get next unloaded resource
			name = Horde3D::queryUnloadedResource();
		}

		return result;
	}


	DLLEXP bool dumpMessages()
	{
		if( !_outf.is_open() )
		{
			// Reset log file
			_outf.setf( ios::fixed );
			_outf.precision( 3 );
			_outf.open( "EngineLog.html", ios::out );
			if( !_outf ) return false;

			_outf << "<style type=\"text/css\">\n";
			
			_outf << "body, html {\n";
			_outf << "background: #000000;\n";
			_outf << "width: 1000px;\n";
			_outf << "font-family: Arial;\n";
			_outf << "font-size: 16px;\n";
			_outf << "color: #C0C0C0;\n";
			_outf << "}\n";

			_outf << "h1 {\n";
			_outf << "color : #FFFFFF;\n";
			_outf << "border-bottom : 1px dotted #888888;\n";
			_outf << "}\n";

			_outf << ".box {\n";
			_outf << "border : 1px dotted #818286;\n";
			_outf << "padding : 5px;\n";
			_outf << "margin: 5px;\n";
			_outf << "width: 950px;\n";
			_outf << "background-color : #292929;\n";
			_outf << "}\n";

			_outf << ".inf {\n";
			_outf << "color: #C0C0C0;\n";
			_outf << "}\n";

			_outf << ".err {\n";
			_outf << "color: #EE1100;\n";
			_outf << "font-weight: bold\n";
			_outf << "}\n";

			_outf << ".war {\n";
			_outf << "color: #FFCC00;\n";
			_outf << "font-weight: bold\n";
			_outf << "}\n";

			_outf << "</style>\n\n";

			_outf << "<h1>Horde3D Log</h1>\n";
			_outf << "<h3>" << Horde3D::getVersionString() << "</h3>\n";
			_outf << "<div class=\"box\">\n";
			_outf << "<table>\n";

			_outf.flush();
		}

		unsigned int level;
		float time;
		string text = Horde3D::getMessage( &level, &time );
		
		while( text != "" )
		{
			_outf << "<tr>\n";
			_outf << "<td width=\"100\">";
			_outf << time;
			_outf << "</td>\n";
			_outf << "<td class=\"";
			
			switch( level )
			{
			case 1:
				_outf << "war";
				break;
			case 2:
				_outf << "err";
				break;
			default:
				_outf << "inf";
			}
			
			_outf << "\">";
			_outf << text.c_str();
			_outf << "</td>\n";
			_outf << "</tr>\n";

			_outf.flush();
			
			text = Horde3D::getMessage( &level, &time );
		}
		
		return true;
	}


	DLLEXP void showText( const char *text, float x, float y, float size,
						  unsigned int layer, unsigned int fontMatRes )
	{
		int pos = 0;
		
		do
		{
			unsigned char ch = (unsigned char)*text++;

			float u0 = 0.0625f * (ch % 16);
			float v0 = 1.0f - 0.0625f * (ch / 16);

			Horde3D::showOverlay( x + size * pos, y, u0, v0 - 0.0625f,
								  x + size * pos + size, y, u0 + 0.0625f,v0 - 0.0625f,
								  x + size * pos + size, y + size, u0 + 0.0625f, v0,
								  x + size * pos, y + size, u0, v0,
								  layer, fontMatRes );
			++pos;
		} while ( *text );
	}


	DLLEXP void freeMem( char **ptr )
	{
		delete[] *ptr; *ptr = 0x0;
	}

	
	DLLEXP bool createTGAImage( const unsigned char *pixels, unsigned int width, unsigned int height,
								unsigned int bpp, char **outData, unsigned int *outSize )
	{
		*outData = 0x0; *outSize = 0;
		
		if( bpp != 24 && bpp != 32 ) return false;

		*outSize = width * height * (bpp / 8) + 18;
		char *data = new char[*outSize];
		*outData = data;

		// Build TGA header
		char c;
		short s;
		c = 0;		memcpy( data, &c, 1 ); data += 1;	// idLength
		c = 0;		memcpy( data, &c, 1 ); data += 1;	// colmapType
		c = 2;		memcpy( data, &c, 1 ); data += 1;	// imageType
		s = 0;		memcpy( data, &s, 2 ); data += 2;	// colmapStart
		s = 0;		memcpy( data, &s, 2 ); data += 2;	// colmapLength
		c = 16;		memcpy( data, &c, 1 ); data += 1;	// colmapBits
		s = 0;		memcpy( data, &s, 2 ); data += 2;	// x
		s = 0;		memcpy( data, &s, 2 ); data += 2;	// y
		s = width;	memcpy( data, &s, 2 ); data += 2;	// width
		s = height;	memcpy( data, &s, 2 ); data += 2;	// height
		c = bpp;	memcpy( data, &c, 1 ); data += 1;	// bpp
		c = 0;		memcpy( data, &c, 1 ); data += 1;	// imageDesc

		// Copy data
		memcpy( data, pixels, width * height * (bpp / 8) );

		return true;
	}
}


// *************************************************************************************************
// DLL entry point
// *************************************************************************************************

#ifdef PLATFORM_WIN
BOOL APIENTRY DllMain( HANDLE hModule, DWORD ul_reason_for_call, LPVOID lpReserved )
{
   switch( ul_reason_for_call )
	{
	case DLL_PROCESS_DETACH:
		// Close log file
		if( Horde3DUtils::_outf.is_open() )
		{
			Horde3DUtils::_outf << "</table>\n";
			Horde3DUtils::_outf << "</div>\n";
			Horde3DUtils::_outf.close();
		}
	break;
	}
	
	return TRUE;
}
#endif
