// *************************************************************************************************
//
// Horde3D
//   Next-Generation Graphics Engine
// --------------------------------------
// Copyright (C) 2006-2007 Nicolas Schulz
//
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//
// *************************************************************************************************

// -------------------------------------------------------------------------------------------------
//
// Math library
//
// Conventions:
//
// - Coordinate system is right-handed with positive y as up axis
// - All rotation angles are counter-clockwise when looking from the positive end of the rotation
//	 axis towards the origin
// - An unrotated view vector points along the negative z-axis
//
// -------------------------------------------------------------------------------------------------

#pragma once

#include <math.h>


// Constants
namespace Math
{
	const long MaxInt32 = 0xFFFFFFFF;
	const float MaxFloat = 3.402823466e+38F;
	
	const float Pi = 3.1415926f;
	const float TwoPi = Pi * 2;
	const float PiHalf = Pi / 2;

	const float Epsilon = 0.000001f;
	const float NaN = *(float *)&MaxInt32;
};


// -------------------------------------------------------------------------------------------------
// General
// -------------------------------------------------------------------------------------------------

inline float degToRad( float f ) 
{
	return f * (Math::Pi / 180.0f);
}

inline float radToDeg( float f ) 
{
	return f * (180.0f / Math::Pi);
}

inline float clamp( float f, float min, float max )
{
	if( f < min ) f = min;
	if( f > max ) f = max;

	return f;
}

inline float minf( float a, float b )
{
	if( a < b ) return a;
	else return b;
}

inline float maxf( float a, float b )
{
	if( a > b ) return a;
	else return b;
}


// -------------------------------------------------------------------------------------------------
// Vector
// -------------------------------------------------------------------------------------------------

class Vec3f
{
public:
	float x, y, z;
	
	
	// ------------
	// Constructors
	// ------------
	Vec3f() : x( 0.0f ), y( 0.0f ), z( 0.0f ) 
	{ 
	}

	explicit Vec3f( const float x, const float y, const float z ) : x( x ), y( y ), z( z ) 
	{
	}
	
	// -----------
	// Comparisons
	// -----------
	bool operator==( const Vec3f &v ) const
	{
		return (x > v.x - Math::Epsilon && x < v.x + Math::Epsilon && 
				y > v.y - Math::Epsilon && y < v.y + Math::Epsilon &&
				z > v.z - Math::Epsilon && z < v.z + Math::Epsilon);
		
	}

	bool operator!=( const Vec3f &v ) const
	{
		return (x < v.x - Math::Epsilon || x > v.x + Math::Epsilon || 
				y < v.y - Math::Epsilon || y > v.y + Math::Epsilon ||
				z < v.z - Math::Epsilon || z > v.z + Math::Epsilon);
		
	}
	
	// ---------------------
	// Artitmetic operations
	// ---------------------
	Vec3f operator-() const
	{
		return Vec3f( -x, -y, -z );
	}

	Vec3f operator+( const Vec3f &v ) const
	{
		return Vec3f( x + v.x, y + v.y, z + v.z );
	}

	void operator+=( const Vec3f &v )
	{
		*this = *this + v;
	}

	Vec3f operator-( const Vec3f &v ) const 
	{
		return Vec3f( x - v.x, y - v.y, z - v.z );
	}

	void operator-=( const Vec3f &v )
	{
		*this = *this - v;
	}

	Vec3f operator*( const float f ) const
	{
		return Vec3f( x * f, y * f, z * f );
	}

	void operator*=( const float f )
	{
		*this = *this * f;
	}

	Vec3f operator/( const float f ) const
	{
		return Vec3f( x / f, y / f, z / f );
	}

	void operator/=( const float f )
	{
		*this = *this / f;
	}

	// -----------
	// Dot product
	// -----------
	float operator*( const Vec3f &v ) const
	{
		return ( x * v.x + y * v.y + z * v.z );
	}

	// -------------
	// Cross Product
	// -------------
	Vec3f crossProduct( const Vec3f &v ) const
	{
		return Vec3f( y * v.z - z * v.y, z * v.x - x * v.z, x * v.y - y * v.x );
	}

	// ----------------
	// Other operations
	// ----------------
	float length() const 
	{
		return sqrtf( *this * *this );
	}

	Vec3f normalized() const
	{
		float l = length();

		if( l != 0 ) return Vec3f( x / l, y / l, z / l );
		else return Vec3f( 0, 0, 0 );
	}

	/*void fromRotation( float angleX, float angleY )
	{
		x = cosf( angleX ) * sinf( angleY ); 
		y = -sinf( angleX );
		z = cosf( angleX ) * cosf( angleY );
	}*/

	Vec3f toRotation() const
	{
		// Assumes that the unrotated view vector is (0, 0, -1)
		Vec3f v;
		
		if( y != 0 ) v.x = atan2( y, sqrtf( x*x + z*z ) );
		if( z != 0 ) v.y = atan2( -x, -z );

		return v;
	}

	Vec3f lerp( const Vec3f &v, float f ) const
	{
		return Vec3f( x + (v.x - x) * f, y + (v.y - y) * f, z + (v.z - z) * f ); 
	}
};


class Vec4f
{
public:
	
	float x, y, z, w;


	Vec4f() : x( 0 ), y( 0 ), z( 0 ), w( 0 )
	{
	}
	
	explicit Vec4f( const float x, const float y, const float z, const float w )
		: x( x ), y( y ), z( z ), w( w )
	{
	}

	explicit Vec4f( Vec3f v ) : x( v.x ), y( v.y ), z( v.z ), w( 1.0f )
	{
	}
};


// -------------------------------------------------------------------------------------------------
// Quaternion
// -------------------------------------------------------------------------------------------------

class Quaternion
{
public:	
	Vec3f v;
	float w;

	// ------------
	// Constructors
	// ------------
	Quaternion() : v( 0.0f, 0.0f, 0.0f ), w( 0.0f ) 
	{ 
	}
	
	explicit Quaternion( const Vec3f v, const float w ) : v( v ), w( w )
	{
	}
	
	Quaternion( const float eulerX, const float eulerY, const float eulerZ )
	{
		Quaternion roll( Vec3f( sinf( eulerX / 2 ), 0, 0 ), cosf( eulerX / 2 ) );
		Quaternion pitch( Vec3f( 0, sinf( eulerY / 2 ), 0 ), cosf( eulerY / 2 ) );
		Quaternion yaw( Vec3f( 0, 0, sinf( eulerZ / 2 ) ), cosf( eulerZ / 2 ) );
	
		// Order: y * x * z
		*this = pitch * roll * yaw;
	}

	// ---------------------
	// Artitmetic operations
	// ---------------------
	Quaternion operator*( const Quaternion &q ) const
	{
		return Quaternion( v.crossProduct( q.v ) + q.v * w + v * q.w, w * q.w - v * q.v );
	}

	Quaternion operator*=( const Quaternion &q )
	{
		*this = *this * q;
	}

	// ----------------
	// Other operations
	// ----------------

	Quaternion slerp( const Quaternion &q, const float t ) const
	{
		float		omega, cosom, sinom, scale0, scale1;
		Quaternion	q1 = q, res;

        // Calc cosine
        cosom = v.x * q.v.x + v.y * q.v.y + v.z * q.v.z + w * q.w;

        // Adjust signs (if necessary)
        if( cosom < 0 ) {
			cosom = -cosom; 
			q1.v *= -1;
			q1.w = -q.w;
        } 

        // Calculate coefficients
		if( (1 - cosom) > Math::Epsilon ) {
			// Standard case (Slerp)
			omega = acosf( cosom );
			sinom = sinf( omega );
			scale0 = sinf( (1 - t) * omega ) / sinom;
			scale1 = sinf( t * omega ) / sinom;
		} 
		else {        
			// Quaternions very close, so do linear interpolation
			scale0 = 1 - t;
			scale1 = t;
		}
		
		// Calculate final values
		res.v = v * scale0 + q1.v * scale1;
		res.w = w * scale0 + q1.w * scale1;

		return res;
	}
};


// -------------------------------------------------------------------------------------------------
// Matrix
// -------------------------------------------------------------------------------------------------

class Matrix4f
{
private:

	Matrix4f( bool noInit )
	{
		// Don't initialize the matrix
	}

public:
	union
	{
		float c[4][4];	// Column major order for OpenGL: c[column][row]
		float x[16];
	};

	// ------------
	// Constructors
	// ------------
	Matrix4f()
	{
		c[0][0] = 1; c[1][0] = 0; c[2][0] = 0; c[3][0] = 0;
		c[0][1] = 0; c[1][1] = 1; c[2][1] = 0; c[3][1] = 0;
		c[0][2] = 0; c[1][2] = 0; c[2][2] = 1; c[3][2] = 0;
		c[0][3] = 0; c[1][3] = 0; c[2][3] = 0; c[3][3] = 1;
	}

	Matrix4f( const float *floatArray16 )
	{
		for( unsigned int i = 0; i < 4; ++i )
		{
			for( unsigned int j = 0; j < 4; ++j )
			{
				c[i][j] = floatArray16[i * 4 + j];
			}
		}
	}

	Matrix4f( const Quaternion &q )
	{
		float wx, wy, wz, xx, yy, yz, xy, xz, zz, x2, y2, z2;

		// Calculate coefficients
		x2 = q.v.x + q.v.x;	y2 = q.v.y + q.v.y;	z2 = q.v.z + q.v.z;
		xx = q.v.x * x2;	xy = q.v.x * y2;	xz = q.v.x * z2;
		yy = q.v.y * y2;	yz = q.v.y * z2;	zz = q.v.z * z2;
		wx = q.w * x2;		wy = q.w * y2;		wz = q.w * z2;


		c[0][0] = 1 - (yy + zz);	c[1][0] = xy - wz;	
		c[2][0] = xz + wy;			c[3][0] = 0;
		c[0][1] = xy + wz;			c[1][1] = 1 - (xx + zz);
		c[2][1] = yz - wx;			c[3][1] = 0;
		c[0][2] = xz - wy;			c[1][2] = yz + wx;
		c[2][2] = 1 - (xx + yy);	c[3][2] = 0;
		c[0][3] = 0;				c[1][3] = 0;
		c[2][3] = 0;				c[3][3] = 1;
	}

	// ---------------------
	// Matrix multiplication
	// ---------------------
	Matrix4f operator*( const Matrix4f &m ) const 
	{
		Matrix4f mf( false );
		
		mf.x[0] = x[0] * m.x[0] + x[4] * m.x[1] + x[8] * m.x[2] + x[12] * m.x[3];
		mf.x[1] = x[1] * m.x[0] + x[5] * m.x[1] + x[9] * m.x[2] + x[13] * m.x[3];
		mf.x[2] = x[2] * m.x[0] + x[6] * m.x[1] + x[10] * m.x[2] + x[14] * m.x[3];
		mf.x[3] = x[3] * m.x[0] + x[7] * m.x[1] + x[11] * m.x[2] + x[15] * m.x[3];

		mf.x[4] = x[0] * m.x[4] + x[4] * m.x[5] + x[8] * m.x[6] + x[12] * m.x[7];
		mf.x[5] = x[1] * m.x[4] + x[5] * m.x[5] + x[9] * m.x[6] + x[13] * m.x[7];
		mf.x[6] = x[2] * m.x[4] + x[6] * m.x[5] + x[10] * m.x[6] + x[14] * m.x[7];
		mf.x[7] = x[3] * m.x[4] + x[7] * m.x[5] + x[11] * m.x[6] + x[15] * m.x[7];

		mf.x[8] = x[0] * m.x[8] + x[4] * m.x[9] + x[8] * m.x[10] + x[12] * m.x[11];
		mf.x[9] = x[1] * m.x[8] + x[5] * m.x[9] + x[9] * m.x[10] + x[13] * m.x[11];
		mf.x[10] = x[2] * m.x[8] + x[6] * m.x[9] + x[10] * m.x[10] + x[14] * m.x[11];
		mf.x[11] = x[3] * m.x[8] + x[7] * m.x[9] + x[11] * m.x[10] + x[15] * m.x[11];

		mf.x[12] = x[0] * m.x[12] + x[4] * m.x[13] + x[8] * m.x[14] + x[12] * m.x[15];
		mf.x[13] = x[1] * m.x[12] + x[5] * m.x[13] + x[9] * m.x[14] + x[13] * m.x[15];
		mf.x[14] = x[2] * m.x[12] + x[6] * m.x[13] + x[10] * m.x[14] + x[14] * m.x[15];
		mf.x[15] = x[3] * m.x[12] + x[7] * m.x[13] + x[11] * m.x[14] + x[15] * m.x[15];

		return mf;
	}

	void fastMult( const Matrix4f &m1, const Matrix4f &m2 )
	{
		x[0] = m1.x[0] * m2.x[0] + m1.x[4] * m2.x[1] + m1.x[8] * m2.x[2] + m1.x[12] * m2.x[3];
		x[1] = m1.x[1] * m2.x[0] + m1.x[5] * m2.x[1] + m1.x[9] * m2.x[2] + m1.x[13] * m2.x[3];
		x[2] = m1.x[2] * m2.x[0] + m1.x[6] * m2.x[1] + m1.x[10] * m2.x[2] + m1.x[14] * m2.x[3];
		x[3] = m1.x[3] * m2.x[0] + m1.x[7] * m2.x[1] + m1.x[11] * m2.x[2] + m1.x[15] * m2.x[3];

		x[4] = m1.x[0] * m2.x[4] + m1.x[4] * m2.x[5] + m1.x[8] * m2.x[6] + m1.x[12] * m2.x[7];
		x[5] = m1.x[1] * m2.x[4] + m1.x[5] * m2.x[5] + m1.x[9] * m2.x[6] + m1.x[13] * m2.x[7];
		x[6] = m1.x[2] * m2.x[4] + m1.x[6] * m2.x[5] + m1.x[10] * m2.x[6] + m1.x[14] * m2.x[7];
		x[7] = m1.x[3] * m2.x[4] + m1.x[7] * m2.x[5] + m1.x[11] * m2.x[6] + m1.x[15] * m2.x[7];

		x[8] = m1.x[0] * m2.x[8] + m1.x[4] * m2.x[9] + m1.x[8] * m2.x[10] + m1.x[12] * m2.x[11];
		x[9] = m1.x[1] * m2.x[8] + m1.x[5] * m2.x[9] + m1.x[9] * m2.x[10] + m1.x[13] * m2.x[11];
		x[10] = m1.x[2] * m2.x[8] + m1.x[6] * m2.x[9] + m1.x[10] * m2.x[10] + m1.x[14] * m2.x[11];
		x[11] = m1.x[3] * m2.x[8] + m1.x[7] * m2.x[9] + m1.x[11] * m2.x[10] + m1.x[15] * m2.x[11];

		x[12] = m1.x[0] * m2.x[12] + m1.x[4] * m2.x[13] + m1.x[8] * m2.x[14] + m1.x[12] * m2.x[15];
		x[13] = m1.x[1] * m2.x[12] + m1.x[5] * m2.x[13] + m1.x[9] * m2.x[14] + m1.x[13] * m2.x[15];
		x[14] = m1.x[2] * m2.x[12] + m1.x[6] * m2.x[13] + m1.x[10] * m2.x[14] + m1.x[14] * m2.x[15];
		x[15] = m1.x[3] * m2.x[12] + m1.x[7] * m2.x[13] + m1.x[11] * m2.x[14] + m1.x[15] * m2.x[15];
	}

	Matrix4f operator*( const float f ) const
	{
		Matrix4f m( *this );
		
		for( unsigned int y = 0; y < 4; ++y )
		{
			for( unsigned int x = 0; x < 4; ++x ) 
			{
				m.c[x][y] *= f;
			}
		}

		return m;
	}

	// ----------------------------
	// Vector-Matrix multiplication
	// ----------------------------
	Vec3f operator*( const Vec3f &v ) const
	{
		return Vec3f( v.x * c[0][0] + v.y * c[1][0] + v.z * c[2][0] + c[3][0],
					  v.x * c[0][1] + v.y * c[1][1] + v.z * c[2][1] + c[3][1],
					  v.x * c[0][2] + v.y * c[1][2] + v.z * c[2][2] + c[3][2] );
	}

	Vec4f operator*( const Vec4f &v ) const
	{
		return Vec4f( v.x * c[0][0] + v.y * c[1][0] + v.z * c[2][0] + c[3][0],
					  v.x * c[0][1] + v.y * c[1][1] + v.z * c[2][1] + c[3][1],
					  v.x * c[0][2] + v.y * c[1][2] + v.z * c[2][2] + c[3][2],
					  v.x * c[0][3] + v.y * c[1][3] + v.z * c[2][3] + c[3][3] );
	}
	
	// ---------------
	// Transformations
	// ---------------
	void translate( const float x, const float y, const float z )
	{
		Matrix4f m;
		
		m.c[3][0] = x;
		m.c[3][1] = y;
		m.c[3][2] = z;

		*this = m * *this;
	}

	void scale( const float x, const float y, const float z )
	{
		Matrix4f m;
		
		m.c[0][0] = x;
		m.c[1][1] = y;
		m.c[2][2] = z;

		*this = m * *this;
	}

	void rotate( const float x, const float y, const float z )
	{
		// Rotation order: YXZ [* Vector]
		*this = Matrix4f( Quaternion( x, y, z ) ) * *this;
	}

	// ---------------
	// Other
	// ---------------

	Matrix4f transposed() const
	{
		Matrix4f m( *this );
		
		for( unsigned int y = 0; y < 4; ++y )
		{
			for( unsigned int x = y + 1; x < 4; ++x ) 
			{
				float tmp = m.c[x][y];
				m.c[x][y] = m.c[y][x];
				m.c[y][x] = tmp;
			}
		}

		return m;
	}

	float determinant() const
	{
		return 
			c[0][3]*c[1][2]*c[2][1]*c[3][0] - c[0][2]*c[1][3]*c[2][1]*c[3][0] - c[0][3]*c[1][1]*c[2][2]*c[3][0] + c[0][1]*c[1][3]*c[2][2]*c[3][0] +
			c[0][2]*c[1][1]*c[2][3]*c[3][0] - c[0][1]*c[1][2]*c[2][3]*c[3][0] - c[0][3]*c[1][2]*c[2][0]*c[3][1] + c[0][2]*c[1][3]*c[2][0]*c[3][1] +
			c[0][3]*c[1][0]*c[2][2]*c[3][1] - c[0][0]*c[1][3]*c[2][2]*c[3][1] - c[0][2]*c[1][0]*c[2][3]*c[3][1] + c[0][0]*c[1][2]*c[2][3]*c[3][1] +
			c[0][3]*c[1][1]*c[2][0]*c[3][2] - c[0][1]*c[1][3]*c[2][0]*c[3][2] - c[0][3]*c[1][0]*c[2][1]*c[3][2] + c[0][0]*c[1][3]*c[2][1]*c[3][2] +
			c[0][1]*c[1][0]*c[2][3]*c[3][2] - c[0][0]*c[1][1]*c[2][3]*c[3][2] - c[0][2]*c[1][1]*c[2][0]*c[3][3] + c[0][1]*c[1][2]*c[2][0]*c[3][3] +
			c[0][2]*c[1][0]*c[2][1]*c[3][3] - c[0][0]*c[1][2]*c[2][1]*c[3][3] - c[0][1]*c[1][0]*c[2][2]*c[3][3] + c[0][0]*c[1][1]*c[2][2]*c[3][3];
	}

	Matrix4f inverted() const
	{
		Matrix4f m( false );

		float d = determinant();
		if( d == 0 ) return m;
		d = 1 / d;
		
		m.c[0][0] = d * (c[1][2]*c[2][3]*c[3][1] - c[1][3]*c[2][2]*c[3][1] + c[1][3]*c[2][1]*c[3][2] - c[1][1]*c[2][3]*c[3][2] - c[1][2]*c[2][1]*c[3][3] + c[1][1]*c[2][2]*c[3][3]);
		m.c[0][1] = d * (c[0][3]*c[2][2]*c[3][1] - c[0][2]*c[2][3]*c[3][1] - c[0][3]*c[2][1]*c[3][2] + c[0][1]*c[2][3]*c[3][2] + c[0][2]*c[2][1]*c[3][3] - c[0][1]*c[2][2]*c[3][3]);
		m.c[0][2] = d * (c[0][2]*c[1][3]*c[3][1] - c[0][3]*c[1][2]*c[3][1] + c[0][3]*c[1][1]*c[3][2] - c[0][1]*c[1][3]*c[3][2] - c[0][2]*c[1][1]*c[3][3] + c[0][1]*c[1][2]*c[3][3]);
		m.c[0][3] = d * (c[0][3]*c[1][2]*c[2][1] - c[0][2]*c[1][3]*c[2][1] - c[0][3]*c[1][1]*c[2][2] + c[0][1]*c[1][3]*c[2][2] + c[0][2]*c[1][1]*c[2][3] - c[0][1]*c[1][2]*c[2][3]);
		m.c[1][0] = d * (c[1][3]*c[2][2]*c[3][0] - c[1][2]*c[2][3]*c[3][0] - c[1][3]*c[2][0]*c[3][2] + c[1][0]*c[2][3]*c[3][2] + c[1][2]*c[2][0]*c[3][3] - c[1][0]*c[2][2]*c[3][3]);
		m.c[1][1] = d * (c[0][2]*c[2][3]*c[3][0] - c[0][3]*c[2][2]*c[3][0] + c[0][3]*c[2][0]*c[3][2] - c[0][0]*c[2][3]*c[3][2] - c[0][2]*c[2][0]*c[3][3] + c[0][0]*c[2][2]*c[3][3]);
		m.c[1][2] = d * (c[0][3]*c[1][2]*c[3][0] - c[0][2]*c[1][3]*c[3][0] - c[0][3]*c[1][0]*c[3][2] + c[0][0]*c[1][3]*c[3][2] + c[0][2]*c[1][0]*c[3][3] - c[0][0]*c[1][2]*c[3][3]);
		m.c[1][3] = d * (c[0][2]*c[1][3]*c[2][0] - c[0][3]*c[1][2]*c[2][0] + c[0][3]*c[1][0]*c[2][2] - c[0][0]*c[1][3]*c[2][2] - c[0][2]*c[1][0]*c[2][3] + c[0][0]*c[1][2]*c[2][3]);
		m.c[2][0] = d * (c[1][1]*c[2][3]*c[3][0] - c[1][3]*c[2][1]*c[3][0] + c[1][3]*c[2][0]*c[3][1] - c[1][0]*c[2][3]*c[3][1] - c[1][1]*c[2][0]*c[3][3] + c[1][0]*c[2][1]*c[3][3]);
		m.c[2][1] = d * (c[0][3]*c[2][1]*c[3][0] - c[0][1]*c[2][3]*c[3][0] - c[0][3]*c[2][0]*c[3][1] + c[0][0]*c[2][3]*c[3][1] + c[0][1]*c[2][0]*c[3][3] - c[0][0]*c[2][1]*c[3][3]);
		m.c[2][2] = d * (c[0][1]*c[1][3]*c[3][0] - c[0][3]*c[1][1]*c[3][0] + c[0][3]*c[1][0]*c[3][1] - c[0][0]*c[1][3]*c[3][1] - c[0][1]*c[1][0]*c[3][3] + c[0][0]*c[1][1]*c[3][3]);
		m.c[2][3] = d * (c[0][3]*c[1][1]*c[2][0] - c[0][1]*c[1][3]*c[2][0] - c[0][3]*c[1][0]*c[2][1] + c[0][0]*c[1][3]*c[2][1] + c[0][1]*c[1][0]*c[2][3] - c[0][0]*c[1][1]*c[2][3]);
		m.c[3][0] = d * (c[1][2]*c[2][1]*c[3][0] - c[1][1]*c[2][2]*c[3][0] - c[1][2]*c[2][0]*c[3][1] + c[1][0]*c[2][2]*c[3][1] + c[1][1]*c[2][0]*c[3][2] - c[1][0]*c[2][1]*c[3][2]);
		m.c[3][1] = d * (c[0][1]*c[2][2]*c[3][0] - c[0][2]*c[2][1]*c[3][0] + c[0][2]*c[2][0]*c[3][1] - c[0][0]*c[2][2]*c[3][1] - c[0][1]*c[2][0]*c[3][2] + c[0][0]*c[2][1]*c[3][2]);
		m.c[3][2] = d * (c[0][2]*c[1][1]*c[3][0] - c[0][1]*c[1][2]*c[3][0] - c[0][2]*c[1][0]*c[3][1] + c[0][0]*c[1][2]*c[3][1] + c[0][1]*c[1][0]*c[3][2] - c[0][0]*c[1][1]*c[3][2]);
		m.c[3][3] = d * (c[0][1]*c[1][2]*c[2][0] - c[0][2]*c[1][1]*c[2][0] + c[0][2]*c[1][0]*c[2][1] - c[0][0]*c[1][2]*c[2][1] - c[0][1]*c[1][0]*c[2][2] + c[0][0]*c[1][1]*c[2][2]);
		
		return m;
	}

	void decompose( Vec3f &trans, Vec3f &rot, Vec3f &scale ) const
	{
		// Getting translation is trivial
		trans = Vec3f( c[3][0], c[3][1], c[3][2] );

		// Scale is length of columns
		scale.x = sqrt( c[0][0] * c[0][0] + c[0][1] * c[0][1] + c[0][2] * c[0][2] );
		scale.y = sqrt( c[1][0] * c[1][0] + c[1][1] * c[1][1] + c[1][2] * c[1][2] );
		scale.z = sqrt( c[2][0] * c[2][0] + c[2][1] * c[2][1] + c[2][2] * c[2][2] );

		if( scale.x == 0 || scale.y == 0 || scale.z == 0 ) return;

		// Detect negative scale with determinant and flip one arbitrary axis
		if( determinant() < 0 ) scale.x = -scale.x;

		// Combined rotation matrix YXZ
		//
		// Cos[y]*Cos[z]+Sin[x]*Sin[y]*Sin[z]	Cos[z]*Sin[x]*Sin[y]-Cos[y]*Sin[z]	Cos[x]*Sin[y]	
		// Cos[x]*Sin[z]						Cos[x]*Cos[z]						-Sin[x]
		// -Cos[z]*Sin[y]+Cos[y]*Sin[x]*Sin[z]	Cos[y]*Cos[z]*Sin[x]+Sin[y]*Sin[z]	Cos[x]*Cos[y]

		rot.x = asinf( -c[2][1] / scale.z );
		
		// Special case: Cos[x] == 0 (when Sin[x] is +/-1)
		float f = fabsf( c[2][1] / scale.z );
		if( f > 0.999f && f < 1.001f )
		{
			// Pin arbitrarily one of y or z to zero
			// Mathematical equivalent of gimbal lock
			rot.y = 0;
			
			// Now: Cos[x] = 0, Sin[x] = +/-1, Cos[y] = 1, Sin[y] = 0
			// => m[0][0] = Cos[z] and m[1][0] = Sin[z]
			rot.z = atan2f( -c[1][0] / scale.y, c[0][0] / scale.x );
		}
		// Standard case
		else
		{
			rot.y = atan2f( c[2][0] / scale.z, c[2][2] / scale.z );
			rot.z = atan2f( c[0][1] / scale.x, c[1][1] / scale.y );
		}
	}

	Vec4f getCol( unsigned int col )
	{
		return Vec4f( x[col * 4 + 0], x[col * 4 + 1], x[col * 4 + 2], x[col * 4 + 3] );
	}

	Vec4f getRow( unsigned int row )
	{
		return Vec4f( x[row + 0], x[row + 4], x[row + 8], x[row + 12] );
	}
};


// -------------------------------------------------------------------------------------------------
// Plane
// -------------------------------------------------------------------------------------------------

class Plane
{
public:
	Vec3f normal; 
	float dist;

	// ------------
	// Constructors
	// ------------
	Plane() 
	{ 
		normal.x = 0; normal.y = 0; normal.z = 0; dist = 0; 
	};

	explicit Plane( const float a, const float b, const float c, const float d )
	{
		normal = Vec3f( a, b, c );
		float len = normal.length();
		normal /= len;	// Normalize
		dist = d / len;
	}

	Plane( const Vec3f &v0, const Vec3f &v1, const Vec3f &v2 )
	{
		normal = v1 - v0;
		normal = normal.crossProduct( v2 - v0 ).normalized();
		dist = -(normal * v0);
	}

	// ----------------
	// Other operations
	// ----------------
	float distToPoint( const Vec3f &v ) const
	{
		return (normal * v) + dist;
	}
};


// -------------------------------------------------------------------------------------------------
// Intersection
// -------------------------------------------------------------------------------------------------

inline bool rayTriangleIntersection( const Vec3f &rayPos, const Vec3f &rayDir, 
									 const Vec3f &vert0, const Vec3f &vert1, const Vec3f &vert2,
									 Vec3f &intPoint, Vec3f &normal ) 
{
	// Calculate plane equation
	Plane plane( vert0, vert1, vert2 );

	// Calculate distance of ray origin and ray end from plane
	float dist0 = plane.distToPoint( rayPos );
	float dist1 = plane.distToPoint( rayPos + rayDir );

	// Assume one sided polygons (TODO)
	//if( dist0 < 0 ) return false;
	
	// Check that ray is intersecting the plane (start and end point are on different sides of plane)
	if( (dist0 > 0 && dist1 > 0) || (dist0 < 0 && dist1 < 0) ) return false;

	// Calculate parametric intersection value (division by 0 is not possible)
	float t = -dist0 / (plane.normal * rayDir);

	// Calculate intersection point
	intPoint = rayPos + rayDir * t;

	// Check if intersection point lies in triangle
	Vec3f v0 = vert0 - intPoint, v1 = vert1 - intPoint, v2 = vert2 - intPoint;
	if( v1.crossProduct(v0) * plane.normal < 0 ) return false;
	if( v2.crossProduct(v1) * plane.normal < 0 ) return false;
	if( v0.crossProduct(v2) * plane.normal < 0 ) return false;

	normal = plane.normal;

	return true;
}


inline bool rayAABBIntersection( const Vec3f &rayPos, const Vec3f &rayDir, 
								 const Vec3f &mins, const Vec3f &maxs ) 
{
	// Check if ray is in bounding box
	if( rayPos.x > mins.x - Math::Epsilon && rayPos.x < maxs.x + Math::Epsilon &&
		rayPos.y > mins.y - Math::Epsilon && rayPos.y < maxs.y + Math::Epsilon &&
		rayPos.z > mins.z - Math::Epsilon && rayPos.z < maxs.z + Math::Epsilon )
	{
		return true;
	}
	
	for( unsigned int i = 0; i < 6; ++i )
	{
		// Calculate plane equation
		Plane plane;

		if( i == 0 )		plane = Plane(  0,  0, -1,  mins.z );
		else if( i == 1 )	plane = Plane(  0,  0,  1, -maxs.z );
		else if( i == 2 )	plane = Plane( -1,  0,  0,  mins.x );
		else if( i == 3 )	plane = Plane(  1,  0,  0, -maxs.x );
		else if( i == 4 )	plane = Plane(  0, -1,  0,  mins.y );
		else if( i == 5 )	plane = Plane(  0,  1,  0, -maxs.y );

		// Check if ray is facing plane (this test is not necessary, only for performance reasons)
		if( rayDir * plane.normal >= 0 ) continue;
		
		// Calculate distance of ray origin and ray end from plane
		float dist0 = plane.distToPoint( rayPos );
		float dist1 = plane.distToPoint( rayPos + rayDir );

		// Check that ray is intersecting the plane (start and end point are on different sides of plane)
		if( (dist0 > 0 && dist1 > 0) || (dist0 < 0 && dist1 < 0) )
		{	
			continue;
		}

		// Calculate parametric intersection value (division by 0 is not possible)
		float t = -dist0 / (plane.normal * rayDir);

		// Calculate intersection point
		Vec3f intPoint = rayPos + rayDir * t;

		// Check that intersection point is within box bounds
		if( intPoint.x > mins.x - Math::Epsilon && intPoint.x < maxs.x + Math::Epsilon &&
			intPoint.y > mins.y - Math::Epsilon && intPoint.y < maxs.y + Math::Epsilon &&
			intPoint.z > mins.z - Math::Epsilon && intPoint.z < maxs.z + Math::Epsilon )
		{
			return true;
		}
	}

	return false;
}
